@@ -23,11 +23,7 @@
 #include <linux/kernel.h>
 #include <linux/version.h>
 #include <linux/sched.h>
-#ifdef KERNEL_2_2
-#   include <linux/slab.h>
-#else
-#   include <linux/malloc.h>
-#endif
+#include <linux/slab.h>
 #include <linux/poll.h>
 
 #include <linux/netdevice.h>
@@ -45,6 +41,7 @@
 #include <linux/ip.h>
 #include <linux/tcp.h>
 #include <net/tcp.h>
+#include <net/ipv6.h>
 
 #ifdef CONFIG_NET_RADIO
 #   include <linux/wireless.h>
@@ -116,6 +113,10 @@ static void VNetBridgePortsChanged(VNetJack *this);
 static int  VNetBridgeIsBridged(VNetJack *this);
 static int  VNetBridgeProcRead(char *page, char **start, off_t off,
                                int count, int *eof, void *data);
+static void VNetBridgeComputeHeaderPosIPv6(struct sk_buff *skb);
+static PacketStatus VNetCallSMACFunc(struct SMACState *state,
+                                     struct sk_buff **skb, void *startOfData,
+                                     SMACFunc func, unsigned int len);
 
 
 /*
@@ -502,17 +503,19 @@ VNetBridgeFree(VNetJack *this) // IN: jack to free
  *----------------------------------------------------------------------
  */
 
-PacketStatus
+static PacketStatus
 VNetCallSMACFunc(struct SMACState *state, // IN: pointer to state
                  struct sk_buff **skb,    // IN/OUT: packet to process
                  void *startOfData,       // IN: points to start of data
-                 SMACFunc func)           // IN: function to be called
+                 SMACFunc func,           // IN: function to be called
+                 unsigned int len)        // IN: length including ETH header
 {
    SMACPackets packets = { {0} };
    PacketStatus status;
 
    packets.orig.skb = *skb;
    packets.orig.startOfData = startOfData;
+   packets.orig.len = len;
 
    status = func(state, &packets);
    if (status != PacketStatusForwardPacket) {
@@ -605,7 +608,7 @@ VNetBridgeReceiveFromVNet(VNetJack        *this, // IN: jack
 
    if (bridge->smac) {
       if (VNetCallSMACFunc(bridge->smac, &skb, skb->data,
-                           SMAC_CheckPacketToHost) !=
+                           SMAC_CheckPacketToHost, skb->len) !=
           PacketStatusForwardPacket) {
          LOG(4, (KERN_NOTICE "bridge-%s: packet dropped .\n",
                  bridge->name));
@@ -851,9 +854,9 @@ VNetBridgeIsDeviceWireless(struct net_device *dev) //IN: sock
    return dev->wireless_handlers != NULL;
 #elif !defined(CONFIG_NET_RADIO)
    return FALSE;
-#elif WIRELESS_EXT > 19
+#elif defined WIRELESS_EXT && WIRELESS_EXT > 19
    return dev->wireless_handlers != NULL;
-#elif WIRELESS_EXT > 12
+#elif defined WIRELESS_EXT && WIRELESS_EXT > 12
    return dev->wireless_handlers != NULL || dev->get_wireless_stats != NULL;
 #else
    return dev->get_wireless_stats != NULL;
@@ -1202,38 +1205,20 @@ VNetBridgeNotify(struct notifier_block *this, // IN: callback data (bridge)
       LOG(2, (KERN_DEBUG "bridge-%s: interface %s is going down\n",
               bridge->name, dev->name));
       if (dev == bridge->dev) {
-         LOG(1, (KERN_DEBUG "bridge-%s: disabling the bridge\n",
+         LOG(1, (KERN_DEBUG "bridge-%s: disabling the bridge on dev down\n",
                  bridge->name));
          VNetBridgeDown(bridge, FALSE);
       }
       break;
 
-   case NETDEV_CHANGE:
-      LOG(2, (KERN_DEBUG "bridge-%s: interface %s is changing\n",
-              bridge->name, dev->name));
-      if (dev == bridge->dev) {
-         LOG(1, (KERN_DEBUG "bridge-%s: disabling the bridge\n",
-                 bridge->name));
-         VNetBridgeDown(bridge, FALSE);
-      } else if (bridge->dev == NULL && VNetBridgeDevCompatible(bridge, dev)) {
-         int errno;
-
-         LOG(1, (KERN_DEBUG "bridge-%s: enabling the bridge\n", bridge->name));
-	 errno = VNetBridgeUp(bridge, FALSE);
-         if (errno != 0) {
-            VNetBridgeNotifyLogBridgeUpError(errno, bridge->name, dev->name);
-         }
-      }
-      break;
-
-
    case NETDEV_UP:
       LOG(2, (KERN_DEBUG "bridge-%s: interface %s is going up\n",
               bridge->name, dev->name));
       if (bridge->dev == NULL && VNetBridgeDevCompatible(bridge, dev)) {
          int errno;
 
-         LOG(1, (KERN_DEBUG "bridge-%s: enabling the bridge\n", bridge->name));
+         LOG(1, (KERN_DEBUG "bridge-%s: enabling the bridge on dev up\n",
+                 bridge->name));
 	 errno = VNetBridgeUp(bridge, FALSE);
          if (errno != 0) {
             VNetBridgeNotifyLogBridgeUpError(errno, bridge->name, dev->name);
@@ -1251,6 +1236,167 @@ VNetBridgeNotify(struct notifier_block *this, // IN: callback data (bridge)
 }
 
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * RangeInLinearSKB --
+ *
+ *      Checks if the given number of bytes from a given offset resides
+ *      within the linear part of the skb.  If not then attempts to
+ *      linearize the skb.
+ *
+ * Results:
+ *      Returns TRUE if the range of bytes is already in the linear
+ *      portion or if linearize succeeded.  Otherwise, returns FALSE if
+ *      the linearize operation fails.
+ *
+ * Side effects:
+ *      As in skb_linearize().
+ *
+ *----------------------------------------------------------------------
+ */
+
+static INLINE_SINGLE_CALLER Bool
+RangeInLinearSKB(struct sk_buff *skb, // IN:
+                 unsigned int start,  // IN:  Start offset
+                 unsigned int length) // IN:  How many bytes
+{
+   if (LIKELY(!compat_skb_is_nonlinear(skb) ||
+              start + length <= compat_skb_headlen(skb))) {
+      /*
+       * Nothing to do.
+       */
+
+      return TRUE;
+   }
+
+   return compat_skb_linearize(skb) == 0;
+}
+
+
+/*
+ * Not all kernel versions have NEXTHDR_MOBILITY defined.
+ */
+
+#ifndef NEXTHDR_MOBILITY
+#  define NEXTHDR_MOBILITY 135 /* Mobility header. */
+#endif
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * VNetBridgeComputeHeaderPosIPv6 --
+ *
+ *      Compute correct position of transport header in IPv6 packets.
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      Transport header pointer updated to point to the PDU contained
+ *      in the packet.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+VNetBridgeComputeHeaderPosIPv6(struct sk_buff *skb) // IN:
+{
+   struct ipv6hdr *ipv6Hdr;
+   unsigned int offset; /* Offset from skb->data. */
+   unsigned int headerLen; /* Length of current header. */
+   uint8 nextHeader;
+
+   /*
+    * Check if the start of the network header is within the linear part of
+    * skb.  If not, then linearize the skb.
+    */
+
+   if (UNLIKELY(compat_skb_network_header(skb) < skb->data ||
+                compat_skb_network_header(skb) >= skb->data +
+                                                  compat_skb_headlen(skb))) {
+      if (compat_skb_linearize(skb)) {
+         return; /* Bail out. */
+      }
+   }
+
+   offset = compat_skb_network_offset(skb);
+   if (!RangeInLinearSKB(skb, offset, sizeof *ipv6Hdr)) {
+      return; /* Bail out. */
+   }
+
+   ipv6Hdr = (struct ipv6hdr *)compat_skb_network_header(skb);
+   headerLen = sizeof *ipv6Hdr;
+   offset += headerLen; /* End of IPv6 header (not including extensions). */
+
+   /*
+    * All IPv6 extension headers begin with a "next header" field (one byte),
+    * and most of them have a "header length" field (as the 2nd byte).  In each
+    * iteration, we find the length of the extension header and add it to
+    * offset from the beginning of skb.  And, in each iteration we update the
+    * next header variable.  When we return from the following for loop, offset
+    * would have incremented by the length of each of the extension header,
+    * and next header type will be something else than an IPv6 extension header
+    * signifying that we have walked through the entire IPv6 header.  We set
+    * the transport header's offset to the value of this offset before exiting
+    * the for loop.
+    */
+
+   nextHeader = ipv6Hdr->nexthdr;
+   for (;;) {
+      switch (nextHeader) {
+      case NEXTHDR_HOP:
+      case NEXTHDR_ROUTING:
+      case NEXTHDR_AUTH:
+      case NEXTHDR_DEST:
+      case NEXTHDR_MOBILITY:
+         /*
+          * We need to check two bytes in the option header:  next header and
+          * header extension length.
+          */
+
+         if (!RangeInLinearSKB(skb, offset, 2)) {
+            return; /* Bail out. */
+         }
+         headerLen = skb->data[offset + 1];
+         if (nextHeader == NEXTHDR_AUTH) {
+            headerLen = (headerLen + 2) << 2; /* See RFC 2402. */
+         } else {
+            headerLen = (headerLen + 1) << 3; /* See ipv6_optlen(). */
+         }
+
+         break;
+
+      case NEXTHDR_FRAGMENT:
+      case NEXTHDR_ESP:
+      case NEXTHDR_NONE:
+         /*
+          * We stop walking if we find the fragment header (NEXTHDR_FRAGMENT).
+          * If the payload is encrypted we may not know the start of the
+          * transport header [1].  So, we just return.  Same applies when
+          * nothing follows this header (NEXTHDR_NONE).
+          * [1]:  http://www.cu.ipv6tf.org/literatura/chap8.pdf
+          */
+
+         return;
+
+       default:
+         /*
+          * We have walked through all IPv6 extension headers.  Let's set the
+          * transport header and return.
+          */
+
+         compat_skb_set_transport_header(skb, offset);
+         return;
+      }
+
+      nextHeader = skb->data[offset];
+      offset += headerLen;
+   }
+}
+
+
 /*
  *----------------------------------------------------------------------
  *
@@ -1276,18 +1422,22 @@ VNetBridgeComputeHeaderPos(struct sk_buff *skb) // IN: buffer to examine
    }
    switch (be16_to_cpu(skb->protocol)) {
       case ETH_P_IP: {
-            struct iphdr *ipHdr = compat_skb_ip_header(skb);
+         struct iphdr *ipHdr = compat_skb_ip_header(skb);
+
+         compat_skb_set_transport_header(skb, compat_skb_network_offset(skb) +
+                                              ipHdr->ihl * 4);
+         break;
+      }
+
+      case ETH_P_IPV6:
+         VNetBridgeComputeHeaderPosIPv6(skb);
+         break;
 
-            compat_skb_set_transport_header(skb, compat_skb_network_offset(skb) +
-                                                 ipHdr->ihl * 4);
-         }
-	 return;
       default:
          LOG(3, (KERN_DEBUG "Unknown EII protocol %04X: csum at %d\n",
-		 be16_to_cpu(skb->protocol), compat_skb_csum_offset(skb)));
-	 break;
+                 be16_to_cpu(skb->protocol), compat_skb_csum_offset(skb)));
+         break;
    }
-   return;
 }
 
 
@@ -1564,8 +1714,15 @@ VNetBridgeReceiveFromDev(struct sk_buff *skb,         // IN: packet to receive
 #  endif
 
    if (bridge->smac) {
+      /*
+       * Wireless driver processes the packet and processes the ethernet header
+       * and the length is reduced by the amount. We need the raw ethernet
+       * packet length hence add the ethernet header length for incoming
+       * packets.
+       */
+
       if (VNetCallSMACFunc(bridge->smac, &skb, compat_skb_mac_header(skb),
-                           SMAC_CheckPacketFromHost) !=
+                           SMAC_CheckPacketFromHost, skb->len + ETH_HLEN) !=
           PacketStatusForwardPacket) {
          LOG(4, (KERN_NOTICE "bridge-%s: packet dropped .\n",
                  bridge->name));
@@ -1573,7 +1730,6 @@ VNetBridgeReceiveFromDev(struct sk_buff *skb,         // IN: packet to receive
       }
    }
 
-#ifdef KERNEL_2_3_15
    skb = skb_share_check(skb, GFP_ATOMIC);
    if (!skb) {
       return 0;
@@ -1583,7 +1739,6 @@ VNetBridgeReceiveFromDev(struct sk_buff *skb,         // IN: packet to receive
     * Unbelievable... Caller sets h.raw = nh.raw before invoking us...
     */
    VNetBridgeComputeHeaderPos(skb);
-#endif
 #endif
 
    skb_push(skb, skb->data - compat_skb_mac_header(skb));
