@@ -31,9 +31,14 @@ struct InitCrossGDT;
 extern Bool Task_AllocCrossGDT(struct InitBlock *initBlock);
 extern int Task_InitCrosspage(VMDriver *vm, struct InitBlock *params);
 extern int Task_InitCrossGDT(struct InitCrossGDT *initCrossGDT);
-extern Bool Task_Switch(VMDriver *vm, Vcpuid vcpuid);
+extern void Task_Switch(VMDriver *vm, Vcpuid vcpuid);
 extern Bool Task_Initialize(void);
 extern void Task_Terminate(void);
+extern MPN Task_GetDummyVMCS(int vmcsId);
+extern MPN Task_GetRootVMCS(uint32 pCPU);
+extern Bool Task_IsVMXDisabledOnAllCPUs(void);
+extern void Task_FreeVMCS(void);
+
 #endif
 
 
