@@ -24,6 +24,8 @@
 
 #if defined(linux) && !defined(VMKERNEL)
 #   include "driver-config.h"
+#   define EXPORT_SYMTAB
+#   include "compat_module.h"
 #   include <linux/string.h> /* memset() in the kernel */
 #elif defined(WINNT_DDK)
 #   include <ntddk.h>
@@ -36,12 +38,8 @@
 #include "vmci_kernel_if.h"
 
 #include "vm_assert.h"
-#ifdef VMKERNEL
-#define LOGLEVEL_MODULE_LEN 0
-#define LOGLEVEL_MODULE VMCIVMK
-#include "log.h"
-#endif // VMKERNEL
 #include "vmci_infrastructure.h"
+#include "vmciHostKernelAPI.h"
 #include "vmciContext.h"
 #include "vmciDatagram.h"
 #include "vmciDriver.h"
@@ -59,7 +57,9 @@
 
 static void VMCIContextFreeContext(VMCIContext *context);
 static Bool VMCIContextExists(VMCIId cid);
-static int VMCIContextFireNotification(VMCIId contextID);
+static int VMCIContextFireNotification(VMCIId contextID,
+                                       VMCIPrivilegeFlags privFlags,
+                                       const char *domain);
 
 /*
  * List of current VMCI contexts.
@@ -72,13 +72,132 @@ static struct {
 } contextList;
 
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * VMCIContextSignalNotify --
+ *
+ *      Sets the notify flag to TRUE.  Assumes that the context lock is
+ *      held.
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static INLINE void
+VMCIContextSignalNotify(VMCIContext *context) // IN:
+{
+#ifndef VMX86_SERVER
+   if (context->notify) {
+      *context->notify = TRUE;
+   }
+#endif
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * VMCIContextClearNotify --
+ *
+ *      Sets the notify flag to FALSE.  Assumes that the context lock is
+ *      held.
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static INLINE void
+VMCIContextClearNotify(VMCIContext *context) // IN:
+{
+#ifndef VMX86_SERVER
+   if (context->notify) {
+      *context->notify = FALSE;
+   }
+#endif
+}
+
+
+#ifndef VMX86_SERVER
+/*
+ *----------------------------------------------------------------------
+ *
+ * VMCIContext_CheckAndSignalNotify --
+ *
+ *      Sets the context's notify flag iff datagrams are pending for this
+ *      context.  Called from VMCISetupNotify(). 
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+VMCIContext_CheckAndSignalNotify(VMCIContext *context) // IN:
+{
+   VMCILockFlags flags;
+
+   ASSERT(context);
+   VMCI_GrabLock(&contextList.lock, &flags);
+   if (context->pendingDatagrams) {
+      VMCIContextSignalNotify(context);
+   }
+   VMCI_ReleaseLock(&contextList.lock, flags);
+}
+#endif
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * VMCIContextGetDomainName --
+ *
+ *      Internal function for retrieving a context domain name, if
+ *      supported by the platform. The returned pointer can only be
+ *      assumed valid while a reference count is held on the given
+ *      context.
+ *
+ * Results:
+ *      Pointer to name if appropriate. NULL otherwise.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static INLINE char *
+VMCIContextGetDomainName(VMCIContext *context) // IN
+{
+#ifdef VMKERNEL
+   return context->domainName;
+#else
+   return NULL;
+#endif
+}
+
+
 /*
  *----------------------------------------------------------------------
  *
  * VMCIContext_Init --
  *
  *      Initializes the VMCI context module.
-*
+ *
  * Results:
  *      None.
  *
@@ -93,9 +212,9 @@ VMCIContext_Init(void)
 {
    contextList.head = NULL;
    VMCI_InitLock(&contextList.lock, "VMCIContextListLock",
-		 VMCI_LOCK_RANK_HIGH);
+		 VMCI_LOCK_RANK_HIGHER);
    VMCI_InitLock(&contextList.firingLock, "VMCIContextFiringLock",
-		 VMCI_LOCK_RANK_MIDDLE);
+		 VMCI_LOCK_RANK_MIDDLE_LOW);
 
    return VMCI_SUCCESS;
 }
@@ -107,7 +226,7 @@ VMCIContext_Init(void)
  * VMCIContext_Exit --
  *
  *      Cleans up the contexts module.
-*
+ *
  * Results:
  *      None.
  *
@@ -145,6 +264,7 @@ int
 VMCIContext_InitContext(VMCIId cid,                   // IN
 			VMCIPrivilegeFlags privFlags, // IN
                         uintptr_t eventHnd,           // IN
+                        int userVersion,              // IN: User's vers no.
                         VMCIContext **outContext)     // OUT
 {
    VMCILockFlags flags;
@@ -152,12 +272,17 @@ VMCIContext_InitContext(VMCIId cid,                   // IN
    int result;
 
    if (privFlags & ~VMCI_PRIVILEGE_ALL_FLAGS) {
-      Log(LGPFX "Invalid flag for VMCI context.\n");
+      VMCILOG((LGPFX"Invalid flag for VMCI context.\n"));
       return VMCI_ERROR_INVALID_ARGS;
    }
+
+   if (userVersion == 0) {
+      return VMCI_ERROR_INVALID_ARGS;
+   }
+
    context = VMCI_AllocKernelMem(sizeof *context, VMCI_MEMORY_NONPAGED);
    if (context == NULL) {
-      Log(LGPFX "Failed to allocate memory for VMCI context.\n");
+      VMCILOG((LGPFX"Failed to allocate memory for VMCI context.\n"));
       return VMCI_ERROR_NO_MEM;
    }
    context->wellKnownArray = NULL;
@@ -167,6 +292,7 @@ VMCIContext_InitContext(VMCIId cid,                   // IN
    context->datagramQueue = NULL;
    context->pendingDatagrams = 0;
    context->datagramQueueSize = 0;
+   context->userVersion = userVersion;
 
    context->wellKnownArray = VMCIHandleArray_Create(0);
    if (context->wellKnownArray == NULL) {
@@ -194,7 +320,7 @@ VMCIContext_InitContext(VMCIId cid,                   // IN
 
    VMCI_InitLock(&context->lock,
                  "VMCIContextLock",
-                 VMCI_LOCK_RANK_HIGH);
+                 VMCI_LOCK_RANK_HIGHER);
    Atomic_Write(&context->refCount, 1);
    
    /* Inititialize host-specific VMCI context. */
@@ -236,6 +362,13 @@ VMCIContext_InitContext(VMCIId cid,                   // IN
    VMCIContext_SetDomainName(context, "");
 #endif
 
+#ifndef VMX86_SERVER
+   context->notify = NULL;
+#  ifdef __linux__
+   context->notifyPage = NULL;
+#  endif
+#endif
+
    *outContext = context;
    return VMCI_SUCCESS;
 
@@ -318,7 +451,8 @@ VMCIContextFreeContext(VMCIContext *context)  // IN
    VMCIHandle tempHandle;
 
    /* Fire event to all contexts interested in knowing this context is dying. */
-   VMCIContextFireNotification(context->cid);
+   VMCIContextFireNotification(context->cid, context->privFlags,
+                               VMCIContextGetDomainName(context));
 
    /*
     * Cleanup all wellknown mappings owned by context. Ideally these would
@@ -368,8 +502,8 @@ VMCIContextFreeContext(VMCIContext *context)  // IN
    ASSERT(context->groupArray);
    tempHandle = VMCIHandleArray_RemoveTail(context->groupArray);
    while (!VMCI_HANDLE_EQUAL(tempHandle, VMCI_INVALID_HANDLE)) {
-      VMCI_LOG((LGPFX"Removing context 0x%x from group 0x%"FMT64"x during release.\n",
-                context->cid, tempHandle));
+      VMCI_DEBUG_LOG((LGPFX"Removing context 0x%x from group 0x%"FMT64
+                      "x during release.\n", context->cid, tempHandle));
       VMCIGroup_RemoveMember(tempHandle, VMCI_MAKE_HANDLE(context->cid,
                                                       VMCI_CONTEXT_RESOURCE_ID));
       tempHandle = VMCIHandleArray_RemoveTail(context->groupArray);
@@ -394,6 +528,12 @@ VMCIContextFreeContext(VMCIContext *context)  // IN
    VMCIHandleArray_Destroy(context->queuePairArray);
    VMCI_CleanupLock(&context->lock);
    VMCIHost_ReleaseContext(&context->hostContext);
+#ifndef VMX86_SERVER
+#  ifdef __linux__
+   /* TODO Windows and Mac OS. */
+   VMCIUnsetNotify(context);
+#  endif
+#endif
    VMCI_FreeKernelMem(context, sizeof *context);
 }
 
@@ -493,14 +633,14 @@ VMCIContext_EnqueueDatagram(VMCIId cid,        // IN: Target VM
    /* Get the target VM's VMCI context. */
    context = VMCIContext_Get(cid);
    if (context == NULL) {
-      Log(LGPFX "Invalid cid\n");
+      VMCILOGThrottled((LGPFX"Invalid cid.\n"));
       return VMCI_ERROR_INVALID_ARGS;
    }
 
    /* Allocate guest call entry and add it to the target VM's queue. */
    dqEntry = VMCI_AllocKernelMem(sizeof *dqEntry, VMCI_MEMORY_NONPAGED);
    if (dqEntry == NULL) {
-      Log(LGPFX "Failed to allocate memory for datagram.\n");
+      VMCILOG((LGPFX"Failed to allocate memory for datagram.\n"));
       VMCIContext_Release(context);
       return VMCI_ERROR_NO_MEM;
    }
@@ -525,13 +665,14 @@ VMCIContext_EnqueueDatagram(VMCIId cid,        // IN: Target VM
       VMCI_ReleaseLock(&context->lock, flags);
       VMCIContext_Release(context);
       VMCI_FreeKernelMem(dqEntry, sizeof *dqEntry);
-      Log(LGPFX "Context 0x%x receive queue is full.\n", cid);
+      VMCILOGThrottled((LGPFX"Context 0x%x receive queue is full.\n", cid));
       return VMCI_ERROR_NO_RESOURCES;
    }
 
    LIST_QUEUE(&dqEntry->listItem, &context->datagramQueue);
    context->pendingDatagrams++;
    context->datagramQueueSize += vmciDgSize;
+   VMCIContextSignalNotify(context);
    VMCIHost_SignalCall(&context->hostContext);
    VMCI_ReleaseLock(&context->lock, flags);
    VMCIContext_Release(context);
@@ -689,7 +830,7 @@ VMCIContext_Release(VMCIContext *context)  // IN
    uint32 refCount;
    ASSERT(context);
    refCount = Atomic_FetchAndDec(&context->refCount);
-   if(refCount == 1) {
+   if (refCount == 1) {
       VMCIContextFreeContext(context);
    }
 }
@@ -734,8 +875,9 @@ VMCIContext_DequeueDatagram(VMCIContext *context, // IN
    VMCI_GrabLock(&context->lock, &flags);
    if (context->pendingDatagrams == 0) {
       VMCIHost_ClearCall(&context->hostContext);
+      VMCIContextClearNotify(context);
       VMCI_ReleaseLock(&context->lock, flags);
-      VMCI_LOG((LGPFX "No datagrams pending.\n"));
+      VMCI_DEBUG_LOG((LGPFX"No datagrams pending.\n"));
       return VMCI_ERROR_NO_MORE_DATAGRAMS;
    }
 
@@ -749,8 +891,8 @@ VMCIContext_DequeueDatagram(VMCIContext *context, // IN
    if (*maxSize < VMCI_DG_SIZE(dqEntry->dg)) {
       *maxSize = VMCI_DG_SIZE(dqEntry->dg);
       VMCI_ReleaseLock(&context->lock, flags);
-      Log(LGPFX "Caller's buffer is too small. It must be at least %"FMTSZ"d "
-          "bytes.\n", *maxSize);
+      VMCILOG((LGPFX"Caller's buffer is too small. It must be at "
+               "least %"FMTSZ"d bytes.\n", *maxSize));
       return VMCI_ERROR_NO_MEM;
    }
    
@@ -759,6 +901,7 @@ VMCIContext_DequeueDatagram(VMCIContext *context, // IN
    context->datagramQueueSize -= VMCI_DG_SIZE(dqEntry->dg);
    if (context->pendingDatagrams == 0) {
       VMCIHost_ClearCall(&context->hostContext);
+      VMCIContextClearNotify(context);
       rv = VMCI_SUCCESS;
    } else {
       /*
@@ -814,6 +957,40 @@ VMCIContext_GetId(VMCIContext *context) // IN:
 }
 
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * VMCIContext_GetPrivFlagsInt --
+ *
+ *      Internal function that retrieves the privilege flags of the given
+ *      VMCI context ID.
+ *
+ * Results:
+ *     Context's privilege flags.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+VMCIPrivilegeFlags
+VMCIContext_GetPrivFlagsInt(VMCIId contextID)  // IN
+{
+   VMCIPrivilegeFlags flags;
+   VMCIContext *context;
+
+   context = VMCIContext_Get(contextID);
+   if (!context) {
+      return VMCI_LEAST_PRIVILEGE_FLAGS;
+   }
+   flags = context->privFlags;
+   VMCIContext_Release(context);
+   return flags;
+}
+
+
+#ifndef VMKERNEL
 /*
  *----------------------------------------------------------------------
  *
@@ -830,20 +1007,16 @@ VMCIContext_GetId(VMCIContext *context) // IN:
  *----------------------------------------------------------------------
  */
 
+#if defined(linux)
+EXPORT_SYMBOL(VMCIContext_GetPrivFlags);
+#endif
+
 VMCIPrivilegeFlags
 VMCIContext_GetPrivFlags(VMCIId contextID)  // IN
 {
-   VMCIPrivilegeFlags flags;
-   VMCIContext *context;
-   
-   context = VMCIContext_Get(contextID);
-   if (!context) {
-      return VMCI_LEAST_PRIVILEGE_FLAGS;
-   }
-   flags = context->privFlags;
-   VMCIContext_Release(context);
-   return flags;
+   return VMCIContext_GetPrivFlagsInt(contextID);
 }
+#endif /* !VMKERNEL */
 
 
 /*
@@ -1016,7 +1189,12 @@ VMCIContext_AddNotification(VMCIId contextID,  // IN:
    if (context == NULL) {
       return VMCI_ERROR_NOT_FOUND;
    }
-   
+
+   if (context->privFlags & VMCI_PRIVILEGE_FLAG_RESTRICTED) {
+      result = VMCI_ERROR_NO_ACCESS;
+      goto out;
+   }
+
    notifierHandle = VMCI_MAKE_HANDLE(remoteCID, VMCI_EVENT_HANDLER);
    VMCI_GrabLock(&contextList.firingLock, &firingFlags);
    VMCI_GrabLock(&context->lock, &flags);
@@ -1026,6 +1204,7 @@ VMCIContext_AddNotification(VMCIId contextID,  // IN:
    }
    VMCI_ReleaseLock(&context->lock, flags);
    VMCI_ReleaseLock(&contextList.firingLock, firingFlags);
+out:
    VMCIContext_Release(context);
    return result;
 }
@@ -1093,8 +1272,10 @@ VMCIContext_RemoveNotification(VMCIId contextID,  // IN:
  */
 
 static int
-VMCIContextFireNotification(VMCIId contextID)  // IN:
-{ 
+VMCIContextFireNotification(VMCIId contextID,             // IN
+                            VMCIPrivilegeFlags privFlags, // IN
+                            const char *domain)           // IN
+{
    uint32 i, arraySize;
    ListItem *next;
    VMCILockFlags flags;
@@ -1120,13 +1301,19 @@ VMCIContextFireNotification(VMCIId contextID)  // IN:
    VMCI_GrabLock(&contextList.firingLock, &firingFlags);
    VMCI_GrabLock(&contextList.lock, &flags);
    LIST_SCAN(next, contextList.head) {
-      VMCIContext *subscriberContext = LIST_CONTAINER(next, VMCIContext,
-						      listItem);
-      if (VMCIHandleArray_HasEntry(subscriberContext->notifierArray, 
-				   contextHandle)) {
-	 VMCIHandleArray_AppendEntry(&subscriberArray,
-				     VMCI_MAKE_HANDLE(subscriberContext->cid,
-						      VMCI_EVENT_HANDLER));
+      VMCIContext *subCtx = LIST_CONTAINER(next, VMCIContext, listItem);
+
+      /*
+       * We only deliver notifications of the removal of contexts, if
+       * the two contexts are allowed to interact.
+       */
+
+      if (VMCIHandleArray_HasEntry(subCtx->notifierArray, contextHandle) &&
+          !VMCIDenyInteraction(privFlags, subCtx->privFlags, domain,
+                               VMCIContextGetDomainName(subCtx))) {
+         VMCIHandleArray_AppendEntry(&subscriberArray,
+                                     VMCI_MAKE_HANDLE(subCtx->cid,
+                                                      VMCI_EVENT_HANDLER));
       }
    }
    VMCI_ReleaseLock(&contextList.lock, flags);
@@ -1156,8 +1343,8 @@ VMCIContextFireNotification(VMCIId contextID)  // IN:
       result = VMCIDatagram_Dispatch(VMCI_HYPERVISOR_CONTEXT_ID,
                                      (VMCIDatagram *)eMsg);
       if (result < VMCI_SUCCESS) {
-         Log(LGPFX "Failed to enqueue event datagram %d for context %d.\n",
-             eMsg->eventData.event, eMsg->hdr.dst.context);
+         VMCILOG((LGPFX"Failed to enqueue event datagram %d for context %d.\n",
+                  eMsg->eventData.event, eMsg->hdr.dst.context));
          /* We continue to enqueue on next subscriber. */
       }
    }
@@ -1214,7 +1401,7 @@ VMCIContext_GetCheckpointState(VMCIId contextID,    // IN:
       array = context->wellKnownArray;
       getContextID = FALSE;
    } else {
-      Log(LGPFX "Invalid cpt state type %d.\n", cptType);
+      VMCILOG((LGPFX"Invalid cpt state type %d.\n", cptType));
       result = VMCI_ERROR_INVALID_ARGS;
       goto release;
    }
@@ -1285,7 +1472,7 @@ VMCIContext_SetCheckpointState(VMCIId contextID, // IN:
 
    if (cptType != VMCI_NOTIFICATION_CPT_STATE &&
        cptType != VMCI_WELLKNOWN_CPT_STATE) {
-      Log(LGPFX "Invalid cpt state type %d.\n", cptType);
+      VMCILOG((LGPFX"Invalid cpt state type %d.\n", cptType));
       return VMCI_ERROR_INVALID_ARGS;
    }
 
@@ -1295,12 +1482,12 @@ VMCIContext_SetCheckpointState(VMCIId contextID, // IN:
 	 result = VMCIContext_AddNotification(contextID, currentID);
       } else if (cptType == VMCI_WELLKNOWN_CPT_STATE) {
 	 result = VMCIDatagramRequestWellKnownMap(currentID, contextID, 
-						  VMCIContext_GetPrivFlags(contextID));
+						  VMCIContext_GetPrivFlagsInt(contextID));
       }
    }
    if (result != VMCI_SUCCESS) {
-      Log(LGPFX "Failed to set cpt state type %d, error %d.\n", 
-	  cptType, result);
+      VMCILOG((LGPFX"Failed to set cpt state type %d, error %d.\n", 
+               cptType, result));
    }
    return result;
 }
@@ -1393,4 +1580,83 @@ out:
    return rv;
 }
 
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * VMCIContextID2HostVmID --
+ *
+ *      Maps a context ID to the host specific (process/world) ID
+ *      of the VM/VMX.
+ *
+ * Results:
+ *      VMCI_SUCCESS on success, error code otherwise.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+VMCIContextID2HostVmID(VMCIId contextID,    // IN
+                       void *hostVmID,      // OUT
+                       size_t hostVmIDLen)  // IN
+{
+   VMCIContext *context;
+   VMCIHostVmID vmID;
+   int result;
+
+   context = VMCIContext_Get(contextID);
+   if (!context) {
+      return VMCI_ERROR_NOT_FOUND;
+   }
+   
+   result = VMCIHost_ContextToHostVmID(&context->hostContext, &vmID);
+   if (result == VMCI_SUCCESS) {
+      if (sizeof vmID == hostVmIDLen) {
+         memcpy(hostVmID, &vmID, hostVmIDLen);
+      } else {
+         result = VMCI_ERROR_INVALID_ARGS;
+      }
+   }
+
+   VMCIContext_Release(context);
+
+   return result;
+}
+
 #endif
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * VMCIContext_SupportsHostQP --
+ *
+ *      Can host QPs be connected to this user process.  The answer is
+ *      FALSE unless a sufficient version number has previously been set
+ *      by this caller.
+ *
+ * Results:
+ *      VMCI_SUCCESS on success, error code otherwise.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+Bool
+VMCIContext_SupportsHostQP(VMCIContext *context)    // IN: Context structure
+{
+#ifdef VMKERNEL
+   return TRUE;
+#else
+   if (!context || context->userVersion < VMCI_VERSION_HOSTQP) {
+      return FALSE;
+   }
+   return TRUE;
+#endif
+}
+
