@@ -36,16 +36,11 @@
 #include "vmci_kernel_if.h"
 
 #include "vm_assert.h"
-#ifdef VMKERNEL
-#define LOGLEVEL_MODULE_LEN 0
-#define LOGLEVEL_MODULE VMCIVMK
-#include "log.h"
-#endif // VMKERNEL
 #include "vmci_defs.h"
 #include "vmci_infrastructure.h"
 #include "vmciContext.h"
 #ifndef VMX86_SERVER
-#include "vmciProcess.h"
+#  include "vmciProcess.h"
 #endif // !VMX86_SERVER
 #include "vmciGroup.h"
 #include "vmciDatagram.h"
@@ -140,7 +135,7 @@ DsRequestCb(void *notifyData,  // IN: callback data
    int written, retval;
    VMCIPrivilegeFlags srcPrivFlags;
 
-   VMCI_LOG((LGPFX "Got request from context: %d\n", msg->src.context));
+   VMCI_DEBUG_LOG((LGPFX"Got request from context: %d\n", msg->src.context));
 
    if (VMCIDatagram_GetPrivFlags(msg->src, &srcPrivFlags) != VMCI_SUCCESS) {
       retval = VMCI_ERROR_INVALID_ARGS;
@@ -149,17 +144,17 @@ DsRequestCb(void *notifyData,  // IN: callback data
    replyMsg->dst = msg->src;
    replyMsg->src = dsAPI.handle;
    DsHandleMessage(VMCI_DG_PAYLOAD(msg), VMCI_DG_PAYLOAD(replyMsg), 
-		   VMCI_DS_MAX_MSG_SIZE, &written, msg->src.context,
+                   VMCI_DS_MAX_MSG_SIZE, &written, msg->src.context,
                    srcPrivFlags);
    replyMsg->payloadSize = written;
 
    /* Send reply back to source handle. */
-   retval = VMCIDatagram_Send(replyMsg);
+   retval = VMCIDatagramSendInt(replyMsg);
 done:
    if (retval >= VMCI_SUCCESS) {
-      VMCI_LOG((LGPFX "Successfully replied with %d bytes\n", written));
+      VMCI_DEBUG_LOG((LGPFX"Successfully replied with %d bytes\n", written));
    } else {
-      Log(LGPFX "Failed to reply to request: %d\n", retval);
+      VMCILOG((LGPFX"Failed to reply to request: %d.\n", retval));
    }
 
    return retval;
@@ -508,22 +503,22 @@ VMCIDs_Init(void)
 
    /* Initialize internal datastructure */
    if (!DsListInit(&dsAPI.registry, 10)) {
-      Log(LGPFX "registry initialization failed\n");
+      VMCILOG((LGPFX"registry initialization failed.\n"));
       return FALSE;
    }
    
    
    /* Setup server handle */
-   if (VMCIDatagram_CreateHndPriv(VMCI_DS_RESOURCE_ID, VMCI_FLAG_WELLKNOWN_DG_HND,
-				  VMCI_PRIVILEGE_FLAG_TRUSTED,
-				  DsRequestCb, NULL,
-				  &dsAPI.handle) < VMCI_SUCCESS) {
-      Log(LGPFX "make handle failed\n");
+   if (VMCIDatagramCreateHndPriv(VMCI_DS_RESOURCE_ID, VMCI_FLAG_WELLKNOWN_DG_HND,
+                                 VMCI_PRIVILEGE_FLAG_TRUSTED,
+                                 DsRequestCb, NULL,
+                                 &dsAPI.handle) < VMCI_SUCCESS) {
+      VMCILOG((LGPFX"make handle failed.\n"));
       return FALSE;
    }
    
    if (!VMCI_HANDLE_EQUAL(dsAPI.handle, VMCI_DS_HANDLE)) {
-      Log(LGPFX "handle inconsistency\n");
+      VMCILOG((LGPFX"handle inconsistency.\n"));
       return FALSE;
    } 
    
@@ -534,8 +529,8 @@ VMCIDs_Init(void)
     */
    dsAPI.groupHandle = VMCIGroup_Create();
    if (VMCI_HANDLE_EQUAL(dsAPI.groupHandle, VMCI_INVALID_HANDLE)) {
-      Log(LGPFX"Failed creating Datagram API group.\n");
-      VMCIDatagram_DestroyHnd(dsAPI.handle);
+      VMCILOG((LGPFX"Failed creating Datagram API group.\n"));
+      VMCIDatagramDestroyHndInt(dsAPI.handle);
       return FALSE;
    }
 
@@ -545,14 +540,14 @@ VMCIDs_Init(void)
                                              1, priv, 0, NULL);
    
    if (result != VMCI_SUCCESS) {
-      Log(LGPFX "Failed to setup privileges: %d\n", result);
+      VMCILOG((LGPFX"Failed to setup privileges: %d.\n", result));
       VMCIGroup_Destroy(dsAPI.groupHandle);
-      VMCIDatagram_DestroyHnd(dsAPI.handle);
+      VMCIDatagramDestroyHndInt(dsAPI.handle);
       return FALSE;
    }
    VMCI_InitLock(&lock,
                  "VMCIDsLock",
-                 VMCI_LOCK_RANK_HIGH);
+                 VMCI_LOCK_RANK_MIDDLE);
    
    dsAPI.isInitialized = TRUE;
    
@@ -587,7 +582,7 @@ VMCIDs_Exit(void)
                                           dsAPI.groupHandle);
    
    VMCIGroup_Destroy(dsAPI.groupHandle);
-   VMCIDatagram_DestroyHnd(dsAPI.handle);
+   VMCIDatagramDestroyHndInt(dsAPI.handle);
 
    DsListDestroy(dsAPI.registry);
 
@@ -991,15 +986,16 @@ DsListRemoveResource(DsList *list,           // IN:
 
    registrationCount = resource->registrationCount;
    if (!registrationCount) {
-      Log(LGPFX "%s called with registrationCount = 0.\n", __FUNCTION__);
+      VMCILOG((LGPFX"%s called with registrationCount = 0.\n",
+               __FUNCTION__));
    }
    
    for (i = 0; i < list->size;) {
       if (VMCI_HANDLE_EQUAL(list->elements[i].handle, handle)) {
          int removeElement = DsListRemoveElement(list, i);
          if (removeElement != VMCI_SUCCESS) {
-            Log(LGPFX "Error: %s returned %d.\n",
-                __FUNCTION__, removeElement);
+            VMCILOG((LGPFX"Error: %s returned %d.\n",
+                     __FUNCTION__, removeElement));
             break;
          }
          count++;
@@ -1010,8 +1006,9 @@ DsListRemoveResource(DsList *list,           // IN:
       }
    }
    if (count != registrationCount) {
-      Log(LGPFX "Error: %s: no. of removed registrations "
-          "= %d, should be %d.\n", __FUNCTION__, count, registrationCount);
+      VMCILOG((LGPFX"Error: %s: no. of removed registrations "
+               "= %d, should be %d.\n", __FUNCTION__, count,
+               registrationCount));
    }
    return count;
 }
