@@ -40,7 +40,7 @@
 typedef struct VMCIProcess VMCIProcess;
 
 int VMCIProcess_Init(void);
-int VMCIProcess_Create(VMCIProcess **process, int eventHnd);
+int VMCIProcess_Create(VMCIProcess **process);
 void VMCIProcess_Destroy(VMCIProcess *process);
 VMCIProcess *VMCIProcess_Get(VMCIId processID);
 #endif // _VMCI_PROCESS_H_
