@@ -36,11 +36,6 @@
 #include "vmci_kernel_if.h"
 
 #include "vm_assert.h"
-#ifdef VMKERNEL
-#define LOGLEVEL_MODULE_LEN 0
-#define LOGLEVEL_MODULE VMCIVMK
-#include "log.h"
-#endif // VMKERNEL 
 #include "vmci_defs.h"
 #include "vmci_handle_array.h"
 #include "vmci_infrastructure.h"
@@ -200,8 +195,8 @@ ResourceAddClient(VMCIResource *resource,
    client = VMCI_AllocKernelMem(sizeof *client, 
                                 VMCI_MEMORY_NONPAGED | VMCI_MEMORY_ATOMIC);
    if (client == NULL) {
-      Log(LGPFX"Failed to create new client for resource %p.\n",
-	 resource);
+      VMCILOG((LGPFX"Failed to create new client for resource %p.\n",
+               resource));
       return VMCI_ERROR_NO_MEM;
    }
    client->handle = clientHandle;
@@ -303,7 +298,7 @@ VMCIResource_Init(void)
 {
    resourceTable = VMCIHashTable_Create(128);
    if (resourceTable == NULL) {
-      Log(LGPFX"Failed creating a resource hash table for VMCI.\n");
+      VMCILOG((LGPFX"Failed creating a resource hash table for VMCI.\n"));
       return VMCI_ERROR_NO_MEM;
    }
 
@@ -401,10 +396,10 @@ VMCIResource_Add(VMCIResource *resource,                // IN
    if (VMCI_HANDLE_EQUAL(resourceHandle, VMCI_INVALID_HANDLE) || 
        VMCI_HANDLE_EQUAL(ownerHandle, VMCI_INVALID_HANDLE) ||
        numValidPrivs < 1) {
-      Log(LGPFX"Invalid arguments resource 0x%x:0x%x, owner 0x%x:0x%x, "
-	  "num valid privs %d.\n",
-	  resourceHandle.context, resourceHandle.resource,
-	  ownerHandle.context, ownerHandle.resource, numValidPrivs);
+      VMCILOG((LGPFX"Invalid arguments resource 0x%x:0x%x, owner 0x%x:0x%x, "
+               "num valid privs %d.\n",
+               resourceHandle.context, resourceHandle.resource,
+               ownerHandle.context, ownerHandle.resource, numValidPrivs));
       return VMCI_ERROR_INVALID_ARGS;
    }
 
@@ -430,13 +425,13 @@ VMCIResource_Add(VMCIResource *resource,                // IN
 
    VMCI_InitLock(&resource->clientsLock,
                  "VMCIResourceClientsLock",
-                 VMCI_LOCK_RANK_MIDDLE);
+                 VMCI_LOCK_RANK_MIDDLE_LOW);
    resource->clients = NULL;
 
    /* Add owner as client with the ownerPrivs privileges. */
    result = ResourceAddClient(resource, ownerHandle, 2, ownerPrivs, 0, NULL);
    if (result != VMCI_SUCCESS) {
-      Log(LGPFX"Failed to create owner client.\n");
+      VMCILOG((LGPFX"Failed to create owner client.\n"));
       VMCI_CleanupLock(&resource->clientsLock);
       return result;
    }
@@ -444,7 +439,7 @@ VMCIResource_Add(VMCIResource *resource,                // IN
    /* Add resource to hashtable. */
    result = VMCIHashTable_AddEntry(resourceTable, &resource->hashEntry);
    if (result != VMCI_SUCCESS) {
-      Log(LGPFX"Failed to add entry to hash table.\n");
+      VMCILOG((LGPFX"Failed to add entry to hash table.\n"));
       VMCI_CleanupLock(&resource->clientsLock);
       return result;
    }
@@ -741,38 +736,38 @@ VMCIResource_AddClientPrivileges(VMCIHandle resourceHandle,
    if (VMCI_HANDLE_EQUAL(resourceHandle, VMCI_INVALID_HANDLE) || 
        VMCI_HANDLE_EQUAL(clientHandle, VMCI_INVALID_HANDLE) ||
        numAllowPrivs + numDenyPrivs < 1) {
-      Log(LGPFX"AddClientPrivs: Invalid args.\n");
+      VMCILOG((LGPFX"AddClientPrivs: Invalid args.\n"));
       return VMCI_ERROR_INVALID_ARGS;
    }
 
 #ifdef VMX86_DEBUG
-   VMCI_LOG((LGPFX"AddClientPrivs: Adding allow privs:\n"));
+   VMCI_DEBUG_LOG((LGPFX"AddClientPrivs: Adding allow privs:\n"));
    for (i = 0; i < numAllowPrivs; i++) {
-      VMCI_LOG((LGPFX"AddClientPrivs: %d. 0x%x.\n", i, allowPrivs[i]));
+      VMCI_DEBUG_LOG((LGPFX"AddClientPrivs: %d. 0x%x.\n", i, allowPrivs[i]));
    }
-   VMCI_LOG((LGPFX"AddClientPrivs: Adding deny privs:\n"));
+   VMCI_DEBUG_LOG((LGPFX"AddClientPrivs: Adding deny privs:\n"));
    for (i = 0; i < numDenyPrivs; i++) {
-      VMCI_LOG((LGPFX"AddClientPrivs: %d. 0x%x.\n", i, denyPrivs[i]));
+      VMCI_DEBUG_LOG((LGPFX"AddClientPrivs: %d. 0x%x.\n", i, denyPrivs[i]));
    }
-   VMCI_LOG((LGPFX"AddClientPrivs: to client 0x%"FMT64"x for resource "
+   VMCI_DEBUG_LOG((LGPFX"AddClientPrivs: to client 0x%"FMT64"x for resource "
              "0x%"FMT64"x.\n", clientHandle, resourceHandle));
 #endif // VMX86_DEBUG
 
    resource = VMCIResource_Get(resourceHandle, VMCI_RESOURCE_TYPE_ANY);
    if (resource == NULL) {
-      Log(LGPFX"AddClientPrivs: No resource.\n");
+      VMCILOG((LGPFX"AddClientPrivs: No resource.\n"));
       return VMCI_ERROR_INVALID_ARGS;
    }
 
    /* Validate privileges up front. */
    result = ResourceValidatePrivileges(resource, numAllowPrivs, allowPrivs);
    if (result != VMCI_SUCCESS) {
-      Log(LGPFX"AddClientPrivs: Invalid allow privs.\n");
+      VMCILOG((LGPFX"AddClientPrivs: Invalid allow privs.\n"));
       goto done;
    }
    result = ResourceValidatePrivileges(resource, numDenyPrivs, denyPrivs);
    if (result != VMCI_SUCCESS) {
-      Log(LGPFX"AddClientPrivs: Invalid deny privs.\n");
+      VMCILOG((LGPFX"AddClientPrivs: Invalid deny privs.\n"));
       goto done;
    }
 
@@ -838,22 +833,22 @@ VMCIResource_RemoveClientPrivileges(VMCIHandle resourceHandle,
    if (VMCI_HANDLE_EQUAL(resourceHandle, VMCI_INVALID_HANDLE) || 
        VMCI_HANDLE_EQUAL(clientHandle, VMCI_INVALID_HANDLE) ||
        numPrivs < 1) {
-      Log(LGPFX"RemoveClientPrivs: Invalid args.\n");
+      VMCILOG((LGPFX"RemoveClientPrivs: Invalid args.\n"));
       return VMCI_ERROR_INVALID_ARGS;
    }
 
 #ifdef VMX86_DEBUG
-   VMCI_LOG((LGPFX"RemoveClientPrivs: Removing privs:\n"));
+   VMCI_DEBUG_LOG((LGPFX"RemoveClientPrivs: Removing privs:\n"));
    for (i = 0; i < numPrivs; i++) {
-      VMCI_LOG((LGPFX"RemoveClientPrivs: %d. 0x%x.\n", i, privs[i]));
+      VMCI_DEBUG_LOG((LGPFX"RemoveClientPrivs: %d. 0x%x.\n", i, privs[i]));
    }
-   VMCI_LOG((LGPFX"RemoveClientPrivs: from client 0x%"FMT64"x for resource "
-             "0x%"FMT64"x.\n", clientHandle, resourceHandle));
+   VMCI_DEBUG_LOG((LGPFX"RemoveClientPrivs: from client 0x%"FMT64"x for "
+                   "resource 0x%"FMT64"x.\n", clientHandle, resourceHandle));
 #endif // VMX86_DEBUG
 
    resource = VMCIResource_Get(resourceHandle, VMCI_RESOURCE_TYPE_ANY);
    if (resource == NULL) {
-      Log(LGPFX"RemoveClientPrivs: Failed getting resource.\n");
+      VMCILOG((LGPFX"RemoveClientPrivs: Failed getting resource.\n"));
       result = VMCI_ERROR_INVALID_ARGS;
       goto done;
    }
@@ -861,7 +856,7 @@ VMCIResource_RemoveClientPrivileges(VMCIHandle resourceHandle,
    /* Validate privileges up front to avoid partial changes of privileges. */
    result = ResourceValidatePrivileges(resource, numPrivs, privs);
    if (result != VMCI_SUCCESS) {
-      Log(LGPFX"RemoveClientPrivs: Invalid privs.\n");
+      VMCILOG((LGPFX"RemoveClientPrivs: Invalid privs.\n"));
       goto done;
    }
 
@@ -869,7 +864,7 @@ VMCIResource_RemoveClientPrivileges(VMCIHandle resourceHandle,
    client = ResourceGetClient(resource, clientHandle);
    if (client == NULL) {
       VMCI_ReleaseLock(&resource->clientsLock, flags);
-      Log(LGPFX"RemoveClientPrivs: No client.\n");
+      VMCILOG((LGPFX"RemoveClientPrivs: No client.\n"));
       result = VMCI_ERROR_INVALID_ARGS;
       goto done;
    }
@@ -892,8 +887,8 @@ VMCIResource_RemoveClientPrivileges(VMCIHandle resourceHandle,
        * This client no longer has any privileges set for resource. We remove it 
        * which also decrements the reference count.
        */
-      VMCI_LOG((LGPFX"RemoveClientPrivs: Removing client 0x%"FMT64"x.\n",
-                clientHandle));
+      VMCI_DEBUG_LOG((LGPFX"RemoveClientPrivs: Removing client 0x%"FMT64"x.\n",
+                      clientHandle));
       ResourceRemoveClient(resource, client);
    }
    ResourceReleaseClient(resource, client);
@@ -933,7 +928,7 @@ VMCIResource_RemoveAllClientPrivileges(VMCIHandle resourceHandle,
 
    if (VMCI_HANDLE_EQUAL(resourceHandle, VMCI_INVALID_HANDLE) || 
        VMCI_HANDLE_EQUAL(clientHandle, VMCI_INVALID_HANDLE)) {
-      Log(LGPFX"RemoveAllClientPrivs: Invalid args.\n");
+      VMCILOG((LGPFX"RemoveAllClientPrivs: Invalid args.\n"));
       return VMCI_ERROR_INVALID_ARGS;
    }
 
@@ -988,7 +983,7 @@ VMCIResource_CheckClientPrivilege(VMCIHandle resourceHandle,       // IN:
    int result = VMCI_ERROR_INVALID_PRIV;
 
    if (VMCI_HANDLE_EQUAL(resourceHandle, VMCI_INVALID_HANDLE)) {
-      Log(LGPFX"CheckClientPriv: Invalid args.\n");
+      VMCILOG((LGPFX"CheckClientPriv: Invalid args.\n"));
       return VMCI_ERROR_INVALID_ARGS;
    }
 
@@ -1038,7 +1033,7 @@ VMCIResource_CheckClientPrivilegePtr(VMCIResource *resource,          // IN:
    if (resource == NULL ||
        VMCI_HANDLE_EQUAL(clientHandle, VMCI_INVALID_HANDLE) || 
        priv >= VMCI_NUM_PRIVILEGES) {
-      Log(LGPFX"CheckClientPrivPtr: Invalid args.\n");
+      VMCILOG((LGPFX"CheckClientPrivPtr: Invalid args.\n"));
       return VMCI_ERROR_INVALID_ARGS;
    }
 
@@ -1098,11 +1093,11 @@ VMCIResource_CheckClientPrivilegePtr(VMCIResource *resource,          // IN:
 	     * higher priority was checked first.
 	     */
 	    if (client->privilege[priv] != VMCI_PRIV_NOT_SET) {
-	       VMCI_LOG((LGPFX"Client 0x%"FMT64"x is a member of group "
-                         "0x%"FMT64"x which has priv 0x%x set to %d for "
-                         "resource 0x%"FMT64"x.\n", 
-                         clientHandle, groupHandle, priv, 
-                         client->privilege[priv], resourceHandle));
+	       VMCI_DEBUG_LOG((LGPFX"Client 0x%"FMT64"x is a member of group "
+                               "0x%"FMT64"x which has priv 0x%x set to %d for "
+                               "resource 0x%"FMT64"x.\n", 
+                               clientHandle, groupHandle, priv, 
+                               client->privilege[priv], resourceHandle));
 	       break;
 	    }
 	    ResourceReleaseClient(resource, client);
@@ -1121,9 +1116,9 @@ VMCIResource_CheckClientPrivilegePtr(VMCIResource *resource,          // IN:
       }
       ResourceReleaseClient(resource, client);
    }
-   VMCI_LOG((LGPFX"Checking if client 0x%"FMT64"x has priv 0x%x for resource "
-             "0x%"FMT64"x, result %d.\n", clientHandle, priv, resourceHandle,
-             result));
+   VMCI_DEBUG_LOG((LGPFX"Checking if client 0x%"FMT64"x has priv 0x%x for "
+                   "resource 0x%"FMT64"x, result %d.\n", clientHandle, priv,
+                   resourceHandle, result));
 
   done:
    VMCI_ReleaseLock(&resource->clientsLock, flags);
