@@ -46,6 +46,7 @@
 #include "cpuid.h"
 #include "memtrack.h"
 #include "hashFunc.h"
+#include "pageUtil.h"
 #if defined(_WIN64)
 #include "x86.h"
 #include "vmmon-asm-x86-64.h"
@@ -96,6 +97,21 @@ static unsigned numLockedPages;
 /* Total virtual machines on this host */
 static unsigned vmCount;
 
+/* Total number of open vmmon file handles. */
+static unsigned fdCount;
+
+
+/*
+ * We implement a list of allocated VM ID's using an array.
+ * The array is initialized with the values 1...MAX_VMS-1, INVALID_VMID.
+ * vmIDsAllocated holds the last VM ID given out and vmIDsUnused
+ * holds the next VM ID to give out.
+ */
+
+#define INVALID_VMID (-1)
+static int vmIDList[MAX_VMS];
+static int vmIDsAllocated;
+static int vmIDsUnused;
 
 /* Max rate requested for fast clock by any virtual machine. */
 static unsigned globalFastClockRate;
@@ -288,7 +304,6 @@ Vmx86HasFreePages(VMDriver *vm,
 
 
 #ifdef VMX86_DEBUG
-
 /*
  *----------------------------------------------------------------------
  *
@@ -330,9 +345,123 @@ Vmx86VMIsRegistered(VMDriver *vm, Bool needsLock)
 
    return found;
 }
-
 #endif
 
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Vmx86_InitIDList --
+ *
+ *       Called when the driver is initialized.
+ *       Set up the list of available VM ID's.
+ *
+ * Results:
+ *       None. Sets up global data.
+ *
+ * Side effects:
+ *       None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+Vmx86_InitIDList(void)
+{
+   int i;
+
+   HostIF_GlobalLock(32);
+
+   for (i = 0; i < MAX_VMS; i++) {
+      vmIDList[i] = i + 1;
+   }
+   vmIDList[MAX_VMS - 1] = INVALID_VMID;
+   vmIDsUnused = 0;
+   vmIDsAllocated = INVALID_VMID;
+
+
+   HostIF_GlobalUnlock(32);
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Vmx86FreeVMID --
+ *
+ *       Return a VM ID to the list of available VM ID's.
+ *
+ * Results:
+ *       None
+ *
+ * Side effects:
+ *       None
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+Vmx86FreeVMID(int vmID) // IN
+{
+   int i;
+
+   ASSERT(HostIF_GlobalLockIsHeld());
+
+   /* Deleting head of the list. */
+   if (vmID == vmIDsAllocated) {
+      int tmp;
+
+      tmp = vmIDList[vmIDsAllocated];
+      vmIDList[vmIDsAllocated] = vmIDsUnused;
+      vmIDsAllocated = tmp;
+      vmIDsUnused = vmID;
+      return;
+   }
+
+   for (i = vmIDsAllocated; vmIDList[i] != INVALID_VMID; i = vmIDList[i]) {
+      if (vmIDList[i] == vmID) {
+         vmIDList[i] = vmIDList[vmID];
+         vmIDList[vmID] = vmIDsUnused;
+         vmIDsUnused = vmID;
+         return;
+      }
+   }
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Vmx86AllocVMID --
+ *
+ *       Grab a VM ID from the list of available VM ID's.
+ *
+ * Results:
+ *       The VM ID, in the range [ 0 ; MAX_VMS ).
+ *
+ * Side effects:
+ *       None
+ *
+ *----------------------------------------------------------------------
+ */
+
+static int
+Vmx86AllocVMID(void)
+{
+   int vmID;
+
+   ASSERT(HostIF_GlobalLockIsHeld());
+
+   vmID = vmIDsUnused;
+   ASSERT(0 <= vmID && vmID < MAX_VMS);
+   vmIDsUnused = vmIDList[vmID];
+   vmIDList[vmID] = vmIDsAllocated;
+   vmIDsAllocated = vmID;
+
+   return vmID;
+}
+
+
 /*
  *----------------------------------------------------------------------
  *
@@ -352,12 +481,17 @@ Vmx86VMIsRegistered(VMDriver *vm, Bool needsLock)
  */
 
 static void
-Vmx86RegisterVMOnList(VMDriver *vm) 
+Vmx86RegisterVMOnList(VMDriver *vm) // IN
 {
+   int vmID;
    VMDriver **vmp;
 
    ASSERT(HostIF_GlobalLockIsHeld());
    vmCount++;
+   vmID = Vmx86AllocVMID();
+   ASSERT(vm->userID == 0);
+   vm->userID = vmID + 1;
+   ASSERT(vm->userID > 0);
    for (vmp = &vmDriverList; *vmp != NULL; vmp = &(*vmp)->nextDriver) {
       if (*vmp == vm) {
          Warning("VM %p already registered on the list of VMs.\n", vm);
@@ -367,6 +501,7 @@ Vmx86RegisterVMOnList(VMDriver *vm)
    *vmp = vm;
 }
 
+
 /*
  *----------------------------------------------------------------------
  *
@@ -403,17 +538,15 @@ Vmx86DeleteVMFromList(VMDriver *vm)
    *vmp = vm->nextDriver;
    vmCount--;
 
+   Vmx86FreeVMID(vm->userID - 1);
    numLockedPages -= vm->memInfo.locked;
 
    /*
     * If no VM is running, reset the configured locked-page limit so
-    * that the next VM to power on sets it appropriately.  Similarly,
-    * reset the pseudo TSC state so that the next VM to initialize is
-    * free to initialize the system wide PTSC however it wants.
+    * that the next VM to power on sets it appropriately.
     */
    if (vmCount == 0) {
       lockedPageLimit.configured = 0;
-      pseudoTSC.initialized = FALSE;
    }
 }
 
@@ -446,9 +579,6 @@ Vmx86FreeAllVMResources(VMDriver *vm)
 
       HostIF_FreeAllResources(vm);
 
-      if (vm->vmhost != NULL) {
-         HostIF_FreeKernelMem(vm->vmhost);
-      }
       HostIF_FreeKernelMem(vm);
    }
 }
@@ -584,19 +714,9 @@ Vmx86_CreateVM(void)
    }
    memset(vm, 0, sizeof *vm);
 
-   ASSERT(sizeof (uintptr_t) == sizeof vm);
-   /*
-    * id needs to be non-zero positive integer, unique for each
-    * VM.  So we just take vm's address, divide it by structure
-    * size (so we get different ids even for back to back
-    * structures), mask top two bits, and add 1.  So we get some
-    * more or less unique value in a range between 1 and 0x40000000.
-    * And typecast everything to unsigned to shut up Windows 64bit
-    * compiler.
-    */
-   vm->id = (unsigned)((((uintptr_t)vm / sizeof *vm) & 0x3FFFFFFF) + 1);
+   vm->userID = 0;
    vm->memInfo.admitted = FALSE;
-   vm->fastSuspResFlag      =  0;
+   vm->fastSuspResFlag = 0;
    for (v = 0; v < MAX_INITBLOCK_CPUS; v++) {
       vm->currentHostCpu[v] = INVALID_HOST_CPU;
    }
@@ -625,9 +745,9 @@ Vmx86_CreateVM(void)
    return vm;
 
 cleanup:
-   /* 
+   /*
     * The VM is not on a list, "vmCount" has not been incremented,
-    * "vm->cowID" is INVALID_COWID, and either the VM's mutex hasn't
+    * "vm->cowID" is INVALID_VMID, and either the VM's mutex hasn't
     * been initialized or we've only taken the global lock and checked
     * a counter since, so we know that the VM has not yet locked any
     * pages.
@@ -667,6 +787,78 @@ Vmx86_ReleaseVM(VMDriver *vm)
 }
 
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * Vmx86_Open --
+ *
+ *      Called on open of the fd.
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *	Bumps fdCount.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+Vmx86_Open(void)
+{
+   HostIF_GlobalLock(123);
+   ASSERT(fdCount < MAX_INT32);
+   if (fdCount < MAX_INT32) {
+      fdCount++;
+   }
+   HostIF_GlobalUnlock(123);
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Vmx86_Close --
+ *
+ *      Called on close of the fd.
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      Decrements fdCount
+ *	May de-initialize ptsc.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+Vmx86_Close(void)
+{
+   HostIF_GlobalLock(124);
+
+   /* 
+    * If fdCount hits MAX_INT32 saturate the counter and leave it at
+    * MAX_INT32. 
+    */
+   ASSERT(fdCount > 0);
+   if (fdCount < MAX_INT32) {
+      fdCount--;
+   }
+
+   /* 
+    * If no VMs are running and there are no open file handles, reset the
+    * pseudo TSC state so that the next VM to initialize is free to
+    * initialize the system wide PTSC however it wants.  See PR 403505.
+    */
+   if (fdCount == 0) {
+      ASSERT(vmCount == 0);
+      pseudoTSC.initialized = FALSE;
+   }
+   HostIF_GlobalUnlock(124);
+}
+
+
 /*
  *------------------------------------------------------------------------------
  *
@@ -850,7 +1042,6 @@ Vmx86_GetNUMAMemStats(VMDriver *curVM,            //IN
    int i;
    NUMA_NodeInfo *curNode;
    int apicId;
-   DEBUG_ONLY(int initialApicId;)
    
    ASSERT(curVM);
    
@@ -878,20 +1069,6 @@ Vmx86_GetNUMAMemStats(VMDriver *curVM,            //IN
       Warning("Processor not part of this node, structures are wrong \n");
    }
    
-   /*
-    * The inital APIC is read from the CPUID instruction, whereas the current
-    * is read from a physical address, done by the HostIF_APIC_ID().
-    * Confirm that they are same.
-    */   
-
-   DEBUG_ONLY({
-      initialApicId = HostIF_GetInitialApicId();
-      if (apicId != initialApicId) {
-         Warning("Vmx86_GetNUMAStats: The current APIC Id of the processor is "
-                 "different from the Inital value\n");
-      }
-   })
-
    vm = curVM;
    HostIF_VMLock(vm, 17);
 
@@ -939,7 +1116,7 @@ Vmx86_InitVM(VMDriver *vm,          // IN
       Warning("Bad magic number for init block 0x%x\n", initParams->magicNumber);
       return 1;
    }
-   if (initParams->numVCPUs >= MAX_INITBLOCK_CPUS) {
+   if (initParams->numVCPUs > MAX_INITBLOCK_CPUS) {
       Warning("Too many VCPUs for init block %d\n", initParams->numVCPUs);
       return 1;
    }
@@ -1581,7 +1758,7 @@ Vmx86_SetDynamicLockedPagesLimit(unsigned limit)
 
 MPN
 Vmx86_LockPage(VMDriver *vm,		     // IN: VMDriver
-	       void* addr,		     // IN: VA of the page to lock
+	       VA64 uAddr,		     // IN: VA of the page to lock
 	       Bool allowMultipleMPNsPerVA)  // IN: allow locking many pages with the same VA
 {
    MPN mpn;
@@ -1592,7 +1769,7 @@ Vmx86_LockPage(VMDriver *vm,		     // IN: VMDriver
    }
 
    HostIF_VMLock(vm, 4);
-   mpn = HostIF_LockPage(vm, addr, allowMultipleMPNsPerVA);
+   mpn = HostIF_LockPage(vm, uAddr, allowMultipleMPNsPerVA);
    HostIF_VMUnlock(vm, 4);
 
    if (!PAGE_LOCK_SUCCESS(mpn)) {
@@ -1619,13 +1796,15 @@ Vmx86_LockPage(VMDriver *vm,		     // IN: VMDriver
  *
  *----------------------------------------------------------------------
  */
+
 int
-Vmx86_UnlockPage(VMDriver *vm, void *addr)
+Vmx86_UnlockPage(VMDriver *vm, // IN
+                 VA64 uAddr)   // IN
 {
    int retval;
    
    HostIF_VMLock(vm, 5);
-   retval = HostIF_UnlockPage(vm, addr);
+   retval = HostIF_UnlockPage(vm, uAddr);
    HostIF_VMUnlock(vm, 5);
 
    if (PAGE_LOCK_SUCCESS(retval)) {
@@ -1654,12 +1833,12 @@ Vmx86_UnlockPage(VMDriver *vm, void *addr)
 int
 Vmx86_UnlockPageByMPN(VMDriver *vm, // IN: VMDriver
 		      MPN mpn,	    // IN: the page to unlock
-		      void* addr)   // IN: optional valid VA for this MPN
+		      VA64 uAddr)   // IN: optional valid VA for this MPN
 {
    int retval;
 
    HostIF_VMLock(vm, 6);
-   retval = HostIF_UnlockPageByMPN(vm, mpn, addr);
+   retval = HostIF_UnlockPageByMPN(vm, mpn, uAddr);
    HostIF_VMUnlock(vm, 6);
 
    if (PAGE_LOCK_SUCCESS(retval)) {
@@ -1681,8 +1860,8 @@ Vmx86_UnlockPageByMPN(VMDriver *vm, // IN: VMDriver
  *
  * Results:
  *      Non-negative value on partial/full completion: actual number of
- *      allocated MPNs. MPNs of the allocated pages are copied to 
- *      the caller's buffer.
+ *      allocated MPNs. MPN32s of the allocated pages are copied to the
+ *      caller's buffer at 'addr'.
  *
  *	Negative system specific error code on error (NTSTATUS on Windows,
  *      etc.)
@@ -1695,8 +1874,8 @@ Vmx86_UnlockPageByMPN(VMDriver *vm, // IN: VMDriver
 
 int
 Vmx86_AllocLockedPages(VMDriver *vm,	     // IN: VMDriver
-		       MPN32 *mpns,	     // OUT: User VA of an array for
-                                             //      allocated MPNs.
+		       VA64 addr,	     // OUT: VA of an array for
+                                             //      allocated MPN32s.
 		       unsigned numPages,    // IN: number of pages to allocate
 		       Bool kernelMPNBuffer) // IN: is the MPN buffer in kernel or user address space?
 {
@@ -1708,7 +1887,7 @@ Vmx86_AllocLockedPages(VMDriver *vm,	     // IN: VMDriver
    }
    
    HostIF_VMLock(vm, 7);
-   allocatedPages = HostIF_AllocLockedPages(vm, mpns, numPages, kernelMPNBuffer); 
+   allocatedPages = HostIF_AllocLockedPages(vm, addr, numPages, kernelMPNBuffer);
    HostIF_VMUnlock(vm, 7);
 
    if (allocatedPages < 0) {
@@ -1741,14 +1920,14 @@ Vmx86_AllocLockedPages(VMDriver *vm,	     // IN: VMDriver
 
 int
 Vmx86_FreeLockedPages(VMDriver *vm,	    // IN: VM instance pointer
-		      const MPN32* mpns,    // IN: user or kernel array of MPNs to free 
+		      VA64 addr,            // IN: user or kernel array of MPNs to free 
 		      unsigned numPages,    // IN: number of pages to free
 		      Bool kernelMPNBuffer) // IN: is the MPN buffer in kernel or user address space?
 {
    int ret;
 
    HostIF_VMLock(vm, 8);
-   ret = HostIF_FreeLockedPages(vm, mpns, numPages, kernelMPNBuffer);
+   ret = HostIF_FreeLockedPages(vm, addr, numPages, kernelMPNBuffer);
    HostIF_VMUnlock(vm, 8);
 
    if (ret == 0) {
@@ -1807,14 +1986,14 @@ Vmx86_IsAnonPage(VMDriver *vm,       // IN: VM instance pointer
  */
 
 int 
-Vmx86_GetLockedPageList(VMDriver* vm,          // IN: VM instance pointer
-                        MPN32* mpns,           // OUT: user mode buffer for MPNs
+Vmx86_GetLockedPageList(VMDriver *vm,          // IN: VM instance pointer
+                        VA64 uAddr,            // OUT: user mode buffer for MPNs
 		        unsigned int numPages) // IN: size of the buffer in MPNs 
 {
    int ret;
       
    HostIF_VMLock(vm, 9);
-   ret = HostIF_GetLockedPageList(vm, mpns, numPages);
+   ret = HostIF_GetLockedPageList(vm, uAddr, numPages);
    HostIF_VMUnlock(vm, 9);
    
    return ret;
@@ -1856,32 +2035,6 @@ Vmx86COWStats(VMMemCOWInfo *info)
 }
 
 
-/*
- *----------------------------------------------------------------------
- *
- * Vmx86COWStatsVM --
- *
- *      Return zeroes...
- *
- * Results:
- *      None.
- *
- * Side effects:
- *      None.
- *
- *----------------------------------------------------------------------
- */
-
-static void
-Vmx86COWStatsVM(VMDriver *vm)
-{
-   VMMemMgmtInfo *memInfo = &vm->memInfo;
-
-   memInfo->breaksAvg = 0;
-   memInfo->sharedPctAvg = 0;
-}
-
-
 /*
  *----------------------------------------------------------------------
  *
@@ -1900,7 +2053,7 @@ Vmx86COWStatsVM(VMDriver *vm)
 
 Bool
 Vmx86_GetMemInfo(VMDriver *curVM,
-                 int32 flags,
+                 Bool curVMOnly,
                  VMMemInfoArgs *outArgs,
                  int outArgsLength)
 {
@@ -1908,8 +2061,6 @@ Vmx86_GetMemInfo(VMDriver *curVM,
    int i;
    int outSize;
    int wantedVMs;
-   Bool curVMOnly = (flags & VMMEM_CUR_VM_ONLY) != 0;
-   Bool getPeriodicStats = (flags & VMMEM_GET_PERIODIC_STATS) != 0;
 
    HostIF_GlobalLock(7);
 
@@ -1935,9 +2086,6 @@ Vmx86_GetMemInfo(VMDriver *curVM,
    Vmx86COWStats(&outArgs->cowInfo);
 
    if (curVM != NULL) {
-      if (getPeriodicStats) {
-         Vmx86COWStatsVM(curVM);
-      }
       if (wantedVMs == 1) {
          outArgs->memInfo[0] = curVM->memInfo;
          outArgs->callerIndex = 0;
@@ -1969,10 +2117,9 @@ Vmx86_GetMemInfo(VMDriver *curVM,
  *
  * Vmx86_GetMemInfoCopy --
  *
- *      Return the information about all VMs by copying the data out to user
- *      memory.
+ *      Return the information about all VMs.
  *
- *      On input, outArgs->numVMs indicates how much space has been allocated
+ *      On input, buf->numVMs indicates how much space has been allocated
  *      for the information. On output, it indicates how much space has been
  *      filled.
  *
@@ -1987,29 +2134,20 @@ Vmx86_GetMemInfo(VMDriver *curVM,
  */
 
 Bool
-Vmx86_GetMemInfoCopy(VMDriver *curVM,        // IN
-                     VMMemInfoArgs *outArgs) // IN/OUT: User VA
+Vmx86_GetMemInfoCopy(VMDriver *curVM,    // IN
+                     VMMemInfoArgs *buf) // IN/OUT
 {
-   VMMemInfoArgs *buf;
    VMDriver *vm;
-   Bool ret = FALSE;
+   Bool ret;
 
    ASSERT(curVM);
 
-   buf = HostIF_AllocKernelMem(VM_GET_MEM_INFO_SIZE(MAX_VMS), TRUE);
-   if (!buf) {
-      goto err0;
-   }
-   if (HostIF_CopyFromUser(buf, outArgs, VM_GET_MEM_INFO_SIZE(1))) {
-      goto err1;
-   }
-
    HostIF_GlobalLock(8);
 
    /* Now that we have the lock, we can read vmCount --hpreg */
    if (buf->numVMs < vmCount) {
-      HostIF_GlobalUnlock(8);
-      goto err1;
+      ret = FALSE;
+      goto exit;
    }
 
    buf->numLockedPages = numLockedPages;
@@ -2018,9 +2156,6 @@ Vmx86_GetMemInfoCopy(VMDriver *curVM,        // IN
    buf->globalMinAllocation = Vmx86CalculateGlobalMinAllocation(minVmMemPct);
    buf->minVmMemPct = minVmMemPct;
    Vmx86COWStats(&buf->cowInfo);
-   if (buf->getPeriodicStats) {
-      Vmx86COWStatsVM(curVM);
-   }
 
    for (vm = vmDriverList, buf->numVMs = 0;
         vm;
@@ -2035,15 +2170,11 @@ Vmx86_GetMemInfoCopy(VMDriver *curVM,        // IN
    }
    ASSERT(buf->numVMs == vmCount);
 
+   ret = TRUE;
+
+exit:
    HostIF_GlobalUnlock(8);
 
-   if (!HostIF_CopyToUser(outArgs, buf, VM_GET_MEM_INFO_SIZE(buf->numVMs))) {
-      ret = TRUE;
-   }
-
-err1:
-   HostIF_FreeKernelMem(buf);
-err0:
    return ret;
 }
 
@@ -2169,8 +2300,9 @@ Vmx86_Admit(VMDriver *curVM,     // IN
       if (mpns == NULL) {
          goto undoAdmission;
       }
-      while(allocatedPages < nonpaged) {
-         pages = Vmx86_AllocLockedPages(curVM, mpns + allocatedPages, 
+      while (allocatedPages < nonpaged) {
+         pages = Vmx86_AllocLockedPages(curVM,
+                                        PtrToVA64(mpns + allocatedPages),
 	                                MIN(ALLOCATE_CHUNK_SIZE, nonpaged - allocatedPages),
 	                                TRUE);
          if (pages <= 0) {
@@ -2185,7 +2317,7 @@ Vmx86_Admit(VMDriver *curVM,     // IN
        */
 
       for (pages = 0; pages < allocatedPages; pages += ALLOCATE_CHUNK_SIZE) {
-         Vmx86_FreeLockedPages(curVM, mpns + pages, 
+         Vmx86_FreeLockedPages(curVM, PtrToVA64(mpns + pages),
                                MIN(ALLOCATE_CHUNK_SIZE, allocatedPages - pages), TRUE);
       }
       HostIF_FreeKernelMem(mpns);
@@ -2214,41 +2346,45 @@ undoAdmission:
 
 
 Bool
-Vmx86_Readmit(VMDriver *curVM, int32 pageDelta)
+Vmx86_Readmit(VMDriver *curVM, OvhdMem_Deltas *delta)
 {
    unsigned globalMinAllocation, newMinAllocation;
    Bool retval = FALSE;
-   int32 nonpaged;
+   int paged;
+   int nonpaged;
 
    HostIF_GlobalLock(31);
    globalMinAllocation = Vmx86CalculateGlobalMinAllocation(minVmMemPct);
    HostIF_VMLock(curVM, 31);
-   nonpaged = curVM->memInfo.nonpaged + pageDelta;
-   if (nonpaged >= 0) {
+   paged = curVM->memInfo.paged + delta->paged;
+   nonpaged = curVM->memInfo.nonpaged + delta->nonpaged;
+   if (nonpaged >= 0 && paged >= (int)curVM->memInfo.mainMemSize) {
       globalMinAllocation -= Vmx86MinAllocation(curVM, minVmMemPct);
-      newMinAllocation = Vmx86MinAllocationFunc(curVM->memInfo.paged,
-                                                nonpaged,
+      newMinAllocation = Vmx86MinAllocationFunc(paged, nonpaged,
                                                 curVM->memInfo.mainMemSize,
                                                 minVmMemPct);
       if (globalMinAllocation + newMinAllocation <= Vmx86LockedPageLimit(curVM) ||
-         pageDelta <= 0) {
-         retval = Vmx86SetMemoryUsage(curVM, curVM->memInfo.paged, nonpaged, minVmMemPct);
+          (delta->paged <= 0 && delta->nonpaged <= 0)) {
+         retval = Vmx86SetMemoryUsage(curVM, paged, nonpaged, minVmMemPct);
       }
    }
    HostIF_VMUnlock(curVM, 31);
    HostIF_GlobalUnlock(31);
    return retval;
 }
-                    
+
+
 /*
  *----------------------------------------------------------------------
  *
- * Vmx86_SetMemUsage --
+ * Vmx86_UpdateMemInfo --
  *
- *      Set the memory usage by this vm based on its memSample data.
+ *      Updates information about this VM with the new data supplied in
+ *      a patch.
  *
  * Results:
- *      Updates memory allocations.
+ *      Sets the memory usage by this vm based on its memSample data and
+ *      updates page sharing stats.
  *
  * Side effects:
  *      None
@@ -2257,14 +2393,16 @@ Vmx86_Readmit(VMDriver *curVM, int32 pageDelta)
  */
 
 void
-Vmx86_SetMemUsage(VMDriver *curVM, 
-		  unsigned usedPct)
+Vmx86_UpdateMemInfo(VMDriver *curVM, 
+		    const VMMemMgmtInfoPatch *patch)
 {
-   if (usedPct <= 100) {
-      HostIF_VMLock(curVM, 13);
-      curVM->memInfo.usedPct = AsPercent(usedPct);
-      HostIF_VMUnlock(curVM, 13);
+   HostIF_VMLock(curVM, 13);
+   if (patch->usedPct <= 100) {
+      curVM->memInfo.usedPct = AsPercent(patch->usedPct);
    }
+   curVM->memInfo.sharedPctAvg = patch->sharedPctAvg;
+   curVM->memInfo.breaksAvg = patch->breaksAvg;
+   HostIF_VMUnlock(curVM, 13);
 }
 
 
@@ -2285,14 +2423,9 @@ Vmx86_SetMemUsage(VMDriver *curVM,
  */
 
 Bool
-Vmx86_PAEEnabled()
+Vmx86_PAEEnabled(void)
 {
-#ifdef VM_X86_64
-   uint64 cr4;
-#else
-   uint32 cr4;
-#endif
-
+   uintptr_t cr4;
    GET_CR4(cr4);
    return (cr4 & CR4_PAE) != 0;
 }
@@ -2303,12 +2436,11 @@ Vmx86_PAEEnabled()
  *
  * Vmx86_VMXEnabled --
  *
- *      Test the VMXE bit as an easy proxy for whether VMX mode 
- *      is enabled. CR4.VMXE is reserved on AMD.
+ *      Test the VMXE bit as an easy proxy for whether VMX operation
+ *      is enabled.
  *
  * Results:
- *      TRUE if CR4.VMXE is set and the CPU is non-AMD (Intel or 
- *      Intel-compatible).
+ *      TRUE if the CPU supports VT and CR4.VMXE is set.
  *
  * Side effects:
  *      None.
@@ -2317,15 +2449,10 @@ Vmx86_PAEEnabled()
  */
 
 Bool
-Vmx86_VMXEnabled()
+Vmx86_VMXEnabled(void)
 {
-#ifdef VM_X86_64
-   uint64 cr4;
-#else
-   uint32 cr4;
-#endif
-
-   if (CPUID_GetVendor() != CPUID_VENDOR_AMD) {
+   if (VT_CapableCPU()) {
+      uintptr_t cr4;
       GET_CR4(cr4);
       return (cr4 & CR4_VMXE) != 0;
    } else {
@@ -2431,9 +2558,7 @@ Vmx86_BrokenCPUHelper(void)
    __SET_MSR(0x8b, 0); // Clear MSR 0x8b
    __GET_CPUID(1, &regs); // Issue CPUID command 1
    msr_val8b = __GET_MSR(0x8b);
-   Log("INTELRDMSR(0x8b) = 0x%08x%08x\n",(uint32)(msr_val8b>>32),(uint32)msr_val8b);
    msr_val17 = __GET_MSR(0x17);
-   Log("INTELRDMSR(0x17) = 0x%08x%08x\n",(uint32)(msr_val17>>32),(uint32)msr_val17);
 
    if (!((family == 6) && (model == 8) && (step == 1))) { 
       /* Only have problems with Coppermine CPU cores. */
@@ -2468,19 +2593,15 @@ Vmx86_BrokenCPUHelper(void)
  *----------------------------------------------------------------------
  */
 
-#ifdef __APPLE__
 Bool
 Vmx86_InCompatMode(void)
 {
-   /*
-    * This routine only runs on Macs at present, which never run in full 64-bit 
-    * mode.  The code below assumes that fact, so we assert it here.
-    */
-   ASSERT_ON_COMPILE(!vm_x86_64);
-
+#if defined __APPLE__ && !vm_x86_64
    return Vmx86_InLongMode();
-}
+#else
+   return FALSE;
 #endif
+}
 
 
 /*
@@ -2501,42 +2622,84 @@ Vmx86_InCompatMode(void)
  *----------------------------------------------------------------------
  */
 
-#ifdef __APPLE__
 Bool
 Vmx86_InLongMode(void)
 {
+#if defined __APPLE__ && !vm_x86_64
    uint64 efer;
 
    efer = __GET_MSR(MSR_EFER);
    return (efer & MSR_EFER_LME) != 0;
-}
+#else
+   return vm_x86_64;
 #endif
+}
 
 
 /*
- *----------------------------------------------------------------------
+ *-----------------------------------------------------------------------------
  *
- * Vmx86LookupVMByID --
+ * Vmx86LookupVMByUserIDLocked --
  *
- *    Finds the vm with the specified ID.
- *    
- *----------------------------------------------------------------------
+ *      Lookup a VM by userID. The caller must hold the global lock.
+ *
+ * Returns:
+ *      On success: Pointer to the driver's VM instance.
+ *      On failure: NULL (not found).
+ *
+ * Side effects:
+ *      None
+ *
+ *-----------------------------------------------------------------------------
  */
 
-VMDriver *
-Vmx86LookupVMByID(int id)
+static VMDriver *
+Vmx86LookupVMByUserIDLocked(int userID) // IN
 {
    VMDriver *vm;
+
    ASSERT(HostIF_GlobalLockIsHeld());
+
    for (vm = vmDriverList; vm != NULL; vm = vm->nextDriver) {
-      if (vm->id == id) {
+      if (vm->userID == userID) {
          return vm;
       }
    }
+
    return NULL;
 }
 
 
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * Vmx86_LookupVMByUserID --
+ *
+ *      Lookup a VM by userID.
+ *
+ * Returns:
+ *      On success: Pointer to the driver's VM instance.
+ *      On failure: NULL (not found).
+ *
+ * Side effects:
+ *      None
+ *
+ *-----------------------------------------------------------------------------
+ */
+
+VMDriver *
+Vmx86_LookupVMByUserID(int userID) // IN
+{
+   VMDriver *vm;
+
+   HostIF_GlobalLock(10);
+   vm = Vmx86LookupVMByUserIDLocked(userID);
+   HostIF_GlobalUnlock(10);
+
+   return vm;
+}
+
+
 /*
  *----------------------------------------------------------------------
  *
@@ -2549,22 +2712,24 @@ Vmx86LookupVMByID(int id)
  *      FALSE if VM was not found.
  *
  * Side effects:
- *      The value we set the flag to is this VM's id.
+ *      The value we set the flag to is this VM's userID.
  *
  *----------------------------------------------------------------------
  */
 
 Bool
-Vmx86_FastSuspResSetOtherFlag(VMDriver *vm, int otherVmId)
+Vmx86_FastSuspResSetOtherFlag(VMDriver *vm,      // IN
+                              int otherVmUserId) // IN
 {
-   VMDriver *otherVM = NULL;
+   VMDriver *otherVM;
+
    HostIF_GlobalLock(35);
-   otherVM = Vmx86LookupVMByID(otherVmId);
-   if (otherVM) {
+   otherVM = Vmx86LookupVMByUserIDLocked(otherVmUserId);
+   if (otherVM != NULL) {
       ASSERT(otherVM->fastSuspResFlag == 0);
-      otherVM->fastSuspResFlag = vm->id;
+      otherVM->fastSuspResFlag = vm->userID;
    } else {
-      Warning("otherVmId (%d) is invalid", otherVmId);
+      Warning("otherVmUserId (%d) is invalid", otherVmUserId);
    }
    HostIF_GlobalUnlock(35);
    return otherVM != NULL;
@@ -2581,7 +2746,7 @@ Vmx86_FastSuspResSetOtherFlag(VMDriver *vm, int otherVmId)
  *      timeout.
  *
  * Returns:
- *      The value of the flag which, if non-zero, should be the id of
+ *      The value of the flag which, if non-zero, should be the userID of
  *      the vm that set it.
  *
  * Side effects:
@@ -2591,7 +2756,8 @@ Vmx86_FastSuspResSetOtherFlag(VMDriver *vm, int otherVmId)
  */
 
 int
-Vmx86_FastSuspResGetMyFlag(VMDriver *vm, Bool blockWait)
+Vmx86_FastSuspResGetMyFlag(VMDriver *vm,   // IN
+                           Bool blockWait) // IN
 {
    int retval = 0;
    int ntries = 1;
@@ -2603,10 +2769,10 @@ Vmx86_FastSuspResGetMyFlag(VMDriver *vm, Bool blockWait)
    }
 
    while (ntries--) {
-      HostIF_GlobalLock(35);
+      HostIF_GlobalLock(6);
       retval = vm->fastSuspResFlag;
       vm->fastSuspResFlag = 0;
-      HostIF_GlobalUnlock(35);
+      HostIF_GlobalUnlock(6);
       if (retval || !ntries) {
          break;
       }
@@ -2760,17 +2926,17 @@ Vmx86_FixHVEnable(Bool force)
    static void (*perCPUFunc)(void*);
 
    /* 
-    * Check CPU vendor and cache "force" at boot. 
+    * Check HV capabilities and cache "force" at boot. 
     * Callers of Vmx86_HVEnabledCPUs can set also "force".
     */
    if (!once) {
       /* Potential race, but idempotent so don't care. */
-      if (CPUID_GetVendor() == CPUID_VENDOR_INTEL) {
+      if (VT_CapableCPU()) {
          perCPUFunc = Vmx86GetVTEnableOnCPU;
-      } else if (CPUID_GetVendor() == CPUID_VENDOR_AMD) {
+      } else if (SVM_CapableCPU()) {
          perCPUFunc = Vmx86GetSVMEnableOnCPU;
       } else {
-         hvCapable = FALSE; // Only Intel / AMD have HV.
+         hvCapable = FALSE;
       }
       if (perCPUFunc) {
          HVEnableData data = { FALSE, FALSE, FALSE, FALSE };
@@ -2803,9 +2969,9 @@ Vmx86_FixHVEnable(Bool force)
       ASSERT(!data.anyNotCapable);
 
       /* VT requires enabled and locked.  SVM only requires enabled. */
-      if (CPUID_GetVendor() == CPUID_VENDOR_INTEL) {
+      if (VT_CapableCPU()) {
          hvEnabled = !(data.anyDisabled || data.anyUnlocked);
-      } else if (CPUID_GetVendor() == CPUID_VENDOR_AMD) {
+      } else if (SVM_CapableCPU()) {
          hvEnabled = !data.anyDisabled;
       }
       Log("HV check: anyNotCapable=%d anyUnlocked=%d anyEnabled=%d anyDisabled=%d\n",
@@ -3025,7 +3191,20 @@ Vmx86GetMSR(void *clientData) // IN/OUT: A Vmx86GetMSRData *
    }
 
    query->logicalCPUs[index].tag = HostIF_GetCurrentPCPU();
-   err = HostIF_SafeRDMSR(query->msrNum, &query->logicalCPUs[index].msrVal);
+
+   /*
+    * We treat BIOS_SIGN_ID (microcode version) specially on Intel,
+    * where the preferred read sequence involves a macro.
+    */
+   if (CPUID_GetVendor() == CPUID_VENDOR_INTEL &&
+       query->msrNum == MSR_BIOS_SIGN_ID) {
+      /* safe to read: MSR_BIOS_SIGN_ID architectural since Pentium Pro */
+      query->logicalCPUs[index].msrVal = INTEL_MICROCODE_VERSION();
+      err = 0;
+   } else {
+      err = HostIF_SafeRDMSR(query->msrNum, &query->logicalCPUs[index].msrVal);
+   }
+
    query->logicalCPUs[index].implemented = (err == 0) ? 1 : 0;
 }
 
