@@ -24,11 +24,7 @@
 #include "compat_module.h"
 #include <linux/version.h>
 #include <linux/sched.h>
-#ifdef KERNEL_2_2
-#   include <linux/slab.h>
-#else
-#   include <linux/malloc.h>
-#endif
+#include <linux/slab.h>
 #include <linux/poll.h>
 
 #include <linux/smp.h>
@@ -486,7 +482,7 @@ init_module(void)
     */
 
    memset(&vnetFileOps, 0, sizeof vnetFileOps);
-   compat_fop_set_owner(&vnetFileOps);
+   vnetFileOps.owner = THIS_MODULE;
    vnetFileOps.read = VNetFileOpRead;
    vnetFileOps.write = VNetFileOpWrite;
    vnetFileOps.poll = VNetFileOpPoll;
@@ -557,33 +553,6 @@ cleanup_module(void)
 }
 
 
-/*
- *----------------------------------------------------------------------
- *
- * VNetIncrModCount --
- *
- *      Increment or decrement the module use count.
- *
- * Results:
- *      None.
- *
- * Side effects:
- *      None.
- *
- *----------------------------------------------------------------------
- */
-
-void
-VNetIncrModCount(int delta) // IN: delta (typically 1 or -1)
-{
-   if (delta > 0) {
-      compat_mod_inc_refcount;
-   } else if (delta < 0) {
-      compat_mod_dec_refcount;
-   }
-}
-
-
 /*
  *----------------------------------------------------------------------
  *
@@ -663,8 +632,6 @@ VNetFileOpOpen(struct inode  *inode, // IN: used to get hub number
 
    LOG(1, (KERN_DEBUG "/dev/vmnet: port on hub %d successfully opened\n", hubNum));
 
-   compat_mod_inc_refcount;
-
    return 0;
 }
 
@@ -706,8 +673,6 @@ VNetFileOpClose(struct inode  *inode, // IN: (unused)
    VNetFree(&port->jack);
    VNetFree(peer);
 
-   compat_mod_dec_refcount;
-
    return 0;
 }
 
@@ -834,9 +799,6 @@ VNetFileOpPoll(struct file *filp, // IN:
  *
  *      SIOCGIFADDR - get ethernet address          - ioarg OUT: 6 bytes
  *      SIOCSIFADDR - set ethernet address          - ioarg IN:  6 bytes
-#ifdef VNET_API_DEPRECATED
- *      SIOCGIFFLAGS - get flags                    - ioarg OUT  4 bytes
-#endif
  *      SIOCSIFFLAGS - set flags                    - ioarg IN:  4 bytes
  *
  *      Private ioctl calls, taken from device-private ioctl space
@@ -897,9 +859,6 @@ VNetFileOpIoctl(struct inode   *inode, // IN:
    // sprintf(vnetHub[hubNum]->devName, "vmnet%d", hubNum);
 
    switch (iocmd) {
-#ifdef VNET_API_DEPRECATED
-   case SIOCSIFBR:
-#endif
    case SIOCSPEER:
    case SIOCBRIDGE:
       if (copy_from_user(name, (void *)ioarg, 8)) {
@@ -1049,9 +1008,6 @@ VNetFileOpIoctl(struct inode   *inode, // IN:
       break;
 #endif
 
-#ifdef VNET_API_DEPRECATED
-   case SIOCGIFBR:
-#endif
    case SIOCGBRSTATUS:
       {
          uint32 flags;
@@ -1081,15 +1037,6 @@ VNetFileOpIoctl(struct inode   *inode, // IN:
       }
       break;
 
-#ifdef VNET_API_DEPRECATED
-   /* We can't think of a good reason to use this ioctl. --gustav,hpreg */
-   case SIOCGIFFLAGS:
-      if (copy_to_user((void *)ioarg, &port->flags, sizeof port->flags)) {
-         return -EFAULT;
-      }
-      break;
-
-#endif
    case SIOCSIFFLAGS:
       if (copy_from_user(&port->flags, (void *)ioarg, sizeof port->flags)) {
          return -EFAULT;
@@ -1851,4 +1798,3 @@ MODULE_LICENSE("GPL v2");
  * by default (i.e., neither mkinitrd nor modprobe will accept it).
  */
 MODULE_INFO(supported, "external");
-
