@@ -37,11 +37,84 @@
 #include "basic_initblock.h"
 #include "x86segdescrs.h"
 #include "rateconv.h"
+#include "overheadmem_types.h"
 
 #ifndef VMX86_SERVER
 #include "numa_defs.h"
 #endif
 
+
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * VA64ToPtr --
+ *
+ *      Convert a VA64 to a pointer.
+ *
+ *      Usage of this function is strictly limited to these 2 cases:
+ *
+ *      1) In a VMX function which does an ioctl to vmmon, and receives a VMX
+ *         pointer as a result.
+ *
+ *      2) In the vmmon code, for the functions which have a VA64 and need
+ *         to call kernel APIs which take pointers.
+ *
+ * Results:
+ *      Virtual address.
+ *
+ * Side effects:
+ *      None
+ *
+ *-----------------------------------------------------------------------------
+ */
+
+static INLINE void *
+VA64ToPtr(VA64 va64) // IN
+{
+#ifdef VM_X86_64
+   ASSERT_ON_COMPILE(sizeof (void *) == 8);
+#else
+   ASSERT_ON_COMPILE(sizeof (void *) == 4);
+   // Check that nothing of value will be lost.
+   ASSERT(!(va64 >> 32));
+#endif
+   return (void *)(uintptr_t)va64;
+}
+
+
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * PtrToVA64 --
+ *
+ *      Convert a pointer to a VA64.
+ *
+ *      Usage of this function is strictly limited to these 2 cases:
+ *
+ *      1) In a VMX function which does an ioctl to vmmon, and passes in a VMX
+ *         pointer.
+ *
+ *      2) In the vmmon code, for the functions which need to pass in a kernel
+ *         pointer to functions which can take either a user or a kernel
+ *         pointer in the same parameter.
+ *
+ * Results:
+ *      Virtual address.
+ *
+ * Side effects:
+ *      None
+ *
+ *-----------------------------------------------------------------------------
+ */
+
+static INLINE VA64
+PtrToVA64(void const *ptr) // IN
+{
+   ASSERT_ON_COMPILE(sizeof ptr <= sizeof (VA64));
+   return (VA64)(uintptr_t)ptr;
+}
+
+
 /*
  * Driver version.
  *
@@ -60,7 +133,7 @@
  *
  */
 
-#define VMMON_VERSION           (209 << 16 | 0)
+#define VMMON_VERSION           (235 << 16 | 0)
 #define VMMON_VERSION_MAJOR(v)  ((uint32) (v) >> 16)
 #define VMMON_VERSION_MINOR(v)  ((uint16) (v))
 
@@ -115,7 +188,7 @@ enum IOCTLCmd {
     * handling 32 bit ioctl syscalls.  Hence FIRST and LAST.  FIRST must be
     * 2001 so that VERSION is 2001 for backwards compatibility.
     */
-#if defined(__linux__) || defined(_WIN32)
+#if defined __linux__ || defined _WIN32
    /* Start at 2001 because legacy code did. */
    IOCTLCMD(FIRST) = 2001,
 #else
@@ -132,13 +205,16 @@ enum IOCTLCmd {
    IOCTLCMD(LATE_INIT_VM),
    IOCTLCMD(RUN_VM),
    IOCTLCMD(LOOK_UP_MPN),
+#if defined __linux__
+   IOCTLCMD(LOOK_UP_LARGE_MPN),
+#endif
    IOCTLCMD(LOCK_PAGE),
    IOCTLCMD(UNLOCK_PAGE),
    IOCTLCMD(APIC_INIT),
    IOCTLCMD(SET_HARD_LIMIT),
    IOCTLCMD(GET_MEM_INFO),
    IOCTLCMD(ADMIT),
-   IOCTLCMD(SET_MEM_USAGE),
+   IOCTLCMD(UPDATE_MEM_INFO),
    IOCTLCMD(READMIT),
    IOCTLCMD(PAE_ENABLED),
 #ifndef __APPLE__
@@ -159,8 +235,8 @@ enum IOCTLCmd {
    IOCTLCMD(COW_CHECK),
    IOCTLCMD(COW_UPDATE_HINT),
    IOCTLCMD(COW_COPY_PAGE),
-   IOCTLCMD(COW_REMOVE_HINT),
    IOCTLCMD(COW_GET_ZERO_MPN),
+   IOCTLCMD(COW_INC_ZERO_REF),
     /* AWE calls */
    IOCTLCMD(ALLOC_LOCKED_PAGES),
    IOCTLCMD(FREE_LOCKED_PAGES),
@@ -186,17 +262,18 @@ enum IOCTLCmd {
    IOCTLCMD(SYNC_GET_TSCS),
    IOCTLCMD(SYNC_SET_TSCS),
 
+   IOCTLCMD(GET_IPI_VECTORS),
+   IOCTLCMD(SEND_IPI),
+
    /*
     * Keep host-specific calls at the end so they can be undefined
     * without renumbering the common calls.
     */
 
-#if __linux__ || defined __APPLE__
+#if defined __linux__ || defined __APPLE__
    IOCTLCMD(SET_UID),		// VMX86_DEVEL only
    IOCTLCMD(ALLOW_CORE_DUMP),
 
-   IOCTLCMD(BROADCAST_IPI),	// SMP 2.2.8+ only
-
    // Not implemented in the main branch.
    IOCTLCMD(REGISTER_PASSTHROUGH_IO),
    IOCTLCMD(REGISTER_PASSTHROUGH_IRQ),
@@ -221,10 +298,9 @@ enum IOCTLCmd {
 
 #if defined _WIN32 || defined __APPLE__
    IOCTLCMD(ALLOC_CONTIG_PAGES),
-   IOCTLCMD(SEND_IPI),
 #endif
 
-#if defined _WIN32 || __linux__
+#if defined _WIN32 || defined __linux__
    IOCTLCMD(ACK_USER_CALL),
 #endif
 
@@ -251,9 +327,12 @@ enum IOCTLCmd {
    IOCTLCMD(FAST_SUSP_RES_SET_OTHER_FLAG),
    IOCTLCMD(FAST_SUSP_RES_GET_MY_FLAG),
 
-#if __linux__
+#if defined __linux__
    IOCTLCMD(SET_HOST_CLOCK_PRIORITY),
    IOCTLCMD(VMX_ENABLED),
+   IOCTLCMD(IOMMU_SETUP_MMU),
+   IOCTLCMD(IOMMU_REGISTER_DEVICE),
+   IOCTLCMD(IOMMU_UNREGISTER_DEVICE),
 #endif
 
    // Must be last.
@@ -282,43 +361,41 @@ enum IOCTLCmd {
 	       METHOD_NEITHER, \
 	       FILE_ANY_ACCESS)
 
-#define IOCTL_VMX86_VERSION		VMIOCTL_BUFFERED(VERSION)
-#define IOCTL_VMX86_CREATE_VM		VMIOCTL_BUFFERED(CREATE_VM)
-#define IOCTL_VMX86_RELEASE_VM		VMIOCTL_BUFFERED(RELEASE_VM)
-#define IOCTL_VMX86_GET_NUM_VMS		VMIOCTL_BUFFERED(GET_NUM_VMS)
-#define IOCTL_VMX86_ALLOC_CROSSGDT	VMIOCTL_BUFFERED(ALLOC_CROSSGDT)
-#define IOCTL_VMX86_INIT_VM		VMIOCTL_BUFFERED(INIT_VM)
-#define IOCTL_VMX86_INIT_CROSSGDT	VMIOCTL_BUFFERED(INIT_CROSSGDT)
-#define IOCTL_VMX86_INIT_NUMA_INFO	VMIOCTL_BUFFERED(INIT_NUMA_INFO)
+#define IOCTL_VMX86_VERSION             VMIOCTL_BUFFERED(VERSION)
+#define IOCTL_VMX86_CREATE_VM           VMIOCTL_BUFFERED(CREATE_VM)
+#define IOCTL_VMX86_RELEASE_VM          VMIOCTL_BUFFERED(RELEASE_VM)
+#define IOCTL_VMX86_GET_NUM_VMS         VMIOCTL_BUFFERED(GET_NUM_VMS)
+#define IOCTL_VMX86_ALLOC_CROSSGDT      VMIOCTL_BUFFERED(ALLOC_CROSSGDT)
+#define IOCTL_VMX86_INIT_VM             VMIOCTL_BUFFERED(INIT_VM)
+#define IOCTL_VMX86_INIT_CROSSGDT       VMIOCTL_BUFFERED(INIT_CROSSGDT)
+#define IOCTL_VMX86_INIT_NUMA_INFO      VMIOCTL_BUFFERED(INIT_NUMA_INFO)
 #define IOCTL_VMX86_GET_NUMA_MEM_STATS  VMIOCTL_BUFFERED(GET_NUMA_MEM_STATS)
-#define IOCTL_VMX86_LATE_INIT_VM	VMIOCTL_BUFFERED(LATE_INIT_VM)
-#define IOCTL_VMX86_RUN_VM		VMIOCTL_NEITHER(RUN_VM)
-#define IOCTL_VMX86_SEND_IPI		VMIOCTL_NEITHER(SEND_IPI)
-#define IOCTL_VMX86_LOOK_UP_MPN		VMIOCTL_BUFFERED(LOOK_UP_MPN)
-#define IOCTL_VMX86_LOCK_PAGE		VMIOCTL_BUFFERED(LOCK_PAGE)
-#define IOCTL_VMX86_UNLOCK_PAGE		VMIOCTL_BUFFERED(UNLOCK_PAGE)
-#define IOCTL_VMX86_APIC_INIT		VMIOCTL_BUFFERED(APIC_INIT)
-#define IOCTL_VMX86_SET_HARD_LIMIT	VMIOCTL_BUFFERED(SET_HARD_LIMIT)
-#define IOCTL_VMX86_GET_MEM_INFO	VMIOCTL_BUFFERED(GET_MEM_INFO)
-#define IOCTL_VMX86_ADMIT  	        VMIOCTL_BUFFERED(ADMIT)
+#define IOCTL_VMX86_LATE_INIT_VM        VMIOCTL_BUFFERED(LATE_INIT_VM)
+#define IOCTL_VMX86_RUN_VM              VMIOCTL_NEITHER(RUN_VM)
+#define IOCTL_VMX86_SEND_IPI            VMIOCTL_NEITHER(SEND_IPI)
+#define IOCTL_VMX86_GET_IPI_VECTORS     VMIOCTL_BUFFERED(GET_IPI_VECTORS)
+#define IOCTL_VMX86_LOOK_UP_MPN         VMIOCTL_BUFFERED(LOOK_UP_MPN)
+#define IOCTL_VMX86_LOCK_PAGE           VMIOCTL_BUFFERED(LOCK_PAGE)
+#define IOCTL_VMX86_UNLOCK_PAGE         VMIOCTL_BUFFERED(UNLOCK_PAGE)
+#define IOCTL_VMX86_APIC_INIT           VMIOCTL_BUFFERED(APIC_INIT)
+#define IOCTL_VMX86_SET_HARD_LIMIT      VMIOCTL_BUFFERED(SET_HARD_LIMIT)
+#define IOCTL_VMX86_GET_MEM_INFO        VMIOCTL_BUFFERED(GET_MEM_INFO)
+#define IOCTL_VMX86_ADMIT               VMIOCTL_BUFFERED(ADMIT)
 #define IOCTL_VMX86_READMIT             VMIOCTL_BUFFERED(READMIT)
-#define IOCTL_VMX86_SET_MEM_USAGE       VMIOCTL_BUFFERED(SET_MEM_USAGE)
-#define IOCTL_VMX86_PAE_ENABLED		VMIOCTL_BUFFERED(PAE_ENABLED)
+#define IOCTL_VMX86_UPDATE_MEM_INFO     VMIOCTL_BUFFERED(UPDATE_MEM_INFO)
+#define IOCTL_VMX86_PAE_ENABLED         VMIOCTL_BUFFERED(PAE_ENABLED)
 #define IOCTL_VMX86_HOST_X86_64         VMIOCTL_BUFFERED(HOST_X86_64)
 #define IOCTL_VMX86_COW_SHARE           VMIOCTL_BUFFERED(COW_SHARE)
 #define IOCTL_VMX86_COW_CHECK           VMIOCTL_BUFFERED(COW_CHECK)
 #define IOCTL_VMX86_COW_UPDATE_HINT     VMIOCTL_BUFFERED(COW_UPDATE_HINT)
 #define IOCTL_VMX86_COW_COPY_PAGE       VMIOCTL_BUFFERED(COW_COPY_PAGE)
-#define IOCTL_VMX86_COW_REMOVE_HINT     VMIOCTL_BUFFERED(COW_REMOVE_HINT)
 #define IOCTL_VMX86_COW_GET_ZERO_MPN    VMIOCTL_BUFFERED(COW_GET_ZERO_MPN)
-#define IOCTL_VMX86_BEEP		VMIOCTL_BUFFERED(BEEP)
-#define IOCTL_VMX86_HARD_LIMIT_MONITOR_STATUS	VMIOCTL_BUFFERED(HARD_LIMIT_MONITOR_STATUS)
+#define IOCTL_VMX86_COW_INC_ZERO_REF    VMIOCTL_BUFFERED(COW_INC_ZERO_REF)
+#define IOCTL_VMX86_BEEP                VMIOCTL_BUFFERED(BEEP)
+#define IOCTL_VMX86_HARD_LIMIT_MONITOR_STATUS   VMIOCTL_BUFFERED(HARD_LIMIT_MONITOR_STATUS)
 #define IOCTL_VMX86_CHANGE_HARD_LIMIT   VMIOCTL_BUFFERED(CHANGE_HARD_LIMIT)
 #define IOCTL_VMX86_ALLOC_CONTIG_PAGES  VMIOCTL_BUFFERED(ALLOC_CONTIG_PAGES)
 #define IOCTL_VMX86_FREE_CONTIG_PAGES   VMIOCTL_BUFFERED(FREE_CONTIG_PAGES)
-#ifdef USE_BLUE_SCREEN
-#define IOCTL_VMX86_BLUE_SCREEN		VMIOCTL_BUFFERED(BLUE_SCREEN)
-#endif
 
 #define IOCTL_VMX86_GET_TOTAL_MEM_USAGE	VMIOCTL_BUFFERED(GET_TOTAL_MEM_USAGE)
 #define IOCTL_VMX86_ACK_USER_CALL       VMIOCTL_BUFFERED(ACK_USER_CALL)
@@ -329,24 +406,24 @@ enum IOCTLCmd {
 #define IOCTL_VMX86_SYNC_SET_TSCS       VMIOCTL_BUFFERED(SYNC_SET_TSCS)
 #define IOCTL_VMX86_READ_PAGE           VMIOCTL_BUFFERED(READ_PAGE)
 #define IOCTL_VMX86_WRITE_PAGE          VMIOCTL_BUFFERED(WRITE_PAGE)
-#define IOCTL_VMX86_LOCK_PAGE_NEW	VMIOCTL_BUFFERED(LOCK_PAGE_NEW)
+#define IOCTL_VMX86_LOCK_PAGE_NEW       VMIOCTL_BUFFERED(LOCK_PAGE_NEW)
 #define IOCTL_VMX86_UNLOCK_PAGE_BY_MPN  VMIOCTL_BUFFERED(UNLOCK_PAGE_BY_MPN)
-#define IOCTL_VMX86_ALLOC_LOCKED_PAGES	VMIOCTL_BUFFERED(ALLOC_LOCKED_PAGES)
+#define IOCTL_VMX86_ALLOC_LOCKED_PAGES  VMIOCTL_BUFFERED(ALLOC_LOCKED_PAGES)
 #define IOCTL_VMX86_FREE_LOCKED_PAGES   VMIOCTL_BUFFERED(FREE_LOCKED_PAGES)
 #define IOCTL_VMX86_GET_LOCKED_PAGES_LIST VMIOCTL_BUFFERED(GET_LOCKED_PAGES_LIST)
 
 #define IOCTL_VMX86_GET_KERNEL_PROC_ADDRESS  VMIOCTL_BUFFERED(GET_KERNEL_PROC_ADDRESS)
-#define IOCTL_VMX86_READ_VA64		VMIOCTL_BUFFERED(READ_VA64)
+#define IOCTL_VMX86_READ_VA64           VMIOCTL_BUFFERED(READ_VA64)
 #define IOCTL_VMX86_SET_MEMORY_PARAMS   VMIOCTL_BUFFERED(SET_MEMORY_PARAMS)
 
 #define IOCTL_VMX86_REMEMBER_KHZ_ESTIMATE VMIOCTL_BUFFERED(REMEMBER_KHZ_ESTIMATE)
 
-#define IOCTL_VMX86_APIC_ID		VMIOCTL_BUFFERED(APIC_ID)
-#define IOCTL_VMX86_SVM_ENABLED_CPU	VMIOCTL_BUFFERED(SVM_ENABLED_CPU)
-#define IOCTL_VMX86_VT_ENABLED_CPU	VMIOCTL_BUFFERED(VT_ENABLED_CPU)
-#define IOCTL_VMX86_VT_SUPPORTED_CPU	VMIOCTL_BUFFERED(VT_SUPPORTED_CPU)
-#define IOCTL_VMX86_GET_ALL_MSRS     	VMIOCTL_BUFFERED(GET_ALL_MSRS)
-#define IOCTL_VMX86_BROKEN_CPU_HELPER	VMIOCTL_BUFFERED(BROKEN_CPU_HELPER)
+#define IOCTL_VMX86_APIC_ID             VMIOCTL_BUFFERED(APIC_ID)
+#define IOCTL_VMX86_SVM_ENABLED_CPU     VMIOCTL_BUFFERED(SVM_ENABLED_CPU)
+#define IOCTL_VMX86_VT_ENABLED_CPU      VMIOCTL_BUFFERED(VT_ENABLED_CPU)
+#define IOCTL_VMX86_VT_SUPPORTED_CPU    VMIOCTL_BUFFERED(VT_SUPPORTED_CPU)
+#define IOCTL_VMX86_GET_ALL_MSRS        VMIOCTL_BUFFERED(GET_ALL_MSRS)
+#define IOCTL_VMX86_BROKEN_CPU_HELPER   VMIOCTL_BUFFERED(BROKEN_CPU_HELPER)
 #define IOCTL_VMX86_COUNT_PRESENT_PAGES	VMIOCTL_BUFFERED(COUNT_PRESENT_PAGES)
 
 #define IOCTL_VMX86_FAST_SUSP_RES_SET_OTHER_FLAG VMIOCTL_BUFFERED(FAST_SUSP_RES_SET_OTHER_FLAG)
@@ -357,7 +434,6 @@ enum IOCTLCmd {
 #define IOCTL_VMX86_CHECK_PSEUDO_TSC         VMIOCTL_BUFFERED(CHECK_PSEUDO_TSC)
 #define IOCTL_VMX86_GET_PSEUDO_TSC           VMIOCTL_NEITHER(GET_PSEUDO_TSC)
 #define IOCTL_VMX86_SET_HOST_CLOCK_PRIORITY  VMIOCTL_BUFFERED(SET_HOST_CLOCK_PRIORITY)
-
 #endif
 
 
@@ -398,32 +474,32 @@ enum IOCTLCmd {
  * -- edward
  */
 
-#define PAGE_LOCK_FAILED              -10001
-#define PAGE_LOCK_LIMIT_EXCEEDED      -10002
-#define PAGE_LOCK_TOUCH_FAILED        -10003
-#define PAGE_LOCK_IN_TRANSITION       -10004
+#define PAGE_LOCK_FAILED              (-10001)
+#define PAGE_LOCK_LIMIT_EXCEEDED      (-10002)
+#define PAGE_LOCK_TOUCH_FAILED        (-10003)
+#define PAGE_LOCK_IN_TRANSITION       (-10004)
 
-#define PAGE_LOCK_SYS_ERROR           -10010
-#define PAGE_LOCK_ALREADY_LOCKED      -10011
-#define PAGE_LOCK_MEMTRACKER_ERROR    -10012
-#define PAGE_LOCK_PHYSTRACKER_ERROR   -10013
-#define PAGE_LOCK_MDL_ERROR           -10014
+#define PAGE_LOCK_SYS_ERROR           (-10010)
+#define PAGE_LOCK_ALREADY_LOCKED      (-10011)
+#define PAGE_LOCK_MEMTRACKER_ERROR    (-10012)
+#define PAGE_LOCK_PHYSTRACKER_ERROR   (-10013)
+#define PAGE_LOCK_MDL_ERROR           (-10014)
 
 #define PAGE_UNLOCK_NO_ERROR		   0
-#define PAGE_UNLOCK_NOT_TRACKED	      -10100
-#define PAGE_UNLOCK_NO_MPN            -10101
-#define PAGE_UNLOCK_NOT_LOCKED        -10102
-#define PAGE_UNLOCK_TOUCH_FAILED      -10103
-#define PAGE_UNLOCK_MISMATCHED_TYPE   -10104
+#define PAGE_UNLOCK_NOT_TRACKED	      (-10100)
+#define PAGE_UNLOCK_NO_MPN            (-10101)
+#define PAGE_UNLOCK_NOT_LOCKED        (-10102)
+#define PAGE_UNLOCK_TOUCH_FAILED      (-10103)
+#define PAGE_UNLOCK_MISMATCHED_TYPE   (-10104)
 
-#define PAGE_LOOKUP_INVALID_ADDR      -10200
-#define PAGE_LOOKUP_BAD_HIGH_ADDR     -10201
-#define PAGE_LOOKUP_ZERO_ADDR         -10202
-#define PAGE_LOOKUP_SMALL_ADDR        -10203
-#define PAGE_LOOKUP_NOT_TRACKED          -10	// added to another code
-#define PAGE_LOOKUP_NO_MPN               -20	// added to another code
-#define PAGE_LOOKUP_NOT_LOCKED           -30	// added to another code
-#define PAGE_LOOKUP_NO_VM                -40	// added to another code
+#define PAGE_LOOKUP_INVALID_ADDR      (-10200)
+#define PAGE_LOOKUP_BAD_HIGH_ADDR     (-10201)
+#define PAGE_LOOKUP_ZERO_ADDR         (-10202)
+#define PAGE_LOOKUP_SMALL_ADDR        (-10203)
+#define PAGE_LOOKUP_NOT_TRACKED          (-10)	// added to another code
+#define PAGE_LOOKUP_NO_MPN               (-20)	// added to another code
+#define PAGE_LOOKUP_NOT_LOCKED           (-30)	// added to another code
+#define PAGE_LOOKUP_NO_VM                (-40)	// added to another code
 
 #define PAGE_LOCK_SUCCESS(mpn) ((int) (mpn) >= 0)
 #define PAGE_LOCK_SOFT_FAILURE(mpn) ((int) (mpn) <= PAGE_LOCK_FAILED && \
@@ -445,7 +521,7 @@ enum IOCTLCmd {
 
 #define PERFCTR_MAX_COUNTERS        2
 #define PERFCTR_INVALID_EVENT       0
-#define PERFCTR_INVALID_IRQ         -1
+#define PERFCTR_INVALID_IRQ         (-1)
 
 typedef struct PerfCtrInfo {
    uint32 eventNum;
@@ -463,7 +539,7 @@ typedef struct VMAPICInfo {
 } VMAPICInfo; 
 
 #ifndef MAX_PROCESSORS
-#define MAX_PROCESSORS 32
+#define MAX_PROCESSORS 64
 #endif
 
 typedef struct TSCSet {
@@ -475,6 +551,12 @@ typedef struct TSCSet {
    Atomic_uint32 valid[2 * ((MAX_PROCESSORS + 63) / 64)];
 } TSCSet;
 
+typedef struct PassthruIOMMUMap {
+   uint64 numPages;      // How many memory pages guest has
+   MPN    mpn[0];        // GPN->PPN mapping
+} PassthruIOMMUMap;
+
+
 static INLINE void
 TSCSet_SetValid(TSCSet *tscSet, unsigned cpuNum)
 {
@@ -512,9 +594,6 @@ typedef struct LockedPageLimit {
  * Data structures for the GET_MEM_INFO and ADMIT ioctls.
  */
 
-#define VMMEM_CUR_VM_ONLY        0x1
-#define VMMEM_GET_PERIODIC_STATS 0x2
-
 typedef struct VMMemMgmtInfo {
    uint32          minAllocation;   // minimum pages for vm
    uint32          maxAllocation;   // maximum pages the vm could lock
@@ -533,7 +612,14 @@ typedef struct VMMemMgmtInfo {
    PShare_MgmtInfo pshareMgmtInfo;  // management info for pshare scan rates
 } VMMemMgmtInfo;
 
-#define VMMEM_COW_HOT_PAGES 5
+typedef struct VMMemMgmtInfoPatch {
+   uint32          breaksAvg;
+   Percent         sharedPctAvg;
+   Percent         usedPct;
+   uint8           _pad[2];
+} VMMemMgmtInfoPatch;
+
+#define VMMEM_COW_HOT_PAGES 10
 
 typedef struct VMMemCOWHotPage {
    MPN             mpn;
@@ -562,8 +648,7 @@ typedef struct VMMemInfoArgs {
    uint32          maxLockedPages;     // effective limit on locked pages
    uint32          callerIndex;        // this vm's index memInfo array
    uint32          numVMs;             // number of running VMs
-   Bool            getPeriodicStats;   // flag to control collection of stats
-   uint8           _pad[3];
+   uint8           _pad[4];
    VMMemMgmtInfo   memInfo[1];
 } VMMemInfoArgs;
 
@@ -573,7 +658,7 @@ typedef struct VMMemInfoArgs {
 typedef struct VMMPNList {
    uint32    mpnCount;   // IN (and OUT on Mac OS)
    uint32    pad;
-   VA64      mpn32List;  // IN: user VA to an array of 32-bit MPNs.
+   VA64      mpn32List;  // IN: User VA of an array of 32-bit MPNs.
 } VMMPNList;
 
 typedef struct VARange {
@@ -586,22 +671,15 @@ typedef struct VARange {
 typedef struct VMMUnlockPageByMPN {
    MPN32     mpn;
    uint32    pad;
-   VA64      va;	    /* optional userspace VA */
+   VA64      uAddr;	    /* IN: User VA of the page (optional). */
 } VMMUnlockPageByMPN;
 
 typedef struct VMMReadWritePage {
    MPN32        mpn; // IN
    uint32       pad;
-   VA64         buf; // IN
+   VA64         uAddr; // IN: User VA of a PAGE_SIZE-large buffer.
 } VMMReadWritePage;
 
-#define VMM_CPP_MAXPAGES 400
-typedef struct VMMCountPresentPages {
-   uint32       numPages;                    /* number of pages in pageList */
-   uint32       pad_numPages;
-   VA64         pageList[VMM_CPP_MAXPAGES];  /* random sample of pages to check */
-} VMMCountPresentPages;
-
 struct passthrough_iorange {
    unsigned short ioBase;   /* Base of range to pass through. */
    unsigned short numPorts; /* Length of range. */
@@ -658,6 +736,11 @@ typedef struct VMNUMAMemStatsArgs {
    uint32	curNUMANode;
    uint32     	numPagesPerNode[NUMA_MAX_NODES]; //For each NUMA node
 } VMNUMAMemStatsArgs;
+
+typedef struct {
+   uint8 vectors[2];
+} IPIVectors;
+
 #endif
  
 /*
@@ -729,25 +812,13 @@ struct VMMonAffinity {
  */
 
 typedef struct VMAllocContiguousMem {
-   VA64   mpn32List;  // IN: 64-bit user VA to an array of 32-bit MPNs.
+   VA64   mpn32List;  // IN: User VA of an array of 32-bit MPNs.
    uint32 mpnCount; // IN
    uint32 order;    // IN
    MPN32  maxMPN;   // IN
    uint32 _pad[1];
 } VMAllocContiguousMem;
 #elif defined __APPLE__
-/*
- * Mac OS uses an ioctl to allocate contiguous locked pages.
- */
-
-typedef struct VMAllocSpecialMem {
-   uint32 numPages; // IN
-   VA32 uAddr;      // OUT
-} VMAllocSpecialMem;
-#endif
-
-
-#if defined __APPLE__
 #   include "iocontrolsMacos.h"
 #endif
 
