@@ -46,7 +46,10 @@
  *
  * ModuleCallType --
  *
- *      Enumaration of support calls done by the module
+ *      Enumeration of support calls done by the module.
+ *
+ *      If anything changes in the enum, please update KStatModuleCallPtrs
+ *      for stats purposes.
  *
  *----------------------------------------------------------------------
  */
@@ -69,11 +72,18 @@ typedef enum ModuleCallType {
    MODULECALL_USERTIMEOUT,
 
    MODULECALL_GET_RECYCLED_PAGE,
-   MODULECALL_RELEASE_ANON_PAGE,
+   MODULECALL_RELEASE_ANON_PAGES,
    MODULECALL_IS_ANON_PAGE,
 
    MODULECALL_YIELD,
 
+   /*
+    * Here "VMX" refers to the Intel VT VMX operation, not the VMware VMX
+    * userland process.
+    */
+   MODULECALL_START_VMX_OP,
+   MODULECALL_ALLOC_VMX_PAGE,
+
    MODULECALL_LAST                   // Number of entries. Must be the last one
 } ModuleCallType;
 
@@ -83,9 +93,6 @@ typedef enum ModuleCallType {
 #define MODULECALL_CROSS_PAGE_LEN    1
 #define MODULECALL_CROSS_PAGE_START  6
 
-// we don't want it in the final version
-//#define USE_BLUE_SCREEN
-
 #define MODULECALL_USERCALL_NONE     300
 
 /*
@@ -142,6 +149,8 @@ typedef struct WSModule {
    uint16 hostToVmm;     // offset from beg of header to Host{32,64}toVmm
    uint16 vmm32ToHost;   // offsets to other routines
    uint16 vmm64ToHost;
+   uint16 handleUD;
+   uint16 handleGP;
 
    union {               // offsets to patches
       struct {
@@ -185,7 +194,7 @@ typedef struct SwitchNMI {
 #ifdef VMX86_SERVER
 #define SHADOW_DR64(cp, n)    (cp->shadowDR[n].ureg64)
 #define SHADOW_DR32(cp, n)    (cp->shadowDR[n].ureg32)
-#else 
+#else
 #define SHADOW_DR64(cp, n)    (cp->_shadowDR[n].ureg64)
 #define SHADOW_DR32(cp, n)    (cp->_shadowDR[n].ureg32)
 #endif
@@ -196,6 +205,99 @@ typedef struct SwitchNMI {
 #define SHADOW_DR(cp, n) SHADOW_DR32(cp, n)
 #endif
 
+
+/*----------------------------------------------------------------------
+ *
+ * MAX_SWITCH_PT_PATCHES
+ *
+ *   This is the maximum number of patches that must be placed into
+ *   the monitor page tables so that two pages of the host GDT and the
+ *   crosspage can be accessed during worldswitch.
+ *
+ *----------------------------------------------------------------------
+ */
+#define MAX_SWITCH_PT_PATCHES 3
+
+/*----------------------------------------------------------------------
+ *
+ * WS_NMI_STRESS
+ *
+ *   When set to non-zero, this causes the NMI-safe worldswitch code
+ *   to be automatically stress tested by simulating NMIs arriving
+ *   between various instructions.
+ *
+ *   When set to zero, normal worldswitch operation occurs.
+ *
+ *   See the worldswitch assembly code for details.
+ *
+ *----------------------------------------------------------------------
+ */
+#define WS_NMI_STRESS 0
+
+
+/*----------------------------------------------------------------------
+ *
+ * VMM64PageTablePatch
+ *
+ *    Describes an entry in the monitor page table which needs to be
+ *    patched during the back-to-host worldswitch.
+ *
+ *    o A patch can appear at any place in the page table, and so four
+ *      items are required to uniquely describe the patch:
+ *
+ *      o level
+ *
+ *        This is the level in the page table to which the patch must
+ *        be applied: L4, L3, L2, L1.  This information is used to
+ *        determine the base of the region of memory which must be
+ *        patched.  The level value corresponds to the following
+ *        regions in monitor memory:
+ *
+ *          MMU_ROOT_64
+ *          MMU_L3_64
+ *          MMU_L2_64
+ *          MON_PAGE_TABLE_64
+ *
+ *        The value zero is reserved to indicate an empty spot in the
+ *        array of patches.
+ *
+ *      o level offset
+ *
+ *        The monitor memory regions corresponding to the page table
+ *        levels may be more than one page in length, so a 'page
+ *        offset' is required to know the starting address of the page
+ *        table page which must be patched in 'level'.
+ *
+ *      o page index
+ *
+ *        The 'index' value specifies the element in the page which
+ *        should be patched.
+ *
+ *      o pte
+ *
+ *        This is the PTE value which will be patched into the monitor
+ *        page table.
+ *
+ *----------------------------------------------------------------------
+ */
+typedef
+#include "vmware_pack_begin.h"
+struct VMM64PageTablePatch {
+#define PTP_EMPTY    (0U) /* Unused array entry. (must be 0) */
+#define PTP_LEVEL_L1 (1U)
+#define PTP_LEVEL_L2 (2U)
+#define PTP_LEVEL_L3 (3U)
+#define PTP_LEVEL_L4 (4U)
+   uint16   level;              /* [0, 4]  (maximal size: 3 bits) */
+   uint16   page;               /* Index of 'page' in 'level'.    */
+   uint32   index;              /* Index of 'pte' in 'page'.      */
+   VM_PDPTE pte;                /* PTE.                           */
+}
+#include "vmware_pack_end.h"
+VMM64PageTablePatch;
+
+#define MAX_DUMMY_VMCSES 16
+
 /*
  *----------------------------------------------------------------------
  *
@@ -211,12 +313,7 @@ typedef struct SwitchNMI {
  *----------------------------------------------------------------------
  */
 
-#define MAX_SWITCH_PT_PATCHES 3
-#define WS_NMI_STRESS 0  // 0: normal operation
-                         // 1: stress test worldswitch NMI handling
-                         //    (requires some real Detroit Muscle)
-
-typedef 
+typedef
 #include "vmware_pack_begin.h"
 struct VMCrossPage {
    /*
@@ -226,9 +323,9 @@ struct VMCrossPage {
    uint32 crosspage_size;
 
    /*
-    * Tiny stack that is used during switching so it can remain valid.  It's 
-    * good to keep the end 16-byte aligned for 64-bit processors.  There must 
-    * be enough room for two interrupt frames (in case of NMI during #DB 
+    * Tiny stack that is used during switching so it can remain valid.  It's
+    * good to keep the end 16-byte aligned for 64-bit processors.  There must
+    * be enough room for two interrupt frames (in case of NMI during #DB
     * handling), plus a half dozen registers.
     *
     * Also, the cache-line stuff below assumes we end on a 64-byte boundary.
@@ -322,8 +419,8 @@ struct VMCrossPage {
    uint64 hostDR[8];         //16
 
    /*
-    * Offset from start of crosspage to where worldswitch code module is 
-    * loaded.  The worldswitch code module is loaded at the very end of the 
+    * Offset from start of crosspage to where worldswitch code module is
+    * loaded.  The worldswitch code module is loaded at the very end of the
     * page that the VMCrossPage struct is loaded at the beginning of.
     *
     * Access fields of the worldswitch code module with the WSMODULE macro.
@@ -339,38 +436,37 @@ struct VMCrossPage {
    Task64      monTask64;                 /* vmm64's task */
 
    /*
-    * These values are used by BackToHost to patch in the crosspage at either 
-    * its host kernel linear address (host/monitor bitsizes match) or at its 
+    * These values are used by BackToHost to patch in the crosspage at either
+    * its host kernel linear address (host/monitor bitsizes match) or at its
     * machine address (host/monitor bitsize mismatch).
     *
-    * The L4 values are used by the 64-bit monitor and the indices index the 
+    * The L4 values are used by the 64-bit monitor and the indices index the
     * 64-bit monitor's L4 (root) page.  They are in the range 0..511.
     *
-    * The L2 values are used by the 32-bit monitor and the indices index the 
-    * 32-bit monitor's L2 (directory) page.  If the table is nonPAE, the number 
-    * is in the range 0..1023.  If the table is PAE, the number is in the range 
-    * 0..2047, as the monitor's L2 pages are virtually contiguous and so they 
+    * The L2 values are used by the 32-bit monitor and the indices index the
+    * 32-bit monitor's L2 (directory) page.  If the table is nonPAE, the number
+    * is in the range 0..1023.  If the table is PAE, the number is in the range
+    * 0..2047, as the monitor's L2 pages are virtually contiguous and so they
     * are treated as a single array.
     *
-    * For 32-bit monitor, the patch is determined by the *host* PAEness.  The 
-    * only mismatch case we allow is host nonPAE, monitor PAE.  In this case, 
-    * the 32-bit BackToHost routine switches to nonPAE mode before switching 
-    * and restores PAE mode before returning out to the monitor.  So as far as 
-    * these patches are concerned, they are done with host PAEness, not monitor 
+    * For 32-bit monitor, the patch is determined by the *host* PAEness.  The
+    * only mismatch case we allow is host nonPAE, monitor PAE.  In this case,
+    * the 32-bit BackToHost routine switches to nonPAE mode before switching
+    * and restores PAE mode before returning out to the monitor.  So as far as
+    * these patches are concerned, they are done with host PAEness, not monitor
     * hapPAEness (is a warm gun...).
     *
     * For nonPAE patches, the setup code guarantees the values fit in 32 bits.
     */
-   uint32 vmm32L2PIs[MAX_SWITCH_PT_PATCHES]; // vmm32: patch indices
-   uint32 vmm64L4PIs[MAX_SWITCH_PT_PATCHES]; // vmm64: patch indices
+   uint32 vmm32L2PIs[MAX_SWITCH_PT_PATCHES];   // vmm32: patch indices
+   uint64 vmm32L2PEs[MAX_SWITCH_PT_PATCHES];   // vmm32: patch entries
 
-   uint64 vmm32L2PEs[MAX_SWITCH_PT_PATCHES]; // vmm32: patch entries
-   uint64 vmm64L4PEs[MAX_SWITCH_PT_PATCHES]; // vmm64: patch entries
+   VMM64PageTablePatch vmm64PTP[MAX_SWITCH_PT_PATCHES]; /* page table patch */
 
    /*
-    * This is only needed when the host is nonPAE and the 32-bit monitor is 
-    * PAE.  It is a nonPAE L1 (table) page that maps the monitor's top 4M 
-    * address space.  It is not kept completely updated, it contains just 
+    * This is only needed when the host is nonPAE and the 32-bit monitor is
+    * PAE.  It is a nonPAE L1 (table) page that maps the monitor's top 4M
+    * address space.  It is not kept completely updated, it contains just
     * enough to switch from monitor to host and back.
     */
    MPN32 crossMonPageTableMPN;
@@ -454,6 +550,32 @@ struct VMCrossPage {
                              //  32-bit host: machine address
                              //  64-bit host: host kernel linear address
 
+   /*
+    * A VMXON page used to return to VMX operation when leaving the monitor,
+    * if we had to leave VMX operation to enter the monitor.  These pages
+    * are allocated per-PCPU, and this field must contain the VMXON page for
+    * the current PCPU when performing a world-switch.
+    */
+   MA     rootVMCS;
+
+   /*
+    * Dummy VMCSes used to ensure that the cached state of a foreign VMCS
+    * gets flushed to memory.
+    */
+   MA     dummyVMCS[MAX_DUMMY_VMCSES];
+
+   /*
+    * The current VMCS of a foreign hypervisor when we leave VMX operation
+    * to disable paging.
+    */
+   MA     foreignVMCS;
+
+   /*
+    * Set if the host is in VMX operation and we need to disable paging
+    * to switch between legacy mode and long mode.
+    */
+   uint32 inVMXOperation;
+
    /*
     * Set by Task_InitCrosspage to tell monitorHosted32.c what host CS to use.
     */
@@ -461,11 +583,11 @@ struct VMCrossPage {
    uint16 hostInitial64CS;
 
    /*
-    * If an NMI happens when switching from the host to the monitor, this flag 
-    * gets set by the monitor and it immediately loops back to the host where 
+    * If an NMI happens when switching from the host to the monitor, this flag
+    * gets set by the monitor and it immediately loops back to the host where
     * the host can forward the NMI to the host OS.  Ditto for MCE.
     */
-   uint32 hostToMonNMI;
+   uint32 retryWorldSwitch;
 
    /*
     * Descriptors and interrupt tables for switchNMI handlers.
@@ -478,7 +600,7 @@ struct VMCrossPage {
 
    uint32 switchMixIDT[76];        // uses CROSSGDT_*CS:MA
 
-   uint32 switchNMI;         // Offset from beginning of crosspage to where 
+   uint32 switchNMI;         // Offset from beginning of crosspage to where
                              // switchNMI.S is loaded (near end of crosspage).
                              // Use macro SWITCHNMI to access switchNMI fields.
 }
@@ -487,7 +609,7 @@ VMCrossPage;
 
 #define SWITCHNMI(crosspage) ((SwitchNMI *)((VA)(crosspage) + (crosspage)->switchNMI))
 
-#define CROSSPAGE_VERSION    (0x17A2 + WS_NMI_STRESS)  // increment by TWO
+#define CROSSPAGE_VERSION    (0x17AC + WS_NMI_STRESS)  // increment by TWO
 
 #if !defined(VMX86_SERVER) && defined(VMM)
 #define CROSS_PAGE  ((VMCrossPage * const) VPN_2_VA(CROSS_PAGE_START))
