@@ -50,6 +50,35 @@
 
 #ifdef __GNUC__
 
+#if defined __APPLE__
+/* PR 352418: GCC produces error if the non-Apple version is used */
+#define ASSERT_ON_COMPILE_SELECTOR_SIZE(expr)
+#else
+/* ASSERT_ON_COMPILE_SELECTOR_SIZE:
+ *
+ *  - Selector must be 16-bits.
+ *  - If a constant is used, it better be only 16-bits.
+ *  - If it's not a constant, it must be Selector-sized. or less.
+ *
+ * Although aesthetically the following looks nicer, gcc is unable
+ * to produce a constant expression for it:
+ *
+ *  ASSERT_ON_COMPILE(sizeof(Selector) == 2 &&                            \
+ *                    ((__builtin_constant_p(expr) ? ((expr) >> 16) == 0) \
+ *                                                 : sizeof(expr) <= 2)
+ */
+#if (__GNUC__ >= 4) && (__GNUC_MINOR__ >= 1)
+#define ASSERT_ON_COMPILE_SELECTOR_SIZE(expr)                                \
+   ASSERT_ON_COMPILE(sizeof(Selector) == 2 &&                                \
+                     ((__builtin_constant_p(expr) && ((expr) >> 16) == 0) || \
+                      sizeof(expr) <= 2))
+#else
+/* gcc 3.3.3 is not able to produce a constant expression (PR 356383) */
+#define ASSERT_ON_COMPILE_SELECTOR_SIZE(expr)
+#endif
+#endif
+
+
 /*
  * [GS]ET_[GI]DT() are defined as macros wrapping a function
  * so we can pass the argument implicitly by reference (requires
@@ -115,202 +144,187 @@ _Get_IDT(_GETSET_DTR_TYPE *dtr)
 }
 
 
-/* Checked against the Intel manual and GCC --hpreg */
-static INLINE void
-SET_LDT(Selector s)
-{
-   __asm__(
-      "lldt %0"
-      :
-      : "rm" (s)
-   );
-}
+#define SET_LDT(expr)                                                   \
+   do {                                                                 \
+      const Selector _set_ldt_sel = (Selector)(expr);                   \
+      ASSERT_ON_COMPILE_SELECTOR_SIZE(expr);                            \
+      /* Checked against the Intel manual and GCC --hpreg */            \
+      __asm__("lldt %0"                                                 \
+              :                                                         \
+              : "rm" (_set_ldt_sel));                                   \
+   } while (0)
 
 
 /* Checked against the Intel manual and GCC --hpreg
-
-   volatile because the LDT can change without the compiler knowing it
-   (when we use lldt). */
-static INLINE Selector
-_Get_ldt(void)
+ *
+ * volatile because the LDT can change without the compiler knowing it
+ * (when we use lldt).
+ */
+static INLINE void
+_GET_LDT(Selector * const result)
 {
-   Selector result;
-
-   __asm__ __volatile__(
-      "sldt %0"
-      : "=rm" (result)
-   );
-
-   return result;
+   __asm__ __volatile__("sldt %0"
+                        : "=rm" (*result));
 }
 
-#define GET_LDT(var) do { \
-   var = _Get_ldt();      \
-} while (0)
+
+#define GET_LDT(var)    \
+   do {                 \
+      _GET_LDT(&(var)); \
+   } while (0)
 
 
-/* Not yet checked against the Intel manual and GCC --slava */
-#define _BUILD_SET_R(func, reg) \
-static INLINE void              \
-func(uintptr_t r)               \
-{                               \
-   __asm__(                     \
-      "mov %0, %%" #reg         \
-      :                         \
-      : "r" (r)                 \
-      : "memory"                \
-   );                           \
-}
-
-_BUILD_SET_R(SET_CR0, cr0)
-_BUILD_SET_R(SET_CR2, cr2)
-_BUILD_SET_R(SET_CR3, cr3)
-_BUILD_SET_R(SET_CR4, cr4)
-
-#if defined __APPLE__
-_BUILD_SET_R(SET_DR0, db0)
-_BUILD_SET_R(SET_DR1, db1)
-_BUILD_SET_R(SET_DR2, db2)
-_BUILD_SET_R(SET_DR3, db3)
-_BUILD_SET_R(SET_DR6, db6)
-_BUILD_SET_R(SET_DR7, db7)
-#else
-_BUILD_SET_R(SET_DR0, dr0)
-_BUILD_SET_R(SET_DR1, dr1)
-_BUILD_SET_R(SET_DR2, dr2)
-_BUILD_SET_R(SET_DR3, dr3)
-_BUILD_SET_R(SET_DR6, dr6)
-_BUILD_SET_R(SET_DR7, dr7)
-#endif
-
+/* Checked against the Intel manual and GCC --thutt */
+#define _BUILD_SET_R(func, reg)        \
+   static INLINE void                  \
+   func(uintptr_t r)                   \
+   {                                   \
+      __asm__("mov %0, %%" #reg        \
+              : /* no outputs */       \
+              :  "r" (r)               \
+              : "memory");             \
+   }
 
 /* Not yet checked against the Intel manual and GCC --slava
+ *
+ * 'volatile' because CRs and DRs can change without the compiler
+ * knowing it (when there is a page fault, when a breakpoint occurs,
+ * and moreover it seems there is no way to teach gcc that smsw
+ * clobbers cr0 for example).
+ *
+ * The parameter is a 'uintptr_t *' so that the size of the actual
+ * parameter must exactly match the size of the hardware register.
+ * This prevents the use of 32-bit variables when building 64-bit
+ * code.
+ */
+#define _BUILD_GET_R(func, reg)                         \
+   static INLINE void                                   \
+   func(uintptr_t *result)                              \
+   {                                                    \
+      __asm__ __volatile__("mov %%" #reg ", %0"         \
+                           : "=r" (*result));           \
+   }
 
-   volatile because CRs and DRs can change without the compiler knowing it
-   (when there is a page fault, when a breakpoint occurs, and moreover it seems
-   there is no way to teach gcc that smsw clobbers cr0 for example). */
-#define _BUILD_GET_R(reg)   \
-static INLINE uintptr_t     \
-_Get_ ## reg(void)          \
-{                           \
-   uintptr_t result;        \
-                            \
-   __asm__ __volatile__(    \
-      "mov %%" #reg ", %0"  \
-      : "=r" (result)       \
-   );                       \
-                            \
-   return result;           \
-}
+_BUILD_SET_R(_SET_CR0, cr0)
+_BUILD_SET_R(_SET_CR2, cr2)
+_BUILD_SET_R(_SET_CR3, cr3)
+_BUILD_SET_R(_SET_CR4, cr4)
+_BUILD_SET_R(_SET_CR8, cr8)
 
-_BUILD_GET_R(cr0)
-#undef GET_CR0
-#define GET_CR0(var) do { \
-   var = _Get_cr0();      \
-} while (0)
-_BUILD_GET_R(cr2)
-#define GET_CR2(var) do { \
-   var = _Get_cr2();      \
-} while (0)
-_BUILD_GET_R(cr3)
-#define GET_CR3(var) do { \
-   var = _Get_cr3();      \
-} while (0)
-_BUILD_GET_R(cr4)
-#define GET_CR4(var) do { \
-   var = _Get_cr4();      \
-} while (0)
+_BUILD_GET_R(_GET_CR0, cr0)
+_BUILD_GET_R(_GET_CR2, cr2)
+_BUILD_GET_R(_GET_CR3, cr3)
+_BUILD_GET_R(_GET_CR4, cr4)
+_BUILD_GET_R(_GET_CR8, cr8)
 
 #if defined __APPLE__
-/* Mac OS gcc 4 is uses DBx instead of DRx register names. */
-_BUILD_GET_R(db0)
-#define GET_DR0(var) do { \
-   var = _Get_db0();      \
-} while (0)
-_BUILD_GET_R(db1)
-#define GET_DR1(var) do { \
-   var = _Get_db1();      \
-} while (0)
-_BUILD_GET_R(db2)
-#define GET_DR2(var) do { \
-   var = _Get_db2();      \
-} while (0)
-_BUILD_GET_R(db3)
-#define GET_DR3(var) do { \
-   var = _Get_db3();      \
-} while (0)
-_BUILD_GET_R(db6)
-#define GET_DR6(var) do { \
-   var = _Get_db6();      \
-} while (0)
-_BUILD_GET_R(db7)
-#define GET_DR7(var) do { \
-   var = _Get_db7();      \
-} while (0)
+/* Mac OS gcc 4 uses DBx instead of DRx register names. */
+_BUILD_SET_R(_SET_DR0, db0)
+_BUILD_SET_R(_SET_DR1, db1)
+_BUILD_SET_R(_SET_DR2, db2)
+_BUILD_SET_R(_SET_DR3, db3)
+_BUILD_SET_R(_SET_DR6, db6)
+_BUILD_SET_R(_SET_DR7, db7)
+
+_BUILD_GET_R(_GET_DR0, db0)
+_BUILD_GET_R(_GET_DR1, db1)
+_BUILD_GET_R(_GET_DR2, db2)
+_BUILD_GET_R(_GET_DR3, db3)
+_BUILD_GET_R(_GET_DR6, db6)
+_BUILD_GET_R(_GET_DR7, db7)
 #else
-_BUILD_GET_R(dr0)
-#define GET_DR0(var) do { \
-   var = _Get_dr0();      \
-} while (0)
-_BUILD_GET_R(dr1)
-#define GET_DR1(var) do { \
-   var = _Get_dr1();      \
-} while (0)
-_BUILD_GET_R(dr2)
-#define GET_DR2(var) do { \
-   var = _Get_dr2();      \
-} while (0)
-_BUILD_GET_R(dr3)
-#define GET_DR3(var) do { \
-   var = _Get_dr3();      \
-} while (0)
-_BUILD_GET_R(dr6)
-#define GET_DR6(var) do { \
-   var = _Get_dr6();      \
-} while (0)
-_BUILD_GET_R(dr7)
-#define GET_DR7(var) do { \
-   var = _Get_dr7();      \
-} while (0)
+_BUILD_SET_R(_SET_DR0, dr0)
+_BUILD_SET_R(_SET_DR1, dr1)
+_BUILD_SET_R(_SET_DR2, dr2)
+_BUILD_SET_R(_SET_DR3, dr3)
+_BUILD_SET_R(_SET_DR6, dr6)
+_BUILD_SET_R(_SET_DR7, dr7)
+
+_BUILD_GET_R(_GET_DR0, dr0)
+_BUILD_GET_R(_GET_DR1, dr1)
+_BUILD_GET_R(_GET_DR2, dr2)
+_BUILD_GET_R(_GET_DR3, dr3)
+_BUILD_GET_R(_GET_DR6, dr6)
+_BUILD_GET_R(_GET_DR7, dr7)
 #endif
 
+#define SET_CR_DR(regType, regNum, expr)                     \
+   do {                                                      \
+      /* Ensure no implicit truncation of 'expr' */          \
+      ASSERT_ON_COMPILE(sizeof(expr) <= sizeof(uintptr_t));  \
+      _SET_##regType##regNum(expr);                          \
+   } while (0)
 
-/* Checked against the Intel manual and GCC --hpreg */
-#define _BUILD_SET_SEG(func, reg) \
-static INLINE void                \
-func(Selector s)                  \
-{                                 \
-   __asm__(                       \
-      "movw %0, %%" #reg          \
-      :                           \
-      : "rm" (s)                  \
-   );                             \
-}
+#define GET_CR_DR(regType, regNum, var) \
+   do {                                 \
+      _GET_##regType##regNum(&(var));   \
+   } while (0)
 
-_BUILD_SET_SEG(SET_DS, ds)
-_BUILD_SET_SEG(SET_ES, es)
-_BUILD_SET_SEG(SET_FS, fs)
-_BUILD_SET_SEG(SET_GS, gs)
-_BUILD_SET_SEG(SET_SS, ss)
+#define SET_CR0(expr) SET_CR_DR(CR, 0, expr)
+#define SET_CR2(expr) SET_CR_DR(CR, 2, expr)
+#define SET_CR3(expr) SET_CR_DR(CR, 3, expr)
+#define SET_CR4(expr) SET_CR_DR(CR, 4, expr)
+#define SET_CR8(expr) SET_CR_DR(CR, 8, expr)
 
+/* Undefine GET_CR0; it is defined in mach_asm.h for SLES cross-compile */
+#undef GET_CR0
+#define GET_CR0(var)  GET_CR_DR(CR, 0, var)
+#define GET_CR2(var)  GET_CR_DR(CR, 2, var)
+#define GET_CR3(var)  GET_CR_DR(CR, 3, var)
+#define GET_CR4(var)  GET_CR_DR(CR, 4, var)
+#define GET_CR8(var)  GET_CR_DR(CR, 8, var)
+
+#define SET_DR0(expr) SET_CR_DR(DR, 0, expr)
+#define SET_DR1(expr) SET_CR_DR(DR, 1, expr)
+#define SET_DR2(expr) SET_CR_DR(DR, 2, expr)
+#define SET_DR3(expr) SET_CR_DR(DR, 3, expr)
+#define SET_DR6(expr) SET_CR_DR(DR, 6, expr)
+#define SET_DR7(expr) SET_CR_DR(DR, 7, expr)
+
+#define GET_DR0(var)  GET_CR_DR(DR, 0, var)
+#define GET_DR1(var)  GET_CR_DR(DR, 1, var)
+#define GET_DR2(var)  GET_CR_DR(DR, 2, var)
+#define GET_DR3(var)  GET_CR_DR(DR, 3, var)
+#define GET_DR6(var)  GET_CR_DR(DR, 6, var)
+#define GET_DR7(var)  GET_CR_DR(DR, 7, var)
+
+#define SET_SEGREG(reg, expr)                                   \
+   do {                                                         \
+      const Selector _set_segreg_sel = (Selector)(expr);        \
+      ASSERT_ON_COMPILE_SELECTOR_SIZE(expr);                    \
+      /* Checked against the Intel manual and GCC --hpreg */    \
+      __asm__("movw %0, %%" #reg                                \
+              :                                                 \
+              : "rm" (_set_segreg_sel));                        \
+   } while (0)
+
+#define SET_DS(expr) SET_SEGREG(ds, expr)
+#define SET_ES(expr) SET_SEGREG(es, expr)
+#define SET_FS(expr) SET_SEGREG(fs, expr)
+#define SET_GS(expr) SET_SEGREG(gs, expr)
+#define SET_SS(expr) SET_SEGREG(ss, expr)
 
 /* Checked against the Intel manual and GCC --hpreg
-
-   volatile because the content of CS can change without the compiler knowing
-   it (when we use call gates). */
-#define _BUILD_GET_SEG(func, reg) \
-static INLINE Selector            \
-func(void)                        \
-{                                 \
-   Selector result;               \
-                                  \
-   __asm__ __volatile__(          \
-      "movw %%" #reg ", %0"       \
-      : "=rm" (result)            \
-   );                             \
-                                  \
-   return result;                 \
-}
+ *
+ * volatile because the content of CS can change without the compiler
+ * knowing it (when we use call gates).
+ *
+ * XXX: The segment register getter functions have not been updated to
+ *      have stricter type checking like many other functions in this
+ *      file because they return a value, rather than taking an
+ *      argument.  Perhaps sometime in the future, a willing soul will
+ *      change these accessors to take an argument and at the same
+ *      time install better type checking.
+ */
+#define _BUILD_GET_SEG(func, reg)                \
+   static INLINE Selector                        \
+   func(void)                                    \
+   {                                             \
+      Selector result;                           \
+      __asm__ __volatile__("movw %%" #reg ", %0" \
+                           : "=rm" (result));    \
+      return result;                             \
+   }
 
 _BUILD_GET_SEG(GET_CS, cs)
 _BUILD_GET_SEG(GET_DS, ds)
@@ -320,41 +334,34 @@ _BUILD_GET_SEG(GET_GS, gs)
 _BUILD_GET_SEG(GET_SS, ss)
 
 
-/* Checked against the Intel manual and GCC --hpreg */
-static INLINE void
-SET_TR(Selector s)
-{
-   __asm__(
-      "ltr %0"
-      :
-      : "rm" (s)
-   );
-}
-
+#define SET_TR(expr)                                                    \
+   do {                                                                 \
+      const Selector _set_tr_sel = (Selector)(expr);                    \
+      ASSERT_ON_COMPILE_SELECTOR_SIZE(expr);                            \
+      /* Checked against the Intel manual and GCC --hpreg */            \
+      __asm__("ltr %0"                                                  \
+              :                                                         \
+              : "rm" (_set_tr_sel));                                    \
+   } while (0)
 
 /* Checked against the Intel manual and GCC --hpreg
 
    volatile because the content of TR can change without the compiler knowing
    it (when we use task gates). */
-static INLINE Selector
-_Get_TR(void)
+static INLINE void
+_GET_TR(Selector * const result)
 {
-   Selector result;
-
-   __asm__ __volatile__(
-      "str %0"
-      : "=rm" (result)
-   );
-
-   return result;
+   __asm__ __volatile__("str %0"
+                        : "=rm" (*result));
 }
 
-#define GET_TR(var) do { \
-   var = _Get_TR();      \
-} while (0)
+#define GET_TR(var)    \
+   do {                \
+      _GET_TR(&(var)); \
+   } while (0)
 
 
-/* Checked against the Intel manual and GCC --hpreg 
+/* Checked against the Intel manual and GCC --hpreg
 
    We use this to restore interrupts, so this cannot be reordered around
    by gcc */
@@ -428,7 +435,7 @@ CLTS(void)
 #define CLEAR_INTERRUPTS()  __asm__ __volatile__ ("cli": : :"memory")
 #endif
 #define RAISE_INTERRUPT(_x)  __asm__ __volatile__("int %0" :: "g" (_x))
-#define RETURN_FROM_INT()   __asm__ __volatile__("iret" :: ) 
+#define RETURN_FROM_INT()   __asm__ __volatile__("iret" :: )
 
 #define NO_INTERRUPTS_BEGIN()	do { \
                                    uintptr_t _flags; \
@@ -447,8 +454,8 @@ CLTS(void)
 #define SET_BYTE_FS(_fsAddr,_val)	__asm__ ("movb %1,%%fs:%0" ::  "m" (*(unsigned*)(_fsAddr)) ,"r" (_val));
 
 static INLINE void
-MEMCOPY_TO_FS(VA to, 
-              char * from, 
+MEMCOPY_TO_FS(VA to,
+              char * from,
               unsigned long n)
 {
    unsigned long i =0;
@@ -467,8 +474,8 @@ MEMCOPY_TO_FS(VA to,
 
 
 static INLINE void
-MEMCOPY_FROM_FS(char * to, 
-                VA from, 
+MEMCOPY_FROM_FS(char * to,
+                VA from,
                 unsigned long n)
 {
    unsigned long i =0;
@@ -483,7 +490,7 @@ MEMCOPY_FROM_FS(char * to,
 }
 /* End of the section whose correctness has NOT been checked */
 
-#elif _MSC_VER  /* !__GNUC__ */
+#elif defined _MSC_VER  /* !__GNUC__ */
 
 #define SET_IDT(_idt)    _Set_IDT(&(_idt))
 #define SET_GDT(_gdt)    _Set_GDT(&(_gdt))
@@ -549,12 +556,12 @@ MEMCOPY_FROM_FS(char * to,
 
 #define RAISE_INTERRUPT(_x)  {__asm int _x }
 #define RETURN_FROM_INT()   {__asm iretd }
-   
+
 
 #define SAVE_FLAGS(x) { \
      __asm pushfd __asm pop eax __asm mov x, eax \
 }
-   
+
 #define RESTORE_FLAGS(x) { \
      __asm push x __asm popfd\
 }
@@ -569,19 +576,19 @@ static INLINE void SET_DS(Selector val)
 
 static INLINE void SET_ES(Selector val)
 {
-   __asm mov ax, val 
+   __asm mov ax, val
    __asm mov es, ax
 }
 
 static INLINE void SET_FS(Selector val)
 {
-   __asm mov ax, val 
+   __asm mov ax, val
    __asm mov fs, ax
 }
 
 static INLINE void SET_GS(Selector val)
 {
-   __asm mov ax, val 
+   __asm mov ax, val
    __asm mov gs, ax
 }
 
@@ -636,37 +643,37 @@ static INLINE Selector GET_CS(void)
 
 #pragma warning( disable : 4035)
 
-static INLINE uint32  GET_WORD_FROM_FS(uint32 *_addr) { 
-	__asm mov eax, _addr 
-    __asm mov eax, fs:[eax] 
+static INLINE uint32  GET_WORD_FROM_FS(uint32 *_addr) {
+	__asm mov eax, _addr
+    __asm mov eax, fs:[eax]
 }
 
-static INLINE uint16  GET_SHORT_FROM_FS(uint16 *_addr) { 
-	__asm mov eax, _addr 
-    __asm mov ax, fs:[eax] 
+static INLINE uint16  GET_SHORT_FROM_FS(uint16 *_addr) {
+	__asm mov eax, _addr
+    __asm mov ax, fs:[eax]
 }
 
-static INLINE uint8  GET_BYTE_FROM_FS(uint8 *_addr) { 
-	__asm mov eax, _addr 
-     __asm mov al, fs:[eax] 
+static INLINE uint8  GET_BYTE_FROM_FS(uint8 *_addr) {
+	__asm mov eax, _addr
+     __asm mov al, fs:[eax]
 }
 
 #pragma warning (default: 4035)
 
 static INLINE void  SET_WORD_FS(uint32 *_addr, uint32 _val) {
-    __asm mov eax, _addr 
+    __asm mov eax, _addr
     __asm mov ebx, _val
     __asm mov fs:[eax], ebx
 }
 
 static INLINE void  SET_SHORT_FS(uint32 *_addr, uint16 _val) {
-    __asm mov eax, _addr 
+    __asm mov eax, _addr
     __asm mov bx, _val
     __asm mov fs:[eax], bx
 }
 
 static INLINE void  SET_BYTE_FS(uint32 *_addr, uint8 _val) {
-    __asm mov eax, _addr 
+    __asm mov eax, _addr
     __asm mov bl, _val
     __asm mov fs:[eax], bl
 }
@@ -715,8 +722,8 @@ static INLINE Selector _Get_TR(void) {
 
 
 static INLINE void
-MEMCOPY_TO_FS(VA to, 
-              char * from, 
+MEMCOPY_TO_FS(VA to,
+              char * from,
               unsigned long n)
 {
    unsigned long i =0;
@@ -741,8 +748,8 @@ MEMCOPY_TO_FS(VA to,
 
 
 static INLINE void
-MEMCOPY_FROM_FS(char * to, 
-                VA from, 
+MEMCOPY_FROM_FS(char * to,
+                VA from,
                 unsigned long n)
 {
    unsigned long i =0;
@@ -776,7 +783,7 @@ static INLINE unsigned CURRENT_CPL(void)
 {
    return SELECTOR_RPL(GET_CS());
 }
-#elif _MSC_VER
+#elif defined _MSC_VER
 static INLINE unsigned CURRENT_CPL(void) {
    volatile Selector _v;
    __asm mov ax, cs _asm mov _v, ax
@@ -817,7 +824,7 @@ static INLINE uint64 __GET_MSR(int cx)
    return msr;
 #endif
 }
-#elif _MSC_VER
+#elif defined _MSC_VER
 #pragma warning( disable : 4035)
 static INLINE uint64 __GET_MSR(int input)
 {
@@ -825,7 +832,7 @@ static INLINE uint64 __GET_MSR(int input)
    __asm mov  ecx, input
    __asm _emit 0x0f __asm _emit 0x32
    __asm pop ecx
-}   
+}
 
 static INLINE void __SET_MSR(int input, uint64 val)
 {
@@ -841,7 +848,7 @@ static INLINE void __SET_MSR(int input, uint64 val)
    __asm pop eax
    __asm pop ecx
    __asm pop edx
-}   
+}
 #pragma warning (default: 4035)
 #else
 #error
@@ -853,13 +860,13 @@ static INLINE void __SET_MSR(int cx, uint64 val)
 #ifdef VM_X86_64
    __asm__ __volatile__(
       "wrmsr"
-      : /* no outputs */			
+      : /* no outputs */
       : "a" ((uint32) val), "d" ((uint32)(val >> 32)), "c" (cx)
     );
 #else
    __asm__ __volatile__(
       "wrmsr"
-      : /* no outputs */			
+      : /* no outputs */
       : "A" (val),
         "c" (cx)
     );
@@ -869,10 +876,10 @@ static INLINE void __SET_MSR(int cx, uint64 val)
 
 
 /*
- * RDMSR/WRMSR access the 64bit MSRs as two 
- * 32 bit quantities, whereas GET_MSR/SET_MSR 
+ * RDMSR/WRMSR access the 64bit MSRs as two
+ * 32 bit quantities, whereas GET_MSR/SET_MSR
  * above access the MSRs as one 64bit quantity.
- */ 
+ */
 #ifdef __GNUC__
 #undef RDMSR
 #undef WRMSR
@@ -919,7 +926,7 @@ static INLINE uint64 RDPMC(int cx)
    return pmc;
 #endif
 }
-#elif _MSC_VER
+#elif defined _MSC_VER
 #ifndef VM_X86_64 // XXX Switch to intrinsics with the new 32 and 64-bit compilers.
 
 static INLINE uint64 RDPMC(int counter)
@@ -951,7 +958,7 @@ static INLINE void WRMSR(uint32 msrNum, uint32 lo, uint32 hi)
    SAVE_FLAGS(flags);     \
    flags &= ~EFLAGS_TF;   \
    RESTORE_FLAGS(flags);  \
-}  
+}
 
 
 static INLINE Bool
