@@ -83,8 +83,6 @@ typedef char           Bool;
 #endif
 
 #ifdef _MSC_VER
-typedef unsigned __int64 uint64;
-typedef signed __int64 int64;
 
 #pragma warning (3 :4505) // unreferenced local function
 #pragma warning (disable :4018) // signed/unsigned mismatch
@@ -95,6 +93,45 @@ typedef signed __int64 int64;
 #pragma warning (disable :4146) // unary minus operator applied to unsigned type, result still unsigned
 #pragma warning (disable :4142) // benign redefinition of type
 
+#endif
+
+#if defined(__APPLE__) || defined(HAVE_STDINT_H)
+
+/*
+ * TODO: This is a C99 standard header.  We should be able to test for
+ * #if __STDC_VERSION__ >= 199901L, but that breaks the Netware build
+ * (which doesn't have stdint.h).
+ */
+
+#include <stdint.h>
+
+typedef uint64_t uint64;
+typedef int64_t int64;
+typedef uint32_t uint32;
+typedef int32_t int32;
+typedef uint16_t uint16;
+typedef int16_t int16;
+typedef uint8_t uint8;
+
+/*
+ * XXX: int8_t is defined to be 'signed char' on Mac hosts.
+ *
+ * Unfortunately, GCC 4.0.1 warns when doing pointer assignment or
+ * comparison between signed char * and char * (even if char is
+ * signed).
+ *
+ * If we want to use int8_t to define int8, we need to go through and
+ * replace uses of char * with signed char * to prevent warnings.
+ */
+typedef char int8;
+
+#else /* !HAVE_STDINT_H */
+
+#ifdef _MSC_VER
+
+typedef unsigned __int64 uint64;
+typedef signed __int64 int64;
+
 #elif __GNUC__
 /* The Xserver source compiles with -ansi -pendantic */
 #ifndef __STRICT_ANSI__
@@ -105,13 +142,13 @@ typedef long int64;
 typedef unsigned long long uint64;
 typedef long long int64;
 #endif
-#elif __FreeBSD__
+#elif defined __FreeBSD__
 typedef unsigned long long uint64;
 typedef long long int64;
 #endif
 #else
 #error - Need compiler define for int64/uint64
-#endif
+#endif /* _MSC_VER */
 
 typedef unsigned int       uint32;
 typedef unsigned short     uint16;
@@ -121,6 +158,8 @@ typedef int       int32;
 typedef short     int16;
 typedef char      int8;
 
+#endif /* HAVE_STDINT_H */
+
 /*
  * FreeBSD (for the tools build) unconditionally defines these in
  * sys/inttypes.h so don't redefine them if this file has already
@@ -143,20 +182,23 @@ typedef char      int8;
 #ifdef HAVE_SYS_INTTYPES_H
 #include <sys/inttypes.h>
 #endif
-#ifdef HAVE_STDINT_H
-#include <stdint.h>
-#endif
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
 
+#ifdef __FreeBSD__
+#include <sys/param.h> /* For __FreeBSD_version */
+#endif
+
 #if !defined(USING_AUTOCONF)
 #   if defined(__FreeBSD__) || defined(sun)
 #      ifdef KLD_MODULE
 #         include <sys/types.h>
 #      else
-#         if (BSD_VERSION >= 50)
-#            include <inttypes.h>
+#         if __FreeBSD_version >= 500043
+#            if !defined(VMKERNEL)
+#               include <inttypes.h>
+#            endif
 #            include <sys/types.h>
 #         else
 #            include <sys/inttypes.h>
@@ -227,6 +269,12 @@ typedef int64 VmTimeVirtualClock;  /* Virtual Clock kept in CPU cycles */
       #define FMTPD      "I"
       #define FMTH       "I"
    #endif
+#elif defined __APPLE__
+   /* Mac OS hosts use the same formatters for 32- and 64-bit. */
+   #define FMT64 "ll"
+   #define FMTSZ "z"
+   #define FMTPD "l"
+   #define FMTH ""
 #elif __GNUC__
    #define FMTH ""
    #if defined(N_PLAT_NLM) || defined(sun) || \
@@ -248,9 +296,14 @@ typedef int64 VmTimeVirtualClock;  /* Virtual Clock kept in CPU cycles */
       || (defined(_POSIX_C_SOURCE) && _POSIX_C_SOURCE >= 200112L) \
       || (defined(_POSIX_VERSION) && _POSIX_VERSION >= 200112L) \
       || (defined(_POSIX2_VERSION) && _POSIX2_VERSION >= 200112L)
-      /* BSD/Darwin, Linux */
+      /* BSD, Linux */
       #define FMTSZ     "z"
-      #define FMTPD     "t"
+
+      #if defined(VM_X86_64)
+         #define FMTPD  "l"
+      #else
+         #define FMTPD  ""
+      #endif
    #else
       /* Systems with a pre-C99 libc */
       #define FMTSZ     "Z"
@@ -262,7 +315,7 @@ typedef int64 VmTimeVirtualClock;  /* Virtual Clock kept in CPU cycles */
    #endif
    #ifdef VM_X86_64
       #define FMT64     "l"
-   #elif defined(sun) || defined(__APPLE__) || defined(__FreeBSD__)
+   #elif defined(sun) || defined(__FreeBSD__)
       #define FMT64     "ll"
    #else
       #define FMT64     "L"
@@ -287,6 +340,9 @@ typedef int64 VmTimeVirtualClock;  /* Virtual Clock kept in CPU cycles */
 #ifdef _MSC_VER
 #define CONST64(c) c##I64
 #define CONST64U(c) c##uI64
+#elif defined __APPLE__
+#define CONST64(c) c##LL
+#define CONST64U(c) c##uLL
 #elif __GNUC__
 #ifdef VM_X86_64
 #define CONST64(c) c##L
@@ -316,6 +372,12 @@ typedef int64 VmTimeVirtualClock;  /* Virtual Clock kept in CPU cycles */
     #define CONST3264U(a) (a)
 #endif
 
+#define MIN_INT16  ((int16)0x8000)
+#define MAX_INT16  ((int16)0x7fff)
+
+#define MIN_UINT16 ((uint16)0)
+#define MAX_UINT16 ((uint16)0xffff)
+
 #define MIN_INT32  ((int32)0x80000000)
 #define MAX_INT32  ((int32)0x7fffffff)
 
@@ -356,7 +418,7 @@ typedef uint32    PageNum;
 typedef unsigned  MemHandle;
 typedef int32     World_ID;
 
-#define INVALID_WORLD_ID ((World_ID)0)
+#define INVALID_WORLD_ID ((World_ID)-1)
 
 typedef World_ID User_CartelID;
 #define INVALID_CARTEL_ID INVALID_WORLD_ID
@@ -367,6 +429,9 @@ typedef User_CartelID User_SessionID;
 typedef User_CartelID User_CartelGroupID;
 #define INVALID_CARTELGROUP_ID INVALID_CARTEL_ID
 
+typedef uint32 Worldlet_ID;
+#define INVALID_WORLDLET_ID ((Worldlet_ID)-1)
+
 /* world page number */
 typedef uint32    WPN;
 
@@ -435,10 +500,11 @@ typedef uint64 MPN64;
  */
 typedef VA32 UserVA32;
 typedef VA64 UserVA64;
-typedef UserVA32 UserVAConst; /* Userspace ptr to data that we may only read. */
+typedef UserVA64 UserVAConst; /* Userspace ptr to data that we may only read. */
+typedef UserVA32 UserVA32Const; /* Userspace ptr to data that we may only read. */
 typedef UserVA64 UserVA64Const; /* Used by 64-bit syscalls until conversion is finished. */
 #ifdef VMKERNEL
-typedef UserVA32 UserVA;
+typedef UserVA64 UserVA;
 #else
 typedef void * UserVA;
 #endif
@@ -458,8 +524,8 @@ typedef void * UserVA;
 #define INVALID_MPN  ((MPN)-1)
 #define MEMREF_MPN   ((MPN)-2)
 #define RESERVED_MPN ((MPN) 0)
-/* Support 39 bits of address space, minus one page. */
-#define MAX_MPN      ((MPN) 0x07ffffff)
+/* Support 43 bits of address space. */
+#define MAX_MPN      ((MPN)0x7fffffff)
 
 #define INVALID_LPN ((LPN)-1)
 #define INVALID_VPN ((VPN)-1)
@@ -483,8 +549,9 @@ typedef void * UserVA;
 #   define FMTVPN ""
 #endif
 
-
+#ifndef EXTERN
 #define EXTERN        extern
+#endif
 #define CONST         const
 
 
@@ -634,6 +701,12 @@ typedef void * UserVA;
 # define PRINTF_DECL(fmtPos, varPos)
 #endif
 
+#if defined(__GNUC__)
+# define SCANF_DECL(fmtPos, varPos) __attribute__((__format__(__scanf__, fmtPos, varPos)))
+#else
+# define SCANF_DECL(fmtPos, varPos)
+#endif
+
 /*
  * UNUSED_PARAM should surround the parameter name and type declaration,
  * e.g. "int MyFunction(int var1, UNUSED_PARAM(int var2))"
@@ -679,6 +752,52 @@ typedef void * UserVA;
 #define ALIGNED(n)
 #endif
 
+/*
+ ***********************************************************************
+ * STRUCT_OFFSET_CHECK --                                    */ /**
+ *
+ * \brief Check if the actual offsef of a member in a structure
+ *        is what is expected
+ *
+ *
+ * \param[in]  STRUCT       Structure the member is a part of.
+ * \param[in]  MEMBER       Member to check the offset of.
+ * \param[in]  OFFSET       Expected offset of MEMBER in STRUCTURE.
+ * \param[in]  DEBUG_EXTRA  Additional bytes to be added to OFFSET to
+ *                          compensate for extra info in debug builds.
+ *
+ ***********************************************************************
+ */
+#ifdef VMX86_DEBUG
+#define STRUCT_OFFSET_CHECK(STRUCT, MEMBER, OFFSET, DEBUG_EXTRA) \
+  ASSERT_ON_COMPILE(vmk_offsetof(STRUCT, MEMBER) == (OFFSET + DEBUG_EXTRA))
+#else
+#define STRUCT_OFFSET_CHECK(STRUCT, MEMBER, OFFSET, DEBUG_EXTRA) \
+  ASSERT_ON_COMPILE(vmk_offsetof(STRUCT, MEMBER) == OFFSET)
+#endif
+
+/*
+ ***********************************************************************
+ * STRUCT_SIZE_CHECK --                                      */ /**
+ *
+ * \brief Check if the actual size of a structure is what is expected
+ *
+ *
+ * \param[in]  STRUCT       Structure whose size is to be checked.
+ * \param[in]  SIZE         Expected size of STRUCT.
+ * \param[in]  DEBUG_EXTRA  Additional bytes to be added to SIZE to
+ *                          compensate for extra info in debug builds.
+ *
+ ***********************************************************************
+ */
+#ifdef VMX86_DEBUG
+#define STRUCT_SIZE_CHECK(STRUCT, SIZE, DEBUG_EXTRA) \
+  ASSERT_ON_COMPILE(sizeof(STRUCT) == (SIZE + DEBUG_EXTRA))
+#else
+#define STRUCT_SIZE_CHECK(STRUCT, SIZE, DEBUG_EXTRA) \
+  ASSERT_ON_COMPILE(sizeof(STRUCT) == SIZE)
+#endif
+
 /*
  * __func__ is a stringified function name that is part of the C99 standard. The block
  * below defines __func__ on older systems where the compiler does not support that
@@ -772,11 +891,11 @@ typedef void * UserVA;
 #endif
 
 /*
- * Format modifier for printing uid_t.  On sun the uid_t is a ulong, but on
- * Linux it's an int.
+ * Format modifier for printing uid_t.  On Solaris 10 and earlier, uid_t
+ * is a ulong, but on other platforms it's an unsigned int.
  * Use this like this: printf("The uid is %"FMTUID".\n", uid);
  */
-#ifdef sun
+#if defined(sun) && !defined(SOL11)
 #   ifdef VM_X86_64
 #      define FMTUID "u"
 #   else
@@ -844,4 +963,25 @@ typedef wchar_t utf16_t;
 typedef uint16 utf16_t;
 #endif
 
+/*
+ * Define for point and rectangle types.  Defined here so they
+ * can be used by other externally facing headers in bora/public.
+ */
+
+typedef struct VMPoint {
+   int x, y;
+} VMPoint;
+
+#if defined _WIN32 && defined USERLEVEL
+struct tagRECT;
+typedef struct tagRECT VMRect;
+#else
+typedef struct VMRect {
+   int left;
+   int top;
+   int right;
+   int bottom;
+} VMRect;
+#endif
+
 #endif  /* _VM_BASIC_TYPES_H_ */
