@@ -41,6 +41,10 @@
 #define APIC_DEFAULT_ADDRESS      0xfee00000
 #define APIC_ADDRESS_ALIGNMENT    4096
 #define APIC_ADDRESS_MASK         (APIC_ADDRESS_ALIGNMENT-1)
+#define APIC_VERSION_0x11         0x11
+#define XAPIC_VERSION_0x14        0x14
+#define APIC_MAXLVT_0x4           0x4
+#define XAPIC_MAXLVT_0x5          0x5
 #define APIC_VERSION_MASK         0xff
 #define APIC_MAX_LVT_MASK         0xff
 #define APIC_MAX_LVT_SHIFT        16
@@ -85,7 +89,7 @@
 #define APIC_PR_Y(_r)         (((_r) & APIC_PR_YMASK) >> 0)
 
 #define APIC_SVR_ONES         0x0000000f
-#define APIC_SVR_VECTOR       0x000000f0
+#define APIC_SVR_VECTOR       0x000000ff
 #define APIC_SVR_APICENABLE   0x00000100
 #define APIC_SVR_FOCUSCHECK   0x00000200
 
@@ -118,6 +122,7 @@
  * APIC_ID_MASK, so we use APIC_ID_BITS to avoid naming conflicts.
  */
 #define APIC_ID_BITS          0x0f000000
+#define XAPIC_ID_BITS         0xff000000
 #define XAPIC_ID_MASK         0xff000000
 #define APIC_ID_SHIFT         24
 #define APIC_LDR_BITS	      0xff000000
@@ -152,6 +157,10 @@
 #define APIC_DEST_ALL_INC     2
 #define APIC_DEST_ALL_EXC     3
 
+/* APIC physical mode broadcasts */
+#define APIC_DEST_BROADCAST   0x0f
+#define XAPIC_DEST_BROADCAST  0xff
+
 /* APIC IPI Command Register format */
 #define APIC_ICRHI_RESERVED		0x00ffffff
 #define APIC_ICRHI_DEST_MASK		0xff000000
