@@ -170,6 +170,7 @@ static INLINE void Desc_SetS(Descriptor *d, uint32 val)        { d->S        = v
 static INLINE void Desc_SetDPL(Descriptor *d, uint32 val)      { d->DPL      = val; }
 static INLINE void Desc_SetPresent(Descriptor *d, uint32 val)  { d->present  = val; }
 static INLINE void Desc_SetDB(Descriptor *d, uint32 val)       { d->DB       = val; }
+static INLINE void Desc_SetLongmode(Descriptor *d, uint32 val) { d->longmode = val; }
 static INLINE void Desc_SetGran(Descriptor *d, uint32 val)     { d->gran     = val; }
 
 static INLINE void Desc64_SetType(Descriptor64 *d, uint32 val)    { d->type     = val; }
@@ -460,10 +461,10 @@ Desc_UnpackLimit(uint32 limit)
    return limit;
 }
 
+
 /*
- * The intel manual is misleading on expand down limit checks.  Valid offsets
- * actually range from the limit+1 to 0xffff 0xffffffff depending on the D
- * bit in the descriptor.  
+ * For expand-down segments, valid offsets range from limit+1 to 
+ * 0xffff or 0xffffffff, depending on the D/B bit in the descriptor.  
  */
 static INLINE Bool
 Desc_InBoundsExpandDown(VA vaddr, VA limit, uint32 size, VA supremum)
@@ -473,6 +474,7 @@ Desc_InBoundsExpandDown(VA vaddr, VA limit, uint32 size, VA supremum)
 }
 
 
+/* For expand-up segments, valid offsets range from 0 to limit. */
 static INLINE Bool
 Desc_InBoundsExpandUp(VA vaddr, VA limit, uint32 size)
 {
@@ -480,39 +482,6 @@ Desc_InBoundsExpandUp(VA vaddr, VA limit, uint32 size)
 }
 
 
-/*
- *----------------------------------------------------------------------
- *
- * Desc_InBounds and Desc_LegalAccess --
- *
- *      Check whether the access is within the bounds of the segment.
- *
- * Results:
- *      TRUE is the access is within the bounds of the segment, 
- *      FALSE otherwise.
- *
- * Side effects:
- *      None.
- *
- *----------------------------------------------------------------------
- */
-static INLINE Bool
-Desc_InBounds(VA vaddr, VA limit, uint32 size, Bool expandUp, VA supremum)
-{
-   ASSERT(supremum == 0xffff || supremum == 0xffffffff);
-   return expandUp ? Desc_InBoundsExpandUp(vaddr, limit, size)
-                   : Desc_InBoundsExpandDown(vaddr, limit, size, supremum);
-}
-
-
-static INLINE Bool
-Desc_LegalAccess(const Descriptor *d, VA vaddr, uint32 size)
-{
-   return Desc_InBounds(vaddr, Desc_ExpandedLimit(d), size, 
-                        !DT_EXPAND_DOWN(*d), Desc_DB(d) ? 0xffffffff : 0xffff);
-}
-
-
 /*
  * Call Gates.
  */
