@@ -81,6 +81,7 @@ MSRQuery;
 #define MSR_BIOS_SIGN_ID      0x0000008b
 #define MSR_PERFCTR0          0x000000c1
 #define MSR_PERFCTR1          0x000000c2
+#define MSR_PLATFORM_INFO     0x000000ce // Intel Nehalem Family
 #define MSR_MTRR_CAP          0x000000fe
 #define MSR_L2CFG             0x0000011e
 #define MSR_SYSENTER_CS       0x00000174
@@ -91,7 +92,6 @@ MSRQuery;
 #define MSR_MCG_CTL           0x0000017b
 #define MSR_EVNTSEL0          0x00000186
 #define MSR_EVNTSEL1          0x00000187
-#define MSR_PERF_STS          0x00000198 // Intel Core Solo and up
 #define MSR_MISC_ENABLE       0x000001a0
 #define MSR_DEBUGCTL          0x000001d9
 #define MSR_EFER              0xc0000080
@@ -100,6 +100,24 @@ MSRQuery;
 #define MSR_KERNELGSBASE      0xc0000102
 #define MSR_TSC_AUX           0xc0000103
 
+/* Intel Core Architecture and later: use only architected counters. */
+#define IA32_MSR_PERF_CAPABILITIES                0x345
+#define MSR_PERF_CAPABILITIES_LBRFMT_SHIFT        0
+#define MSR_PERF_CAPABILITIES_LBRFMT_MASK         0x3f
+#define MSR_PERF_CAPABILITIES_PEBSTRAP            (1u << 6)
+#define MSR_PERF_CAPABILITIES_PEBSSAVEARCHREGS    (1u << 7)
+#define MSR_PERF_CAPABILITIES_PEBSRECORDFMT_SHIFT 8
+#define MSR_PERF_CAPABILITIES_PEBSRECORDFMT_MASK  0xf
+#define MSR_PERF_CAPABILITIES_FREEZE_WHILE_SMM    (1u << 12)
+
+typedef enum {
+   SL_PMC_FLAGS_NONE             = 0x00, /* No flags.                      */
+   SL_PMC_FLAGS_LBR_VA32         = 0x01, /* LBR format: 32-bit VA.         */
+   SL_PMC_FLAGS_LBR_LA64         = 0x02, /* LBR format: 64-bit LA.         */
+   SL_PMC_FLAGS_LBR_VA64         = 0x04, /* LBR format: 64-bit VA.         */
+   SL_PMC_FLAGS_LBR_PACKED_VA32  = 0x08, /* LBR format: 2x32-bit VAs.      */
+} StateLoggerPMCFlags;
+
 #define MSR_MTRR_BASE0        0x00000200
 #define MSR_MTRR_MASK0        0x00000201
 #define MSR_MTRR_BASE1        0x00000202
@@ -149,19 +167,28 @@ MSRQuery;
 #define MSR_LASTBRANCH_2     0x000001dd // Intel Pentium4 Family
 #define MSR_LASTBRANCH_3     0x000001de // Intel Pentium4 Family
 
-/* Enhanced SpeedStep MSRs */
+#define CORE_LBR_SIZE        8
+#define CORE2_LBR_SIZE       4
+
+/* Power Management MSRs */
 #define MSR_PERF_STATUS      0x00000198 // Current Performance State (ro)
 #define MSR_PERF_CTL         0x00000199 // Target Performance State (rw)
+#define MSR_POWER_CTL        0x000001fc // Power Control Register
+#define MSR_CST_CONFIG_CTL   0x000000e2 // C-state Configuration (CORE)
+#define MSR_MISC_PWR_MGMT    0x000001aa // Misc Power Management (NHM)
+
+/* MSR_POWER_CTL bits (Intel) */
+#define MSR_POWER_CTL_C1E    0x00000001 // C1E enable (NHM)
 
 /* P-State Hardware Coordination Feedback Capability (Intel) */
-#define MSR_MPERF            0x000000E7 // Maximum Performance (rw)
-#define MSR_APERF            0x000000E8 // Actual Performance (rw)
+#define MSR_MPERF            0x000000e7 // Maximum Performance (rw)
+#define MSR_APERF            0x000000e8 // Actual Performance (rw)
 
 /* Software Controlled Clock Modulation and Thermal Monitors (Intel) */
-#define MSR_CLOCK_MODULATION 0x0000019A // Thermal Monitor Control (rw)
-#define MSR_THERM_INTERRUPT  0x0000019B // Thermal Interrupt Control (rw)
-#define MSR_THERM_STATUS     0x0000019C // Thermal Monitor Status (rw)
-#define MSR_THERM2_CTL       0x0000019D // Thermal Monitor 2 Control (ro)
+#define MSR_CLOCK_MODULATION 0x0000019a // Thermal Monitor Control (rw)
+#define MSR_THERM_INTERRUPT  0x0000019b // Thermal Interrupt Control (rw)
+#define MSR_THERM_STATUS     0x0000019c // Thermal Monitor Status (rw)
+#define MSR_THERM2_CTL       0x0000019d // Thermal Monitor 2 Control (ro)
 
 /* MSR_MISC_ENABLE bits (Intel) */
 #define MSR_MISC_ENABLE_FOPCODE_COMPAT  (1LL<<2)
@@ -170,6 +197,7 @@ MSRQuery;
 #define MSR_MISC_ENABLE_TM2             (1LL<<13) // Enable Thermal Monitor 2
 #define MSR_MISC_ENABLE_ESS             (1LL<<16) // Enable Enhanced SpeedStep
 #define MSR_MISC_ENABLE_LIMIT_CPUID     (1LL<<22) // Enable CPUID maxval
+#define MSR_MISC_ENABLE_C1E             (1LL<<25) // C1E enable (Merom/Penryn)
 #define MSR_MISC_ENABLE_TURBO_DISABLE   (1LL<<38) // Turbo Mode Disabled
 
 /* DebugCtlMSR bits */
@@ -192,7 +220,7 @@ MSRQuery;
 #define MSR_EFER_LMA         0x0000000000000400ULL  /* Long mode active: r/o */
 #define MSR_EFER_NXE         0x0000000000000800ULL  /* No-exec enable:   r/w */
 #define MSR_EFER_SVME        0x0000000000001000ULL  /* SVM(AMD) enabled? r/w */
-#define MSR_EFER_TRUNCE      0x0000000000002000ULL  /* trun. seg. enable:r/w */
+#define MSR_EFER_LMSLE       0x0000000000002000ULL  /* LM seg lim enable:r/w */
 #define MSR_EFER_FFXSR       0x0000000000004000ULL  /* Fast FXSAVE:      r/w */
 #define MSR_EFER_MBZ         0xffffffffffff8200ULL  /* Must be zero (resrvd) */
 
@@ -212,16 +240,16 @@ MSRQuery;
 #define MSR_K8_TOPMEM2       0xc001001d
 
 /* AMD "Greyhound" P-state MSRs */
-#define MSR_GH_PSTATE_LIMIT      0xC0010061  // P-state Limit Register
-#define MSR_GH_PSTATE_CONTROL    0xC0010062  // P-state Control Register [2:0]
-#define MSR_GH_PSTATE_STATUS     0xC0010063  // P-state Status Register [2:0]
-#define MSR_GH_PSTATE0           0xC0010064  // P-state 0
-#define MSR_GH_PSTATE1           0xC0010065  // P-state 1
-#define MSR_GH_PSTATE2           0xC0010066  // P-state 2
-#define MSR_GH_PSTATE3           0xC0010067  // P-state 3
-#define MSR_GH_PSTATE4           0xC0010068  // P-state 4
-#define MSR_GH_COFVID_CONTROL    0xC0010070  // COFVID Control Register
-#define MSR_GH_COFVID_STATUS     0xC0010071  // COFVID Status Register
+#define MSR_GH_PSTATE_LIMIT      0xc0010061  // P-state Limit Register
+#define MSR_GH_PSTATE_CONTROL    0xc0010062  // P-state Control Register [2:0]
+#define MSR_GH_PSTATE_STATUS     0xc0010063  // P-state Status Register [2:0]
+#define MSR_GH_PSTATE0           0xc0010064  // P-state 0
+#define MSR_GH_PSTATE1           0xc0010065  // P-state 1
+#define MSR_GH_PSTATE2           0xc0010066  // P-state 2
+#define MSR_GH_PSTATE3           0xc0010067  // P-state 3
+#define MSR_GH_PSTATE4           0xc0010068  // P-state 4
+#define MSR_GH_COFVID_CONTROL    0xc0010070  // COFVID Control Register
+#define MSR_GH_COFVID_STATUS     0xc0010071  // COFVID Status Register
 
 /* Syscall/Sysret related MSRs (x86_64) */
 #define MSR_STAR             0xc0000081 // Also present on Athlons.
@@ -251,9 +279,17 @@ MSRQuery;
 #define MTRR_TYPE_WB          6
 
 /*
- * PERF_STS bits
+ * PERF_STATUS bits
  */
-#define MSR_PERF_STS_BUS_RATIO_MAX_SHIFT 40
-#define MSR_PERF_STS_BUS_RATIO_MAX_MASK  0x1f
+#define MSR_PERF_STATUS_MAX_BUS_RATIO_SHIFT 40
+#define MSR_PERF_STATUS_MAX_BUS_RATIO_MASK  0x1f
+
+/*
+ * PLATFORM_INFO bits
+ */
+#define MSR_PLATFORM_INFO_MAX_NONTURBO_RATIO_SHIFT 8
+#define MSR_PLATFORM_INFO_MAX_NONTURBO_RATIO_MASK 0xff
+
+
 
 #endif /* _X86MSR_H_ */
