@@ -64,29 +64,17 @@ sys_ioctl(unsigned int fd, unsigned int cmd, unsigned long arg);
 #include "vmci_infrastructure.h"
 #include "vmciCommonInt.h"
 #include "vmciDriver.h"
+#include "vmciHostKernelAPI.h"
 #include "vmciDatagram.h"
 #include "vmciProcess.h"
 #include "vmciDsInt.h"
 #include "vmciGroup.h"
 #include "vmciQueuePair.h"
 
-/* Compatibility reference counting */
-#ifdef KERNEL_2_4_0
-#   define compat_fop_set_owner(_pFop) do { \
-   (_pFop)->owner = THIS_MODULE;            \
-} while (0)
-#   define compat_mod_inc_refcount
-#   define compat_mod_dec_refcount
-#else
-#   define compat_fop_set_owner(_pFop)
-#   define compat_mod_inc_refcount do { \
-   MOD_INC_USE_COUNT;                   \
-} while (0)
-#   define compat_mod_dec_refcount do { \
-   MOD_DEC_USE_COUNT;                   \
-} while (0)
-#endif
-
+#include "compat_uaccess.h"
+#include "compat_highmem.h"
+#include "compat_mm.h"
+#include "pgtbl.h"
 
 /*
  * Per-instance driver state
@@ -98,6 +86,7 @@ typedef struct VMCILinux {
       VMCIProcess *process;
       VMCIDatagramProcess *dgmProc;
    } ct;
+   int userVersion;
    VMCIObjType ctType;
 #if defined(HAVE_COMPAT_IOCTL) || defined(HAVE_UNLOCKED_IOCTL)
    struct semaphore lock;
@@ -129,6 +118,8 @@ typedef struct VMCILinuxState {
 
 static struct VMCILinuxState linuxState;
 
+static int VMCISetupNotify(VMCIContext *context, VA notifyUVA);
+
 
 /*
  *----------------------------------------------------------------------
@@ -186,8 +177,21 @@ register_ioctl32_handlers(void)
 #ifndef HAVE_COMPAT_IOCTL
    {
       int i;
+
       for (i = IOCTL_VMCI_FIRST; i < IOCTL_VMCI_LAST; i++) {
-         int retval = register_ioctl32_conversion(i, LinuxDriver_Ioctl32_Handler);
+         int retval = register_ioctl32_conversion(i,
+                                                  LinuxDriver_Ioctl32_Handler);
+
+         if (retval) {
+            Warning("Fail to register ioctl32 conversion for cmd %d\n", i);
+            return retval;
+         }
+      }
+
+      for (i = IOCTL_VMCI_FIRST2; i < IOCTL_VMCI_LAST2; i++) {
+         int retval = register_ioctl32_conversion(i,
+                                                  LinuxDriver_Ioctl32_Handler);
+
          if (retval) {
             Warning("Fail to register ioctl32 conversion for cmd %d\n", i);
             return retval;
@@ -204,8 +208,18 @@ unregister_ioctl32_handlers(void)
 #ifndef HAVE_COMPAT_IOCTL
    {
       int i;
+
       for (i = IOCTL_VMCI_FIRST; i < IOCTL_VMCI_LAST; i++) {
          int retval = unregister_ioctl32_conversion(i);
+
+         if (retval) {
+            Warning("Fail to unregister ioctl32 conversion for cmd %d\n", i);
+         }
+      }
+
+      for (i = IOCTL_VMCI_FIRST2; i < IOCTL_VMCI_LAST2; i++) {
+         int retval = unregister_ioctl32_conversion(i);
+
          if (retval) {
             Warning("Fail to unregister ioctl32 conversion for cmd %d\n", i);
          }
@@ -253,7 +267,7 @@ init_module(void)
     */
 
    memset(&vmuser_fops, 0, sizeof vmuser_fops);
-   compat_fop_set_owner(&vmuser_fops);
+   vmuser_fops.owner = THIS_MODULE;
    vmuser_fops.poll = LinuxDriverPoll;
 #ifdef HAVE_UNLOCKED_IOCTL
    vmuser_fops.unlocked_ioctl = LinuxDriver_UnlockedIoctl;
@@ -350,15 +364,13 @@ LinuxDriver_Open(struct inode *inode, // IN
 {
    VMCILinux *vmciLinux;
 
-   compat_mod_inc_refcount;
-
    vmciLinux = kmalloc(sizeof *vmciLinux, GFP_KERNEL);
    if (vmciLinux == NULL) {
-      compat_mod_dec_refcount;
       return -ENOMEM;
    }
    memset(vmciLinux, 0, sizeof *vmciLinux);
    vmciLinux->ctType = VMCIOBJ_NOT_SET;
+   vmciLinux->userVersion = 0;
 #if defined(HAVE_COMPAT_IOCTL) || defined(HAVE_UNLOCKED_IOCTL)
    init_MUTEX(&vmciLinux->lock);
 #endif
@@ -418,7 +430,6 @@ LinuxDriver_Close(struct inode *inode, // IN
 
    kfree(vmciLinux);
    filp->private_data = NULL;
-   compat_mod_dec_refcount;
    return 0;
 }
 
@@ -513,9 +524,31 @@ LinuxDriver_Ioctl(struct inode *inode,
          retval = -EFAULT;
          break;
       }
+
+      vmciLinux->userVersion = verFromUser;
    }
       /* Fall through. */
    case IOCTL_VMCI_VERSION:
+      /*
+       * The basic logic here is:
+       *
+       * If the user sends in a version of 0 tell it our version.
+       * If the user didn't send in a version, tell it our version.
+       * If the user sent in an old version, tell it -its- version.
+       * If the user sent in an newer version, tell it our version.
+       *
+       * The rationale behind telling the caller its version is that
+       * Workstation 6.5 required that VMX and VMCI kernel module were
+       * version sync'd.  All new VMX users will be programmed to
+       * handle the VMCI kernel module version.
+       */
+
+      if (vmciLinux->userVersion > 0 &&
+          vmciLinux->userVersion < VMCI_VERSION_HOSTQP) {
+         retval = vmciLinux->userVersion;
+      } else {
+         retval = VMCI_VERSION;
+      }
       retval = VMCI_VERSION;
       break;
 
@@ -542,15 +575,15 @@ LinuxDriver_Ioctl(struct inode *inode,
          goto init_release;
       }
 
-      if ((retval = VMCIContext_InitContext(initBlock.cid, initBlock.flags,
-					    -1, &vmciLinux->ct.context)) <
-          VMCI_SUCCESS) {
+      retval = VMCIContext_InitContext(initBlock.cid, initBlock.flags,
+                                       0 /* Unused */, vmciLinux->userVersion,
+                                       &vmciLinux->ct.context);
+      if (retval < VMCI_SUCCESS) {
          Log("VMCI: Error initializing context.\n");
          retval = retval == VMCI_ERROR_DUPLICATE_ENTRY ? -EEXIST : -EINVAL;
          goto init_release;
       }
 
-
       /* 
        * Copy cid to userlevel, we do this to allow the VMX to enforce its 
        * policy on cid generation.
@@ -587,7 +620,7 @@ LinuxDriver_Ioctl(struct inode *inode,
          goto create_release;
       }
 
-      if (VMCIProcess_Create(&vmciLinux->ct.process, -1) < VMCI_SUCCESS) {
+      if (VMCIProcess_Create(&vmciLinux->ct.process) < VMCI_SUCCESS) {
          Log("VMCI: Error initializing process.\n");
          retval = -EINVAL;
          goto create_release;
@@ -619,7 +652,8 @@ LinuxDriver_Ioctl(struct inode *inode,
       }
 
       /* Create process and datagram. */
-      if (VMCIDatagramProcess_Create(&dgmProc, &dgCreateInfo) < VMCI_SUCCESS) {
+      if (VMCIDatagramProcess_Create(&dgmProc, &dgCreateInfo,
+                                     0 /* Unused */) < VMCI_SUCCESS) {
 	 retval = -EINVAL;
 	 goto create_dg_release;
       }
@@ -685,8 +719,8 @@ LinuxDriver_Ioctl(struct inode *inode,
          break;
       }
 
-      VMCI_LOG(("VMCI: Datagram dst handle 0x%"FMT64"x, src handle 0x%"FMT64"x, "
-	        "payload size %"FMT64"u.\n", 
+      VMCI_DEBUG_LOG(("VMCI: Datagram dst handle 0x%"FMT64"x, "
+	        "src handle 0x%"FMT64"x, payload size %"FMT64"u.\n",
                 dg->dstHandle, dg->srcHandle, dg->payloadSize));
 
       /* Get source context id. */
@@ -770,14 +804,18 @@ LinuxDriver_Ioctl(struct inode *inode,
       QueuePairList_Lock();
 
       {
-	 QueuePairPageStore pageStore = { queuePairAllocInfo.producePageFile,
+	 QueuePairPageStore pageStore = { TRUE,
+					  queuePairAllocInfo.producePageFile,
 					  queuePairAllocInfo.consumePageFile,
 					  queuePairAllocInfo.producePageFileSize,
-					  queuePairAllocInfo.consumePageFileSize };
+					  queuePairAllocInfo.consumePageFileSize,
+					  0,
+					  0 };
 
 	 result = QueuePair_Alloc(queuePairAllocInfo.handle,
 	                          queuePairAllocInfo.peer,
 				  queuePairAllocInfo.flags,
+                                  VMCI_NO_PRIVILEGE_FLAGS,
 				  queuePairAllocInfo.produceSize,
 				  queuePairAllocInfo.consumeSize,
 				  &pageStore,
@@ -804,6 +842,8 @@ LinuxDriver_Ioctl(struct inode *inode,
       VMCIQueuePairPageFileInfo *info = (VMCIQueuePairPageFileInfo *)ioarg;
       int32 result;
       VMCIId cid;
+      Bool useUVA;
+      int size;
 
       if (vmciLinux->ctType != VMCIOBJ_CONTEXT) {
          Log("VMCI: IOCTL_VMCI_QUEUEPAIR_SETPAGEFILE only valid for "
@@ -812,39 +852,97 @@ LinuxDriver_Ioctl(struct inode *inode,
          break;
       }
 
-      retval = copy_from_user(&pageFileInfo, (void *)ioarg,
-                              sizeof pageFileInfo);
+      if (VMCIContext_SupportsHostQP(vmciLinux->ct.context)) {
+         useUVA = TRUE;
+         size = sizeof *info;
+      } else {
+         /*
+          * An older VMX version won't supply the UVA of the page
+          * files backing the queue pair contents (and headers)
+          */
+
+         useUVA = FALSE;
+         size = sizeof(VMCIQueuePairPageFileInfo_NoHostQP);
+      }
+
+      retval = copy_from_user(&pageFileInfo, (void *)ioarg, size);
       if (retval) {
          retval = -EFAULT;
          break;
       }
 
-      cid = VMCIContext_GetId(vmciLinux->ct.context);
-      QueuePairList_Lock();
+      /*
+       * Communicate success pre-emptively to the caller.  Note that
+       * the basic premise is that it is incumbent upon the caller not
+       * to look at the info.result field until after the ioctl()
+       * returns.  And then, only if the ioctl() result indicates no
+       * error.  We send up the SUCCESS status before calling
+       * SetPageStore() store because failing to copy up the result
+       * code means unwinding the SetPageStore().
+       *
+       * It turns out the logic to unwind a SetPageStore() opens a can
+       * of worms.  For example, if a host had created the QueuePair
+       * and a guest attaches and SetPageStore() is successful but
+       * writing success fails, then ... the host has to be stopped
+       * from writing (anymore) data into the QueuePair.  That means
+       * an additional test in the VMCI_Enqueue() code path.  Ugh.
+       */
 
-      {
-	 QueuePairPageStore pageStore = { pageFileInfo.producePageFile,
-					  pageFileInfo.consumePageFile,
-					  pageFileInfo.producePageFileSize,
-					  pageFileInfo.consumePageFileSize };
-	 
-	 result = QueuePair_SetPageStore(pageFileInfo.handle,
-					 &pageStore,
-					 vmciLinux->ct.context);
-      }
-      Log("VMCI: IOCTL_VMCI_QUEUEPAIR_SETPAGEFILE cid = %u result = %d.\n",
-          cid, result);
+      result = VMCI_SUCCESS;
       retval = copy_to_user(&info->result, &result, sizeof result);
-      if (retval) {
-         retval = -EFAULT;
-         if (result >= VMCI_SUCCESS) {
-            result = QueuePair_ResetPageStore(pageFileInfo.handle,
-					      vmciLinux->ct.context);
-            ASSERT(result >= VMCI_SUCCESS);
+      if (retval == 0) {
+         cid = VMCIContext_GetId(vmciLinux->ct.context);
+         QueuePairList_Lock();
+
+         {
+            QueuePairPageStore pageStore = { TRUE,
+                                             pageFileInfo.producePageFile,
+                                             pageFileInfo.consumePageFile,
+                                             pageFileInfo.producePageFileSize,
+                                             pageFileInfo.consumePageFileSize,
+                                             useUVA ? pageFileInfo.produceVA : 0,
+                                             useUVA ? pageFileInfo.consumeVA : 0 };
+
+            result = QueuePair_SetPageStore(pageFileInfo.handle,
+                                            &pageStore,
+                                            vmciLinux->ct.context);
          }
+         QueuePairList_Unlock();
+
+         if (result < VMCI_SUCCESS) {
+            Log("VMCI: IOCTL_VMCI_QUEUEPAIR_SETPAGEFILE cid = %u result = %d.\n",
+                cid, result);
+
+            retval = copy_to_user(&info->result, &result, sizeof result);
+            if (retval != 0) {
+               /*
+                * Note that in this case the SetPageStore() call
+                * failed but we were unable to communicate that to the
+                * caller (because the copy_to_user() call failed).
+                * So, if we simply return an error (in this case
+                * -EFAULT) then the caller will know that the
+                * SetPageStore failed even though we couldn't put the
+                * result code in the result field and indicate exactly
+                * why it failed.
+                *
+                * That says nothing about the issue where we were once
+                * able to write to the caller's info memory and now
+                * can't.  Something more serious is probably going on
+                * than the fact that SetPageStore() didn't work.
+                */
+               retval = -EFAULT;
+            }
+         }
+
+      } else {
+         /*
+          * In this case, we can't write a result field of the
+          * caller's info block.  So, we don't even try to
+          * SetPageStore().
+          */
+         retval = -EFAULT;
       }
 
-      QueuePairList_Unlock();
       break;
    }
 
@@ -917,7 +1015,7 @@ LinuxDriver_Ioctl(struct inode *inode,
 
       cid = VMCIContext_GetId(vmciLinux->ct.context);
       result = VMCIDatagramRequestWellKnownMap(mapInfo.wellKnownID, cid,
-					       VMCIContext_GetPrivFlags(cid));
+					       VMCIContext_GetPrivFlagsInt(cid));
       retval = copy_to_user(&info->result, &result, sizeof result);
       if (retval) {
          retval = -EFAULT;
@@ -1097,10 +1195,38 @@ LinuxDriver_Ioctl(struct inode *inode,
 
    case IOCTL_VMCI_GET_CONTEXT_ID: {
       VMCIId cid = VMCI_HOST_CONTEXT_ID;
+
       retval = copy_to_user((void *)ioarg, &cid, sizeof cid);
       break;
    }
 
+   case IOCTL_VMCI_SET_NOTIFY: {
+      VMCISetNotifyInfo notifyInfo;
+
+      if (vmciLinux->ctType != VMCIOBJ_CONTEXT) {
+         Log("VMCI: IOCTL_VMCI_SET_NOTIFY only valid for contexts.\n");
+         retval = -EINVAL;
+         break;
+      }
+
+      retval = copy_from_user(&notifyInfo, (void *)ioarg, sizeof notifyInfo);
+      if (retval) {
+         retval = -EFAULT;
+         break;
+      }
+
+      notifyInfo.result = VMCISetupNotify(vmciLinux->ct.context,
+                                          (VA)notifyInfo.notifyUVA);
+
+      retval = copy_to_user((void *)ioarg, &notifyInfo, sizeof notifyInfo);
+      if (retval) {
+         retval = -EFAULT;
+         break;
+      }
+
+      break;
+   }
+
    default:
       Warning("Unknown ioctl %d\n", iocmd);
       retval = -EINVAL;
@@ -1142,6 +1268,221 @@ LinuxDriver_UnlockedIoctl(struct file *filp,
 #endif
 
 
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * VMCIUserVAInvalidPointer --
+ *
+ *      Checks if a given user VA is valid or not.  Copied from
+ *      bora/modules/vmnet/linux/hostif.c:VNetUserIfInvalidPointer().  TODO
+ *      libify the common code.
+ *
+ * Results:
+ *      TRUE iff invalid.
+ *
+ * Side effects:
+ *      None.
+ *
+ *-----------------------------------------------------------------------------
+ */
+
+static INLINE Bool
+VMCIUserVAInvalidPointer(VA uva,      // IN:
+                         size_t size) // IN:
+{
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
+   return !access_ok(VERIFY_WRITE, (void *)uva, size);
+#else
+   return verify_area(VERIFY_READ, (void *)uva, size) ||
+          verify_area(VERIFY_WRITE, (void *)uva, size);
+#endif
+}
+
+
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * VMCIUserVALockPage --
+ *
+ *      Lock physical page backing a given user VA.  Copied from
+ *      bora/modules/vmnet/linux/userif.c:UserIfLockPage().  TODO libify the
+ *      common code.
+ *
+ * Results:
+ *      Pointer to struct page on success, NULL otherwise.
+ *
+ * Side effects:
+ *      None.
+ *
+ *-----------------------------------------------------------------------------
+ */
+
+static INLINE struct page *
+VMCIUserVALockPage(VA addr) // IN:
+{
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 19)
+   struct page *page = NULL;
+   int retval;
+
+   down_read(&current->mm->mmap_sem);
+   retval = get_user_pages(current, current->mm, addr,
+                           1, 1, 0, &page, NULL);
+   up_read(&current->mm->mmap_sem);
+
+   if (retval != 1) {
+      return NULL;
+   }
+
+   return page;
+#else
+   struct page *page;
+   struct page *check;
+   volatile int c;
+
+   /*
+    * Establish a virtual to physical mapping by touching the physical
+    * page. Because the address is valid, there is no need to check the return
+    * value here --hpreg
+    */
+   compat_get_user(c, (char *)addr);
+
+   page = PgtblVa2Page(addr);
+   if (page == NULL) {
+      /* The mapping went away --hpreg */
+      return NULL;
+   }
+
+   /* Lock the physical page --hpreg */
+   get_page(page);
+
+   check = PgtblVa2Page(addr);
+   if (check != page) {
+      /*
+       * The mapping went away or was modified, so we didn't lock the right
+       * physical page --hpreg
+       */
+
+      /* Unlock the physical page --hpreg */
+      put_page(page);
+
+      return NULL;
+   }
+
+   /* We locked the right physical page --hpreg */
+   return page;
+#endif
+}
+
+
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * VMCIMapBoolPtr --
+ *
+ *      Lock physical page backing a given user VA and maps it to kernel
+ *      address space.  The range of the mapped memory should be within a
+ *      single page otherwise an error is returned.  Copied from
+ *      bora/modules/vmnet/linux/userif.c:VNetUserIfMapUint32Ptr().  TODO
+ *      libify the common code.
+ *
+ * Results:
+ *      0 on success, negative error code otherwise.
+ *
+ * Side effects:
+ *      None.
+ *
+ *-----------------------------------------------------------------------------
+ */
+
+static INLINE int
+VMCIMapBoolPtr(VA notifyUVA,     // IN:
+               struct page **p,  // OUT:
+               Bool **notifyPtr) // OUT:
+{
+   if (VMCIUserVAInvalidPointer(notifyUVA, sizeof **notifyPtr) ||
+       (((notifyUVA + sizeof **notifyPtr - 1) & ~(PAGE_SIZE - 1)) !=
+        (notifyUVA & ~(PAGE_SIZE - 1)))) {
+      return -EINVAL;
+   }
+
+   *p = VMCIUserVALockPage(notifyUVA);
+   if (*p == NULL) {
+      return -EAGAIN;
+   }
+
+   *notifyPtr = (Bool *)((uint8 *)kmap(*p) + (notifyUVA & (PAGE_SIZE - 1)));
+   return 0;
+}
+
+
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * VMCISetupNotify --
+ *
+ *      Sets up a given context for notify to work.  Calls VMCIMapBoolPtr()
+ *      which maps the notify boolean in user VA in kernel space.
+ *
+ * Results:
+ *      VMCI_SUCCESS on success, error code otherwise.
+ *
+ * Side effects:
+ *      None.
+ *
+ *-----------------------------------------------------------------------------
+ */
+
+static int
+VMCISetupNotify(VMCIContext *context, // IN:
+                VA notifyUVA)         // IN:
+{
+   int retval;
+
+   if (context->notify) {
+      Warning("VMCI:  Notify mechanism is already set up.\n");
+      return VMCI_ERROR_DUPLICATE_ENTRY;
+   }
+
+   retval =
+      VMCIMapBoolPtr(notifyUVA, &context->notifyPage, &context->notify) == 0 ?
+         VMCI_SUCCESS : VMCI_ERROR_GENERIC;
+   if (retval == VMCI_SUCCESS) {
+      VMCIContext_CheckAndSignalNotify(context);
+   }
+
+   return retval;
+}
+
+
+/*
+ *-----------------------------------------------------------------------------
+ *
+ * VMCIUnsetNotify --
+ *
+ *      Reverts actions set up by VMCISetupNotify().  Unmaps and unlocks the
+ *      page mapped/locked by VMCISetupNotify().
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      None.
+ *
+ *-----------------------------------------------------------------------------
+ */
+
+void
+VMCIUnsetNotify(VMCIContext *context) // IN:
+{
+   if (context->notifyPage) {
+      kunmap(context->notifyPage);
+      put_page(context->notifyPage);
+      context->notify = NULL;
+      context->notifyPage = NULL;
+   }
+}
+
+
 MODULE_AUTHOR("VMware, Inc.");
 MODULE_DESCRIPTION("VMware Virtual Machine Communication Interface (VMCI).");
 MODULE_LICENSE("GPL v2");
@@ -1152,11 +1493,3 @@ MODULE_LICENSE("GPL v2");
  * by default (i.e., neither mkinitrd nor modprobe will accept it).
  */
 MODULE_INFO(supported, "external");
-
-/*
- * Starting with SLE10sp2, Novell requires that IHVs sign a support agreement
- * with them and mark their kernel modules as externally supported via a
- * change to the module header. If this isn't done, the module will not load
- * by default (i.e., neither mkinitrd nor modprobe will accept it).
- */
-MODULE_INFO(supported, "external");
