@@ -41,6 +41,8 @@
 
 #include <linux/poll.h>
 
+#include "usercalldefs.h"
+
 /*
  * Power Management: hook resume to work around
  * BIOS bugs where VT is not properly enabled after S4
@@ -48,15 +50,12 @@
  * entering the monitor and taking a #GP at the first VMXON.
  * Workaround: at any resume, apply VT fixups.
  *
- * 2.?.? - 2.3.?  => uses apm_* power APIs
  * 2.4.0 - 2.6.10 => uses pm_* power APIs
  *     (deprecated in 2.6.11)
  * 2.?.? - present => uses struct device_driver callbacks
  *     (GPL only)
  */
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
-#   define DO_PM22
-#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11)
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11)
 #   define DO_PM24
 #else
 #   define DO_PM26
@@ -71,9 +70,6 @@
  */
 #endif
 
-#ifdef DO_PM22
-#   include <linux/apm_bios.h>
-#endif
 #ifdef DO_PM24
 #   include <linux/pm.h>
 #endif
@@ -112,6 +108,7 @@ sys_ioctl(unsigned int fd, unsigned int cmd, unsigned long arg);
 #include "cpuid_info.h"
 #include "circList.h"
 #include "x86msr.h"
+#include "iommu.h"
 
 #ifdef VMX86_DEVEL
 #include "private.h"
@@ -122,8 +119,6 @@ sys_ioctl(unsigned int fd, unsigned int cmd, unsigned long arg);
 
 #include "vmmonInt.h"
 
-#include "phystrack.h"
-
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
 int errno;       // required for compat_exit()
 #endif
@@ -136,12 +131,9 @@ typedef enum {
    LINUXDRIVERPM_RESUME,
 } LinuxDriverPMState;
 
-#if defined(DO_PM22) || defined(DO_PM24)
+#if defined(DO_PM24)
 static int LinuxDriverPMImpl(LinuxDriverPMState state);
 #endif
-#ifdef DO_PM22
-static int LinuxDriverAPMCallback(apm_event_t event);
-#endif
 #ifdef DO_PM24
 static struct pm_dev *LinuxDriverPMDev;
 static int LinuxDriverPM24Callback(struct pm_dev *dev, pm_request_t rqst, void *data);
@@ -151,6 +143,12 @@ static int LinuxDriverPM24Callback(struct pm_dev *dev, pm_request_t rqst, void *
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
 #define VMW_NOPAGE_2624
 #endif
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0) && \
+    (defined(CONFIG_SMP) || defined(CONFIG_PREEMPT))
+#  define kernel_locked_by_current() kernel_locked()
+#else
+#  define kernel_locked_by_current() 0
+#endif
 
 struct VMXLinuxState linuxState;
 
@@ -181,12 +179,9 @@ static int LinuxDriverFault(struct vm_area_struct *vma, struct vm_fault *fault);
 #elif defined(VMW_NOPAGE_261)
 static struct page *LinuxDriverNoPage(struct vm_area_struct *vma,
                            unsigned long address, int *type);
-#elif defined(KERNEL_2_4_0)
+#else
 static struct page *LinuxDriverNoPage(struct vm_area_struct *vma,
 			   unsigned long address, int unused);
-#else
-static unsigned long LinuxDriverNoPage(struct vm_area_struct *vma,
-			   unsigned long address, int write_access);
 #endif
 static int LinuxDriverMmap(struct file *filp, struct vm_area_struct *vma);
 
@@ -238,8 +233,8 @@ VMX86_RegisterMonitor(int value)
 
 #ifdef VM_X86_64
 #ifndef HAVE_COMPAT_IOCTL
-static int 
-LinuxDriver_Ioctl32_Handler(unsigned int fd, unsigned int iocmd, 
+static int
+LinuxDriver_Ioctl32_Handler(unsigned int fd, unsigned int iocmd,
                             unsigned long ioarg, struct file * filp)
 {
    int ret;
@@ -404,7 +399,7 @@ init_module(void)
    linuxState.pollTimer.data = 0;
    linuxState.pollTimer.function = LinuxDriverPollTimeout;
 
-   linuxState.fastClockThread = 0;
+   linuxState.fastClockThread = NULL;
    linuxState.fastClockRate = 0;
    linuxState.fastClockPriority = -20;
 
@@ -419,7 +414,7 @@ init_module(void)
     */
 
    memset(&vmuser_fops, 0, sizeof vmuser_fops);
-   compat_fop_set_owner(&vmuser_fops);
+   vmuser_fops.owner = THIS_MODULE;
    vmuser_fops.poll = LinuxDriverPoll;
 #ifdef VMW_HAVE_UNLOCKED_IOCTL
    vmuser_fops.unlocked_ioctl = LinuxDriver_UnlockedIoctl;
@@ -482,6 +477,7 @@ init_module(void)
    tscTimer.expires = jiffies + 4 * HZ;
    add_timer(&tscTimer);
 
+   Vmx86_InitIDList();
 
    /*
     * Workaround for buggy BIOSes that don't handle VT enable well.
@@ -493,12 +489,9 @@ init_module(void)
     * (2.6.10 kernels and LOWER; later kernels are GPL-only symbols.)
     */
    Vmx86_FixHVEnable(FALSE);
-#ifdef DO_PM22
-   apm_register_callback(&LinuxDriverAPMCallback);
-#endif
 #ifdef DO_PM24
-   LinuxDriverPMDev = pm_register(PM_UNKNOWN_DEV, 
-                                  PM_SYS_UNKNOWN, 
+   LinuxDriverPMDev = pm_register(PM_UNKNOWN_DEV,
+                                  PM_SYS_UNKNOWN,
                                   &LinuxDriverPM24Callback);
 #endif
 #ifdef DO_PM26
@@ -530,10 +523,10 @@ cleanup_module(void)
       pm_unregister(LinuxDriverPMDev);
    }
 #endif
-#ifdef DO_PM22
-   apm_unregister_callback(&LinuxDriverAPMCallback);
-#endif
 
+   if (Task_IsVMXDisabledOnAllCPUs()) {
+      Task_FreeVMCS();
+   }
    unregister_ioctl32_handlers();
 
    /*
@@ -562,6 +555,10 @@ cleanup_module(void)
 
 
    Vmx86_DestroyNUMAInfo();
+
+#ifdef HOSTED_IOMMU_SUPPORT
+   IOMMU_ModuleCleanup();
+#endif
 }
 
 
@@ -586,11 +583,8 @@ LinuxDriver_Open(struct inode *inode, // IN
 {
    VMLinux *vmLinux;
 
-   compat_mod_inc_refcount;
-
    vmLinux = kmalloc(sizeof *vmLinux, GFP_KERNEL);
    if (vmLinux == NULL) {
-      compat_mod_dec_refcount;
       return -ENOMEM;
    }
    memset(vmLinux, 0, sizeof *vmLinux);
@@ -601,6 +595,8 @@ LinuxDriver_Open(struct inode *inode, // IN
    filp->private_data = vmLinux;
    LinuxDriverQueue(vmLinux);
 
+   Vmx86_Open();
+
    return 0;
 }
 
@@ -610,8 +606,8 @@ LinuxDriver_Open(struct inode *inode, // IN
  *
  * LinuxDriverAllocPages --
  *
- *    Allocate physically contiguous block of memory with specified order.  
- *    Pages in the allocated block are configured so that caller can pass 
+ *    Allocate physically contiguous block of memory with specified order.
+ *    Pages in the allocated block are configured so that caller can pass
  *    independent pages to the VM.
  *
  * Results:
@@ -721,12 +717,18 @@ LinuxDriver_Close(struct inode *inode, // IN
    vmLinux = (VMLinux *)filp->private_data;
    ASSERT(vmLinux);
 
+#ifdef HOSTED_IOMMU_SUPPORT
+   IOMMU_VMCleanup(vmLinux);
+#endif
+
    LinuxDriverDequeue(vmLinux);
    if (vmLinux->vm != NULL) {
       Vmx86_ReleaseVM(vmLinux->vm);
       vmLinux->vm = NULL;
-   } 
-   
+   }
+
+   Vmx86_Close();
+
    /*
     * Destroy all low memory allocations.
     * We are closing the struct file here, so clearly no other process
@@ -762,7 +764,6 @@ LinuxDriver_Close(struct inode *inode, // IN
 
    kfree(vmLinux);
    filp->private_data = NULL;
-   compat_mod_dec_refcount;
    return 0;
 }
 
@@ -793,7 +794,7 @@ static INLINE_SINGLE_CALLER void
 LinuxDriverQueuePoll(void)
 {
    unsigned long flags;
-   
+
    spin_lock_irqsave(&pollQueueLock, flags);
 
    /*
@@ -1060,22 +1061,6 @@ LinuxDriverPollTimeout(unsigned long clientData)
 }
 
 
-/*
- *----------------------------------------------------------------------
- *
- * LinuxDriverIPIHandler  --
- *
- *      Null IPI handler - for monitor to notice AIO completion
- *
- *----------------------------------------------------------------------
- */
-void
-LinuxDriverIPIHandler(void *info)
-{
-   return;
-}
-
-
 /*
  *-----------------------------------------------------------------------------
  *
@@ -1102,14 +1087,10 @@ static int LinuxDriverFault(struct vm_area_struct *vma, //IN
 static struct page *LinuxDriverNoPage(struct vm_area_struct *vma, //IN
 				      unsigned long address, 	  //IN
 				      int *type)		  //OUT: Fault type
-#elif defined(KERNEL_2_4_0)
+#else
 static struct page *LinuxDriverNoPage(struct vm_area_struct *vma, //IN
 				      unsigned long address, 	  //IN
 				      int unused)		  //nothing
-#else
-static unsigned long LinuxDriverNoPage(struct vm_area_struct *vma,//IN
-				       unsigned long address,	  //IN
-				       int write_access)	  //IN
 #endif
 {
 	VMLinux *vmLinux = (VMLinux *) vma->vm_file->private_data;
@@ -1118,7 +1099,7 @@ static unsigned long LinuxDriverNoPage(struct vm_area_struct *vma,//IN
 
 #ifdef VMW_NOPAGE_2624
 	pg = fault->pgoff;
-#else	
+#else
 	pg = ((address - vma->vm_start) >> PAGE_SHIFT) + compat_vm_pgoff(vma);
 #endif
 	pg = VMMON_MAP_OFFSET(pg);
@@ -1134,13 +1115,11 @@ static unsigned long LinuxDriverNoPage(struct vm_area_struct *vma,//IN
 #ifdef VMW_NOPAGE_2624
 	fault->page = page;
 	return 0;
-#elif defined(KERNEL_2_4_0)
+#else
 #ifdef VMW_NOPAGE_261
         *type = VM_FAULT_MINOR;
 #endif
 	return page;
-#else
-	return page_address(page);
 #endif
 }
 
@@ -1211,7 +1190,7 @@ static int LinuxDriverAllocContig(VMLinux *vmLinux,
    /* 16 pages looks like a good limit... */
    if (size > VMMON_MAX_LOWMEM_PAGES) {
       return -ENOMEM;
-   } 
+   }
    /* Sorry. Only one mmap per one open. */
    down(&vmLinux->lock4Gb);
    if (vmLinux->size4Gb) {
@@ -1235,8 +1214,8 @@ static int LinuxDriverAllocContig(VMLinux *vmLinux,
    vmLinux->size4Gb = size;
    up(&vmLinux->lock4Gb);
    vma->vm_ops = &vmuser_mops;
-   /* 
-    * It seems that SuSE's 2.6.4-52 needs this.  Hopefully 
+   /*
+    * It seems that SuSE's 2.6.4-52 needs this.  Hopefully
     * it will not break anything else.
     *
     * It breaks on post 2.6.14 kernels, so get rid of it on them.
@@ -1340,7 +1319,7 @@ LinuxDriverSyncCallHook(void *data)
    SyncFuncArgs *args = (SyncFuncArgs *)data;
    unsigned cpu = smp_processor_id();
 
-   /* 
+   /*
     * We need to careful about reading cpu_online_map on kernels that
     * have hot add/remove cpu support.  The kernel's smp_call_function
     * blocks hot add from occuring between the time it computes the set
@@ -1374,7 +1353,7 @@ LinuxDriverSyncCallHook(void *data)
 
    /* Now simultaneously call the routine. */
    success = args->func(args->data, cpu);
-   
+
    if (!timeRemaining || !success) {
       /* Indicate that we either timed out or the callback failed. */
       Atomic_Inc(&args->failures);
@@ -1410,19 +1389,19 @@ LinuxDriverSyncCallOnEachCPU(SyncFunc func,
 
    args.func = func;
    args.data = data;
-   
+
    Atomic_Write(&args.numCPUs, 0); // Must be calculated inside the callback.
    Atomic_Write(&args.ready, 0);
    Atomic_Write(&args.failures, 0);
    Atomic_Write(&args.done, 0);
 
    compat_preempt_disable();
-   /* 
+   /*
     * Call all other CPUs, but do not wait so we can enter the callback
     * on this CPU too.
     */
    compat_smp_call_function(LinuxDriverSyncCallHook, &args, 0);
-   /* 
+   /*
     * smp_call_function doesn't return until all cpus have been
     * interrupted.  It's safe to disable interrupts now that all other
     * cpus are in their IPI handlers.
@@ -1435,13 +1414,13 @@ LinuxDriverSyncCallOnEachCPU(SyncFunc func,
    RESTORE_FLAGS(flags);
    compat_preempt_enable();
 
-   /* 
+   /*
     * Wait for everyone else to finish so we can get an accurate
     * failures count.
     */
    while (Atomic_Read(&args.done) != Atomic_Read(&args.numCPUs)) ;
 
-   /* 
+   /*
     * This routine failed if any CPU bailed out early to avoid deadlock,
     * or the callback routine failed on any CPU.  Both conditions are
     * recorded in the failures field.
@@ -1531,19 +1510,24 @@ LinuxDriverWriteTSC(void *data,   // IN: TSC values
 static Bool
 LinuxDriverSyncReadTSCs(TSCSet *tscSet) // OUT: TSC values
 {
-   TSCSet tmpTSCSet;
+   TSCSet *tmpTSCSet = HostIF_AllocKernelMem(sizeof *tmpTSCSet, TRUE);
    unsigned i;
    Bool okay = FALSE;
 
+   if (tmpTSCSet == NULL) {
+      return FALSE;
+   }
+
    /* Loop at least twice to warm up the cache. */
    for (i = 0; i < 1; i++) {
-      memset(&tmpTSCSet, 0, sizeof(tmpTSCSet));
-      if (LinuxDriverSyncCallOnEachCPU(LinuxDriverReadTSC, &tmpTSCSet)) {
+      memset(tmpTSCSet, 0, sizeof *tmpTSCSet);
+      if (LinuxDriverSyncCallOnEachCPU(LinuxDriverReadTSC, tmpTSCSet)) {
          /* We return the last successful simultaneous read of the TSCs. */
-         memcpy(tscSet, &tmpTSCSet, sizeof(tmpTSCSet));
+         memcpy(tscSet, tmpTSCSet, sizeof *tmpTSCSet);
          okay = TRUE;
       }
    }
+   HostIF_FreeKernelMem(tmpTSCSet);
    return okay;
 }
 
@@ -1578,6 +1562,10 @@ LinuxDriverSyncWriteTSCs(TSCSet *tscSet) // IN: TSC values
  *
  *      Main path for UserRPC
  *
+ *      Be VERY careful with stack usage; gcc's stack allocation is iffy
+ *      and allocations from individual "case" statements do not overlap,
+ *      so it is easy to use kilobytes of stack space here.
+ *
  * Results:
  *
  * Side effects:
@@ -1611,7 +1599,7 @@ LinuxDriver_Ioctl(struct inode *inode,
 	 retval = -ENOMEM;
 	 break;
       }
-      retval = vmLinux->vm->id;
+      retval = vmLinux->vm->userID;
       break;
    }
 
@@ -1665,15 +1653,15 @@ LinuxDriver_Ioctl(struct inode *inode,
    case IOCTL_VMX86_INIT_CROSSGDT: {
       InitCrossGDT initCrossGDT;
 
-      retval = HostIF_CopyFromUser(&initCrossGDT, 
-                                   (char *)ioarg, 
+      retval = HostIF_CopyFromUser(&initCrossGDT,
+                                   (char *)ioarg,
                                    sizeof initCrossGDT);
       if ((retval == 0) && Task_InitCrossGDT(&initCrossGDT)) {
          retval = -EIO;
       }
       break;
    }
-   
+
    case IOCTL_VMX86_INIT_NUMA_INFO: {
       NUMAInfoArgs *initParams;
       unsigned int numNodes;
@@ -1681,13 +1669,13 @@ LinuxDriver_Ioctl(struct inode *inode,
       retval = HostIF_CopyFromUser(&numNodes, (char*)ioarg, sizeof(uint32));
       if (retval != 0) {
 	 break;
-      } 
+      }
       if (numNodes > NUMA_MAX_NODES) {
          retval = -EINVAL;
          break;
       }
-      initParams = HostIF_AllocKernelMem(NUMA_INFO_ARGS_SIZE(*initParams, 
-                                                             numNodes), FALSE); 
+      initParams = HostIF_AllocKernelMem(NUMA_INFO_ARGS_SIZE(*initParams,
+                                                             numNodes), FALSE);
       if (!initParams) {
          retval = -EINVAL;
          break;
@@ -1742,12 +1730,11 @@ LinuxDriver_Ioctl(struct inode *inode,
          break;
       }
       unlock_kernel();
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
-#  if defined(CONFIG_SMP) || defined(CONFIG_PREEMPT)
-      ASSERT(!kernel_locked());
-#  endif
-#endif
-      retval = Vmx86_RunVM(vmLinux->vm, vcpuid);
+      if (kernel_locked_by_current()) {
+         retval = USERCALL_VMX86KRNLLOCK;
+      } else {
+         retval = Vmx86_RunVM(vmLinux->vm, vcpuid);
+      }
       lock_kernel();
       break;
 
@@ -1759,47 +1746,101 @@ LinuxDriver_Ioctl(struct inode *inode,
 #endif
       break;
 
-   case IOCTL_VMX86_LOCK_PAGE:
-   case IOCTL_VMX86_LOCK_PAGE_NEW:
-   case IOCTL_VMX86_UNLOCK_PAGE:
-   case IOCTL_VMX86_UNLOCK_PAGE_BY_MPN:
+   case IOCTL_VMX86_LOCK_PAGE: {
+      VA64 uAddr;
+      MPN mpn;
+
       if (vmLinux->vm == NULL) {
 	 retval = -EINVAL;
 	 break;
       }
-   case IOCTL_VMX86_LOOK_UP_MPN: {
-      void *addr = (void*)ioarg;
-      MPN   mpn  = 0;
-
-      switch (iocmd) { 
-      case IOCTL_VMX86_LOOK_UP_MPN:
-	 mpn = HostIF_LookupUserMPN(vmLinux->vm, addr);
-	 break;
-      case IOCTL_VMX86_LOCK_PAGE:
-	 mpn = Vmx86_LockPage(vmLinux->vm, addr, FALSE);
-	 break;
-      case IOCTL_VMX86_LOCK_PAGE_NEW: 
-	 mpn = Vmx86_LockPage(vmLinux->vm, addr, TRUE);
-	 break;
-      case IOCTL_VMX86_UNLOCK_PAGE:
-	 mpn = Vmx86_UnlockPage(vmLinux->vm, addr);
-	 break;
-      case IOCTL_VMX86_UNLOCK_PAGE_BY_MPN: {
-	 VMMUnlockPageByMPN args;
-
-	 retval = HostIF_CopyFromUser(&args, (void*)ioarg, sizeof args);
-	 if (retval != 0) {
-	    goto pageLockEnd;
-	 }
-	 mpn = Vmx86_UnlockPageByMPN(vmLinux->vm, args.mpn, (void*)(VA)args.va);
-      }
+      retval = HostIF_CopyFromUser(&uAddr, (void *)ioarg, sizeof uAddr);
+      if (retval) {
+         break;
       }
+      mpn = Vmx86_LockPage(vmLinux->vm, uAddr, FALSE);
+      retval = (int)mpn;
+      // Make sure mpn is within 32 bits.
+      ASSERT(mpn == (MPN)retval);
+   } break;
+
+   case IOCTL_VMX86_LOCK_PAGE_NEW: {
+      VA64 uAddr;
+      MPN mpn;
+
+      if (vmLinux->vm == NULL) {
+	 retval = -EINVAL;
+	 break;
+      }
+      retval = HostIF_CopyFromUser(&uAddr, (void *)ioarg, sizeof uAddr);
+      if (retval) {
+         break;
+      }
+      mpn = Vmx86_LockPage(vmLinux->vm, uAddr, TRUE);
+      retval = (int)mpn;
+      // Make sure mpn is within 32 bits.
+      ASSERT(mpn == (MPN)retval);
+   } break;
+
+   case IOCTL_VMX86_UNLOCK_PAGE: {
+      VA64 uAddr;
+      MPN mpn;
+
+      if (vmLinux->vm == NULL) {
+	 retval = -EINVAL;
+	 break;
+      }
+      retval = HostIF_CopyFromUser(&uAddr, (void *)ioarg, sizeof uAddr);
+      if (retval) {
+         break;
+      }
+      mpn = Vmx86_UnlockPage(vmLinux->vm, uAddr);
+      retval = (int)mpn;
+      // Make sure mpn is within 32 bits.
+      ASSERT(mpn == (MPN)retval);
+   } break;
+
+   case IOCTL_VMX86_UNLOCK_PAGE_BY_MPN: {
+      VMMUnlockPageByMPN args;
+      MPN mpn;
+
+      if (vmLinux->vm == NULL) {
+	 retval = -EINVAL;
+	 break;
+      }
+      retval = HostIF_CopyFromUser(&args, (void *)ioarg, sizeof args);
+      if (retval) {
+         break;
+      }
+      mpn = Vmx86_UnlockPageByMPN(vmLinux->vm, args.mpn, args.uAddr);
+      retval = (int)mpn;
+      // Make sure mpn is within 32 bits.
+      ASSERT(mpn == (MPN)retval);
+   } break;
+
+   case IOCTL_VMX86_LOOK_UP_MPN: {
+      VA64 uAddr;
+      MPN mpn;
+
+      retval = HostIF_CopyFromUser(&uAddr, (void *)ioarg, sizeof uAddr);
+      if (retval) {
+         break;
+      }
+      mpn = HostIF_LookupUserMPN(vmLinux->vm, uAddr);
+      retval = (int)mpn;
+      // Make sure mpn is within 32 bits.
+      ASSERT(mpn == (MPN)retval);
+   } break;
+
+#if defined(__linux__) && defined(VMX86_DEVEL) && defined(VM_X86_64)
+  case IOCTL_VMX86_LOOK_UP_LARGE_MPN: {
+      void *addr = (void *)ioarg;
+      MPN   mpn; 
+      mpn = HostIF_LookupLargeMPN(addr);
       retval = (int)mpn;
-      // make sure mpn is within 32 bits
-      ASSERT(mpn == (MPN)retval); 
-pageLockEnd:
       break;
    }
+#endif
 
    case IOCTL_VMX86_GET_NUM_VMS: {
       retval = Vmx86_GetNumVMs();
@@ -1834,57 +1875,90 @@ pageLockEnd:
       if (retval != 0) {
 	 break;
       }
-      Vmx86_Admit(vmLinux->vm, &args); 
+      Vmx86_Admit(vmLinux->vm, &args);
       retval = HostIF_CopyToUser((void *)ioarg, &args, sizeof args);
       break;
    }
-      
+
    case IOCTL_VMX86_READMIT: {
-      int32 pageDelta;
+      OvhdMem_Deltas delta;
       if (vmLinux->vm == NULL) {
          retval = -EINVAL;
          break;
       }
-      retval = HostIF_CopyFromUser(&pageDelta, (void *)ioarg, sizeof pageDelta);
+      retval = HostIF_CopyFromUser(&delta, (void *)ioarg, sizeof delta);
       if (retval != 0) {
          break;
       }
-      if (!Vmx86_Readmit(vmLinux->vm, pageDelta)) {
+      if (!Vmx86_Readmit(vmLinux->vm, &delta)) {
          retval = -1;
       }
 
       break;
    }
 
-   case IOCTL_VMX86_SET_MEM_USAGE: {
-      int32 usedPct;
+   case IOCTL_VMX86_UPDATE_MEM_INFO: {
+      VMMemMgmtInfoPatch patch;
       if (vmLinux->vm == NULL) {
          retval = -EINVAL;
          break;
       }
-      retval = HostIF_CopyFromUser(&usedPct, (void *)ioarg, sizeof usedPct);
+      retval = HostIF_CopyFromUser(&patch, (void *)ioarg, sizeof patch);
       if (retval != 0) {
 	 break;
       }
-      Vmx86_SetMemUsage(vmLinux->vm, usedPct);
+      Vmx86_UpdateMemInfo(vmLinux->vm, &patch);
       break;
-   } 
-   
+   }
+
    case IOCTL_VMX86_GET_MEM_INFO: {
+      VA64 uAddr;
+      VMMemInfoArgs *userVA;
+      VMMemInfoArgs in;
+      VMMemInfoArgs *out;
+
       if (vmLinux->vm == NULL) {
 	 retval = -EINVAL;
 	 break;
       }
-      if (!Vmx86_GetMemInfoCopy(vmLinux->vm, (VMMemInfoArgs *) ioarg)) {
-         retval = -EINVAL;
+
+      retval = HostIF_CopyFromUser(&uAddr, (void *)ioarg, sizeof uAddr);
+      if (retval) {
+         break;
       }
-      break;
-   }
+
+      userVA = VA64ToPtr(uAddr);
+      retval = HostIF_CopyFromUser(&in, userVA, sizeof in);
+      if (retval) {
+         break;
+      }
+
+      if (in.numVMs < 1 || in.numVMs > MAX_VMS) {
+         retval = -EINVAL;
+         break;
+      }
+      out = HostIF_AllocKernelMem(VM_GET_MEM_INFO_SIZE(in.numVMs), TRUE);
+      if (!out) {
+         retval = -ENOMEM;
+         break;
+      }
+
+      *out = in;
+      if (!Vmx86_GetMemInfoCopy(vmLinux->vm, out)) {
+         HostIF_FreeKernelMem(out);
+         retval = -ENOBUFS;
+         break;
+      }
+
+      retval = HostIF_CopyToUser(userVA, out,
+                                 VM_GET_MEM_INFO_SIZE(out->numVMs));
+      HostIF_FreeKernelMem(out);
+   } break;
 
    case IOCTL_VMX86_PAE_ENABLED:
       retval = Vmx86_PAEEnabled();
       break;
-      
+
    case IOCTL_VMX86_VMX_ENABLED:
       retval = Vmx86_VMXEnabled();
       break;
@@ -1896,51 +1970,15 @@ pageLockEnd:
       }
       retval = Vmx86_HVEnabledCPUs();
       break;
-     
+
    case IOCTL_VMX86_VT_SUPPORTED_CPU:
       retval = Vmx86_VTSupportedCPU();
       break;
-     
+
    case IOCTL_VMX86_BROKEN_CPU_HELPER:
       retval = HostIF_BrokenCPUHelper();
       break;
 
-   case IOCTL_VMX86_COUNT_PRESENT_PAGES: {
-      VMMCountPresentPages *args;
-
-      /*
-       * Allocate temp buffer in heap so we don't overflow kernel stack.
-       */
-      retval = -ENOMEM;
-      args = HostIF_AllocKernelMem(sizeof *args, FALSE);
-      if (args != NULL) {
-
-         /*
-          * Copy buffer from user space, maybe faulting it in and maybe 
-          * getting an error because it's not addressable.
-          */
-         retval = HostIF_CopyFromUser(args, (void *)ioarg, sizeof *args);
-         if (retval == 0) {
-            if (args->numPages > VMM_CPP_MAXPAGES) {
-               retval = -EINVAL;
-            } else {
-
-               /*
-                * Now it's ok to scan the list and not have to worry 
-                * about it being inaccessable.
-                */
-               retval = HostIF_CountPresentPages(args);
-            }
-         }
-
-         /*
-          * Done with temp buffer, free it off.
-          */
-         HostIF_FreeKernelMem(args);
-      }
-      break;
-   }
-
    case IOCTL_VMX86_HOST_X86_64:
 #ifdef VM_X86_64
       retval = TRUE;
@@ -1950,18 +1988,18 @@ pageLockEnd:
       break;
 
    case IOCTL_VMX86_APIC_INIT: {
-      VMAPICInfo info; 
-      Bool setVMPtr; 
+      VMAPICInfo info;
+      Bool setVMPtr;
       Bool probe;
 
       retval = HostIF_CopyFromUser(&info, (VMAPICInfo *)ioarg,
                                    sizeof info);
       if (retval != 0) {
-         break; 
+         break;
       }
       setVMPtr = ((info.flags & APIC_FLAG_DISABLE_NMI) != 0);
       probe = ((info.flags & APIC_FLAG_PROBE) != 0);
-      
+
       if (vmLinux->vm == NULL) {
 	 retval = -EINVAL;
 	 break;
@@ -1977,8 +2015,8 @@ pageLockEnd:
 
    case IOCTL_VMX86_SET_HOST_CLOCK_RATE: {
       if (vmLinux->vm == NULL) {
-	 retval = -EINVAL;
-	 break;
+         retval = -EINVAL;
+         break;
       }
       retval = -Vmx86_SetHostClockRate(vmLinux->vm, (int)ioarg);
       break;
@@ -2004,9 +2042,37 @@ pageLockEnd:
       }
       break;
 
-   case IOCTL_VMX86_BROADCAST_IPI:
-      retval = compat_smp_call_function(LinuxDriverIPIHandler, NULL, 1);
+   case IOCTL_VMX86_SEND_IPI: {
+      VCPUSet ipiTargets;
+
+      if (vmLinux->vm == NULL) {
+         retval = -EINVAL;
+         break;
+      }
+
+      retval = HostIF_CopyFromUser(&ipiTargets, (VCPUSet *) ioarg,
+                                   sizeof ipiTargets);
+
+      if (retval == 0) {
+         HostIF_IPI(vmLinux->vm, ipiTargets, TRUE);
+      }
+
       break;
+   }
+
+   case IOCTL_VMX86_GET_IPI_VECTORS: {
+      IPIVectors ipiVectors;
+
+      ipiVectors.vectors[0] = CALL_FUNCTION_VECTOR;
+#ifdef CALL_FUNCTION_SINGLE_VECTOR
+      ipiVectors.vectors[1] = CALL_FUNCTION_SINGLE_VECTOR;
+#else
+      ipiVectors.vectors[1] = 0;
+#endif
+
+      retval = HostIF_CopyToUser((void *)ioarg, &ipiVectors, sizeof ipiVectors);
+      break;
+   }
 
    case IOCTL_VMX86_GET_KHZ_ESTIMATE:
       retval = Vmx86_GetkHzEstimate(&linuxState.startTime);
@@ -2034,105 +2100,105 @@ pageLockEnd:
       Vmx86_CompleteUserCall(vmLinux->vm, vcpuid);
       break;
 
-   case IOCTL_VMX86_GET_ALL_CPUID:
-      {
-         /*
-          * XXX The whole memory manipulation logic should go to the common
-          *     code, like we do for IOCTL_VMX86_GET_MEM_INFO.
-          */
+   case IOCTL_VMX86_GET_ALL_CPUID: {
+      VA64 uAddr;
+      CPUIDQuery *userVA;
+      CPUIDQuery in;
+      CPUIDQuery *out;
 
-         CPUIDQuery *userVA = (CPUIDQuery *)ioarg;
-         CPUIDQuery in;
-         CPUIDQuery *out;
-
-         retval = HostIF_CopyFromUser(&in, userVA, sizeof in);
-         if (retval) {
-            break;
-         }
-
-	 /*
-          * Some kernels panic on kmalloc request larger than 128KB.
-          * XXX This test should go inside HostIF_AllocKernelMem() then.
-          */
-	 if (  in.numLogicalCPUs
-             > (131072 - sizeof *out) / sizeof out->logicalCPUs[0]) {
-	    retval = -EINVAL;
-	    break;
-	 }
-         out = HostIF_AllocKernelMem(
-            sizeof *out + in.numLogicalCPUs * sizeof out->logicalCPUs[0],
-            TRUE);
-         if (!out) {
-            retval = -ENOMEM;
-            break;
-         }
-
-	 *out = in;
-         if (!HostIF_GetAllCpuInfo(out)) {
-            HostIF_FreeKernelMem(out);
-            retval = -ENOBUFS;
-            break;
-         }
-
-         retval = HostIF_CopyToUser((int8 *)userVA + sizeof *userVA,
-            &out->logicalCPUs[0],
-            out->numLogicalCPUs * sizeof out->logicalCPUs[0]);
-         HostIF_FreeKernelMem(out);
+      retval = HostIF_CopyFromUser(&uAddr, (void *)ioarg, sizeof uAddr);
+      if (retval) {
+         break;
       }
-      break;
 
-   case IOCTL_VMX86_GET_ALL_MSRS:
-      {
-         /*
-          * XXX The whole memory manipulation logic should go to the common
-          *     code, like we do for IOCTL_VMX86_GET_MEM_INFO.
-          */
-
-         MSRQuery *userVA = (MSRQuery *)ioarg;
-         MSRQuery in;
-         MSRQuery *out;
-
-         retval = HostIF_CopyFromUser(&in, userVA, sizeof in);
-         if (retval) {
-            break;
-         }
-
-	 /*
-          * Some kernels panic on kmalloc request larger than 128KB.
-          * XXX This test should go inside HostIF_AllocKernelMem() then.
-          */
-	 if (  in.numLogicalCPUs
-             > (131072 - sizeof *out) / sizeof out->logicalCPUs[0]) {
-	    retval = -EINVAL;
-	    break;
-	 }
-         out = HostIF_AllocKernelMem(
-            sizeof *out + in.numLogicalCPUs * sizeof out->logicalCPUs[0],
-            TRUE);
-         if (!out) {
-            retval = -ENOMEM;
-            break;
-         }
-
-	 *out = in;
-         if (!Vmx86_GetAllMSRs(out)) {
-            HostIF_FreeKernelMem(out);
-            retval = -ENOBUFS;
-            break;
-         }
-
-         retval = HostIF_CopyToUser((int8 *)userVA + sizeof *userVA,
-            &out->logicalCPUs[0],
-            out->numLogicalCPUs * sizeof out->logicalCPUs[0]);
-         HostIF_FreeKernelMem(out);
+      userVA = VA64ToPtr(uAddr);
+      retval = HostIF_CopyFromUser(&in, userVA, sizeof in);
+      if (retval) {
+         break;
       }
-      break;
+
+      /*
+       * Some kernels panic on kmalloc request larger than 128KB.
+       * XXX This test should go inside HostIF_AllocKernelMem() then.
+       */
+      if (  in.numLogicalCPUs
+          > (131072 - sizeof *out) / sizeof out->logicalCPUs[0]) {
+         retval = -EINVAL;
+         break;
+      }
+      out = HostIF_AllocKernelMem(
+         sizeof *out + in.numLogicalCPUs * sizeof out->logicalCPUs[0],
+         TRUE);
+      if (!out) {
+         retval = -ENOMEM;
+         break;
+      }
+
+      *out = in;
+      if (!HostIF_GetAllCpuInfo(out)) {
+         HostIF_FreeKernelMem(out);
+         retval = -ENOBUFS;
+         break;
+      }
+
+      retval = HostIF_CopyToUser((int8 *)userVA + sizeof *userVA,
+         &out->logicalCPUs[0],
+         out->numLogicalCPUs * sizeof out->logicalCPUs[0]);
+      HostIF_FreeKernelMem(out);
+   } break;
+
+   case IOCTL_VMX86_GET_ALL_MSRS: {
+      VA64 uAddr;
+      MSRQuery *userVA;
+      MSRQuery in;
+      MSRQuery *out;
+
+      retval = HostIF_CopyFromUser(&uAddr, (void *)ioarg, sizeof uAddr);
+      if (retval) {
+         break;
+      }
+
+      userVA = VA64ToPtr(uAddr);
+      retval = HostIF_CopyFromUser(&in, userVA, sizeof in);
+      if (retval) {
+         break;
+      }
+
+      /*
+       * Some kernels panic on kmalloc request larger than 128KB.
+       * XXX This test should go inside HostIF_AllocKernelMem() then.
+       */
+      if (  in.numLogicalCPUs
+          > (131072 - sizeof *out) / sizeof out->logicalCPUs[0]) {
+         retval = -EINVAL;
+         break;
+      }
+      out = HostIF_AllocKernelMem(
+         sizeof *out + in.numLogicalCPUs * sizeof out->logicalCPUs[0],
+         TRUE);
+      if (!out) {
+         retval = -ENOMEM;
+         break;
+      }
+
+      *out = in;
+      if (!Vmx86_GetAllMSRs(out)) {
+         HostIF_FreeKernelMem(out);
+         retval = -ENOBUFS;
+         break;
+      }
+
+      retval = HostIF_CopyToUser((int8 *)userVA + sizeof *userVA,
+         &out->logicalCPUs[0],
+         out->numLogicalCPUs * sizeof out->logicalCPUs[0]);
+      HostIF_FreeKernelMem(out);
+   } break;
 
    case IOCTL_VMX86_ALLOC_LOCKED_PAGES:
    case IOCTL_VMX86_FREE_LOCKED_PAGES:
       {
          VMMPNList req;
-	 
+
 	 retval = HostIF_CopyFromUser(&req, (void*)ioarg, sizeof req);
 	 if (retval) {
 	    break;
@@ -2142,12 +2208,12 @@ pageLockEnd:
 	    break;
 	 }
          if (iocmd == IOCTL_VMX86_ALLOC_LOCKED_PAGES) {
-            retval = Vmx86_AllocLockedPages(vmLinux->vm, 
-                                            (MPN32*)(VA)req.mpn32List, 
+            retval = Vmx86_AllocLockedPages(vmLinux->vm,
+                                            req.mpn32List,
                                             req.mpnCount, FALSE);
          } else {
-            retval = Vmx86_FreeLockedPages(vmLinux->vm, 
-                                           (MPN32*)(VA)req.mpn32List, 
+            retval = Vmx86_FreeLockedPages(vmLinux->vm,
+                                           req.mpn32List,
                                            req.mpnCount, FALSE);
          }
       }
@@ -2156,7 +2222,7 @@ pageLockEnd:
    case IOCTL_VMX86_GET_LOCKED_PAGES_LIST:
       {
          VMMPNList req;
-	 
+
 	 retval = HostIF_CopyFromUser(&req, (void*)ioarg, sizeof req);
 	 if (retval) {
 	    break;
@@ -2165,8 +2231,8 @@ pageLockEnd:
 	    retval = -EINVAL;
 	    break;
 	 }
-	 retval = Vmx86_GetLockedPageList(vmLinux->vm, 
-                                          (MPN32*)(VA)req.mpn32List, 
+	 retval = Vmx86_GetLockedPageList(vmLinux->vm,
+                                          req.mpn32List,
                                           req.mpnCount);
       }
       break;
@@ -2176,33 +2242,33 @@ pageLockEnd:
          if (HostIF_CopyFromUser(&var, (void *)ioarg, sizeof(struct VARange)) != 0) {
             retval = -EINVAL;
          } else {
-            retval = HostIF_MarkLockedVARangeClean(vmLinux->vm, var.addr, 
+            retval = HostIF_MarkLockedVARangeClean(vmLinux->vm, var.addr,
                                                    var.len, var.bv);
          }
       }
       break;
 
-   case IOCTL_VMX86_READ_PAGE: 
+   case IOCTL_VMX86_READ_PAGE:
       {
          VMMReadWritePage req;
-	 
+
 	 retval = HostIF_CopyFromUser(&req, (void*)ioarg, sizeof req);
 	 if (retval) {
 	    break;
 	 }
-	 retval = HostIF_ReadPage(req.mpn, (void*)(VA)req.buf, FALSE);
+	 retval = HostIF_ReadPage(req.mpn, req.uAddr, FALSE);
 	 break;
       }
 
-   case IOCTL_VMX86_WRITE_PAGE: 
+   case IOCTL_VMX86_WRITE_PAGE:
       {
          VMMReadWritePage req;
-	 
+
 	 retval = HostIF_CopyFromUser(&req, (void*)ioarg, sizeof req);
 	 if (retval) {
 	    break;
 	 }
-	 retval = HostIF_WritePage(req.mpn, (void*)(VA)req.buf, FALSE);
+	 retval = HostIF_WritePage(req.mpn, req.uAddr, FALSE);
 	 break;
       }
 
@@ -2212,6 +2278,12 @@ pageLockEnd:
       break;
    }
 
+   case IOCTL_VMX86_COW_INC_ZERO_REF:
+   {
+      retval = -ENOTTY;
+      break;
+   }
+
    case IOCTL_VMX86_COW_GET_ZERO_MPN:
    {
       retval = -ENOTTY;
@@ -2229,13 +2301,7 @@ pageLockEnd:
       break;
    }
 
-   case IOCTL_VMX86_COW_COPY_PAGE: 
-   {
-      retval = -ENOTTY;
-      break;
-   } 
-   
-   case IOCTL_VMX86_COW_REMOVE_HINT: 
+   case IOCTL_VMX86_COW_COPY_PAGE:
    {
       retval = -ENOTTY;
       break;
@@ -2280,7 +2346,7 @@ pageLockEnd:
    }
 
    case IOCTL_VMX86_APIC_ID: {
-      uint8 apicId; 
+      uint8 apicId;
 
       apicId = HostIF_APIC_ID();
       retval = HostIF_CopyToUser((void *)ioarg, &apicId, sizeof apicId);
@@ -2348,9 +2414,9 @@ pageLockEnd:
       if (retval != 0) {
 	 break;
       }
-      params.usingRefClock = Vmx86_CheckPseudoTSC(&params.lastTSC, 
+      params.usingRefClock = Vmx86_CheckPseudoTSC(&params.lastTSC,
 						  &params.lastRC);
-						
+
       retval = HostIF_CopyToUser((void *)ioarg, &params, sizeof params);
       break;
    }
@@ -2380,28 +2446,55 @@ pageLockEnd:
    }
 
    case IOCTL_VMX86_SYNC_GET_TSCS: {
-      TSCSet tscSet;
-      if (!LinuxDriverSyncReadTSCs(&tscSet)) {
-         retval = -EBUSY;
-         break;
+      TSCSet *tscSet = HostIF_AllocKernelMem(sizeof *tscSet, TRUE);
+      if (tscSet != NULL) {
+         if (LinuxDriverSyncReadTSCs(tscSet)) {
+            retval = HostIF_CopyToUser((void *)ioarg, tscSet, sizeof *tscSet);
+          } else {
+            retval = -EBUSY;
+         }
+         HostIF_FreeKernelMem(tscSet);
+      } else {
+         retval = -ENOMEM;
       }
-      retval = HostIF_CopyToUser((void *)ioarg, &tscSet, sizeof tscSet);
       break;
    }
 
    case IOCTL_VMX86_SYNC_SET_TSCS: {
-      TSCSet tscSet;
-      retval = HostIF_CopyFromUser(&tscSet, (void *)ioarg, sizeof tscSet);
-      if (retval != 0) {
-         break;
-      }
-      if (!LinuxDriverSyncWriteTSCs(&tscSet)) {
-         retval = -EBUSY;
-         break;
+      TSCSet *tscSet = HostIF_AllocKernelMem(sizeof *tscSet, TRUE);
+      if (tscSet != NULL) {
+         retval = HostIF_CopyFromUser(tscSet, (void *)ioarg, sizeof *tscSet);
+         if (retval == 0) {
+            if (LinuxDriverSyncWriteTSCs(tscSet)) {
+               ; // Success
+            } else {
+               retval = -EBUSY;
+            }
+         }
+         HostIF_FreeKernelMem(tscSet);
+      } else {
+         retval = -ENOMEM;
       }
       break;
    }
-   
+
+#ifdef HOSTED_IOMMU_SUPPORT
+   case IOCTL_VMX86_IOMMU_SETUP_MMU: {
+      retval = IOMMU_SetupMMU(vmLinux, (PassthruIOMMUMap *)ioarg);
+      break;
+   }
+
+   case IOCTL_VMX86_IOMMU_REGISTER_DEVICE: {
+      retval = IOMMU_RegisterDevice(vmLinux, (uint32)ioarg);
+      break;
+   }
+
+   case IOCTL_VMX86_IOMMU_UNREGISTER_DEVICE: {
+      retval = IOMMU_UnregisterDevice((uint32)ioarg);
+      break;
+   }
+#endif
+
    default:
       Warning("Unknown ioctl %d\n", iocmd);
       retval = -EINVAL;
@@ -2553,7 +2646,7 @@ error:
 }
 
 
-#if defined(DO_PM22) || defined(DO_PM24)
+#if defined(DO_PM24)
 /*
  *----------------------------------------------------------------------
  *
@@ -2569,7 +2662,7 @@ error:
  *
  *----------------------------------------------------------------------
  */
-static int 
+static int
 LinuxDriverPMImpl(LinuxDriverPMState state)  // IN
 {
    if (state == LINUXDRIVERPM_RESUME) {
@@ -2597,7 +2690,7 @@ LinuxDriverPMImpl(LinuxDriverPMState state)  // IN
  *----------------------------------------------------------------------
  */
 #ifdef DO_PM24
-static int 
+static int
 LinuxDriverPM24Callback(struct pm_dev *dev, pm_request_t rqst, void *data)
 {
    if (rqst == PM_SUSPEND) {
@@ -2609,20 +2702,6 @@ LinuxDriverPM24Callback(struct pm_dev *dev, pm_request_t rqst, void *data)
    }
 }
 #endif
-#ifdef DO_PM22
-static int 
-LinuxDriverAPMCallback(apm_event_t event)   // IN
-{
-   switch (event) {
-   case APM_SYS_SUSPEND:
-   case APM_USER_SUSPEND:
-      return LinuxDriverPMImpl(LINUXDRIVERPM_SUSPEND);
-   case APM_NORMAL_RESUME:
-      return LinuxDriverPMImpl(LINUXDRIVERPM_RESUME);
-   }
-   return 0;
-}
-#endif
 
 MODULE_AUTHOR("VMware, Inc.");
 MODULE_DESCRIPTION("VMware Virtual Machine Monitor.");
@@ -2634,4 +2713,3 @@ MODULE_LICENSE("GPL v2");
  * by default (i.e., neither mkinitrd nor modprobe will accept it).
  */
 MODULE_INFO(supported, "external");
-
