@@ -21,8 +21,8 @@
  * It appeared between 2.6.25 and 2.6.26-rc1.
  */
 
-#include <linux/autoconf.h>
-#include <linux/version.h>
+#include "compat_version.h"
+#include "compat_autoconf.h"
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 25)
 #   error This compile test intentionally fails.
