@@ -23,11 +23,7 @@
 #include <linux/kernel.h>
 #include <linux/version.h>
 #include <linux/sched.h>
-#ifdef KERNEL_2_2
-#   include <linux/slab.h>
-#else
-#   include <linux/malloc.h>
-#endif
+#include <linux/slab.h>
 #include <linux/poll.h>
 
 #include <linux/netdevice.h>
@@ -68,16 +64,14 @@ static struct net_device_stats *VNetNetifGetStats(struct net_device *dev);
 static int  VNetNetifSetMAC(struct net_device *dev, void *addr);
 static void VNetNetifSetMulticast(struct net_device *dev);
 #if 0
-#ifdef KERNEL_2_3_43
 static void VNetNetifTxTimeout(struct net_device *dev);
 #endif
-#endif
 
 static int  VNetNetIfProcRead(char *page, char **start, off_t off,
                               int count, int *eof, void *data);
 
+
 #if 0
-#ifdef KERNEL_2_3_43
 /*
  *----------------------------------------------------------------------
  *
@@ -108,7 +102,6 @@ VNetNetifTxTimeout(struct net_device *dev) // IN:
    netif_wake_queue(dev);
 }
 #endif
-#endif
 
 
 /*
@@ -160,7 +153,6 @@ VNetNetIfSetup(struct net_device *dev)  // IN:
    dev->get_stats = VNetNetifGetStats;
    dev->set_mac_address = VNetNetifSetMAC;
    dev->set_multicast_list = VNetNetifSetMulticast;
-#ifdef KERNEL_2_3_43
    /*
     * We cannot stuck... If someone will report problems under
     * low memory conditions or some such, we should enable it.
@@ -168,7 +160,6 @@ VNetNetIfSetup(struct net_device *dev)  // IN:
 #if 0
    dev->tx_timeout = VNetNetifTxTimeout;
 #endif
-#endif
 #endif /* HAVE_NET_DEVICE_OPS */
 
 #if 0
@@ -480,12 +471,6 @@ VNetNetifOpen(struct net_device *dev) // IN:
     */
 
    netif_start_queue(dev);
-#ifndef KERNEL_2_3_43
-   /* Softnet does not have interrupt hack ... */
-   dev->interrupt = 0;
-   /* ... and sets start for us */
-   dev->start = 1;
-#endif
    // xxx need to change flags
    return 0;
 }
@@ -533,10 +518,6 @@ VNetNetifProbe(struct net_device *dev) // IN: unused
 int
 VNetNetifClose(struct net_device *dev) // IN:
 {
-#ifndef KERNEL_2_3_43
-   /* Softnet generic layer clears it for us */
-   dev->start = 0;
-#endif
    netif_stop_queue(dev);
    // xxx need to change flags
    return 0;
@@ -577,20 +558,9 @@ VNetNetifStartXmit(struct sk_buff    *skb, // IN:
     * If this ever occurs the queue layer is doing something evil!
     */
 
-#ifndef KERNEL_2_3_43
-   /* Softnet does not play with tbusy... */
-   if (test_and_set_bit(0, (void*)&dev->tbusy) != 0)
-   {
-      LOG(0, (KERN_NOTICE "%s: transmitter access conflict.\n", dev->name));
-      return 1;
-   }
-#endif
    VNetSend(&netIf->port.jack, skb);
 
    netIf->stats.tx_packets++;
-#ifndef KERNEL_2_3_43
-   dev->tbusy = 0;
-#endif
    dev->trans_start = jiffies;
 
    return 0;
