@@ -21,8 +21,8 @@
  * This API change was introduced between 2.6.12-rc1 and 2.6.12-rc2.
  */
 
-#include <linux/autoconf.h>
-#include <linux/version.h>
+#include "compat_version.h"
+#include "compat_autoconf.h"
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 13)
 #include <net/sock.h>
