@@ -23,11 +23,7 @@
 #include <linux/kernel.h>
 #include <linux/version.h>
 #include <linux/sched.h>
-#ifdef KERNEL_2_2
-#   include <linux/slab.h>
-#else
-#   include <linux/malloc.h>
-#endif
+#include <linux/slab.h>
 #include <linux/poll.h>
 
 #include <linux/netdevice.h>
@@ -702,8 +698,7 @@ VNetUserIfRead(VNetPort    *port, // IN
 
    add_wait_queue(&userIf->waitQueue, &wait);
    for (;;) {
-      current->state = TASK_INTERRUPTIBLE;
-      mb();
+      set_current_state(TASK_INTERRUPTIBLE);
       skb = skb_peek(&userIf->packetQueue);
       if (skb && (skb->len > count)) {
          skb = NULL;
@@ -740,7 +735,7 @@ VNetUserIfRead(VNetPort    *port, // IN
       }
       schedule();
    }
-   current->state = TASK_RUNNING;
+   __set_current_state(TASK_RUNNING);
    remove_wait_queue(&userIf->waitQueue, &wait);
    if (! skb) {
       return ret;
